@@ -84,10 +84,10 @@ filetype_builtin(const call_info_t *call_info)
 	{
 #ifndef _WIN32
 		const mode_t mode = curr_view->dir_entry[fnum].mode;
-		var_val.string = (char *)get_mode_str(mode);
+		var_val.const_string = get_mode_str(mode);
 #else
 		const FileType type = curr_view->dir_entry[fnum].type;
-		var_val.string = (char *)get_type_str(type);
+		var_val.const_string = get_type_str(type);
 #endif
 	}
 	return var_new(VTYPE_STRING, var_val);
