@@ -31,8 +31,9 @@
 #define CP_RC "cp " PACKAGE_DATA_DIR "/" VIFMRC " ~/.vifm"
 #endif
 
+#include <unistd.h> /* _SC_ARG_MAX sysconf() */
+
 #include <assert.h> /* assert() */
-#include <errno.h>
 #include <limits.h> /* INT_MIN */
 #include <stddef.h> /* size_t */
 #include <stdio.h> /* FILE snprintf() */
@@ -57,12 +58,21 @@
 #include "../filelist.h"
 #include "../opt_handlers.h"
 #include "../status.h"
+#include "../trash.h"
 #include "../ui.h"
 #include "hist.h"
 
 /* Maximum supported by the implementation length of line in vifmrc file. */
 #define MAX_VIFMRC_LINE_LEN 4*1024
 
+/* Default value of shell command used if $SHELL environment variable isn't
+ * set. */
+#ifndef _WIN32
+#define DEFAULT_SHELL_CMD "sh"
+#else
+#define DEFAULT_SHELL_CMD "cmd"
+#endif
+
 config_t cfg;
 
 static void find_home_dir(void);
@@ -100,7 +110,6 @@ static void save_into_history(const char item[], hist_t *hist, int len);
 void
 init_config(void)
 {
-	cfg.num_bookmarks = 0;
 	cfg.vim_filter = 0;
 	cfg.show_one_window = 0;
 	cfg.history_len = 15;
@@ -181,11 +190,7 @@ init_config(void)
 	strcat(cfg.log_file, "/startup-log");
 #endif
 
-#ifndef _WIN32
-	cfg.shell = strdup(env_get_def("SHELL", "sh"));
-#else
-	cfg.shell = strdup(env_get_def("SHELL", "cmd"));
-#endif
+	cfg_set_shell(env_get_def("SHELL", DEFAULT_SHELL_CMD));
 
 #ifndef _WIN32
 	/* Maximum argument length to pass to the shell */
@@ -437,6 +442,8 @@ store_config_paths(void)
 	snprintf(cfg.config_dir, sizeof(cfg.config_dir), "%s", env_get(VIFM_EV));
 	snprintf(cfg.trash_dir, sizeof(cfg.trash_dir), "%s/" TRASH, cfg.config_dir);
 	snprintf(cfg.log_file, sizeof(cfg.log_file), "%s/" LOG, cfg.config_dir);
+
+	(void)set_trash_dir(cfg.trash_dir);
 }
 
 /* ensures existence of configuration directory */
@@ -484,7 +491,7 @@ create_help_file(void)
 	LOG_FUNC_ENTER;
 
 	char command[] = CP_HELP;
-	(void)my_system(command);
+	(void)vifm_system(command);
 }
 
 /* Copies example vifmrc file from shared files to the ~/.vifm directory. */
@@ -494,7 +501,7 @@ create_rc_file(void)
 	LOG_FUNC_ENTER;
 
 	char command[] = CP_RC;
-	(void)my_system(command);
+	(void)vifm_system(command);
 }
 #endif
 
@@ -504,28 +511,8 @@ add_default_bookmarks(void)
 {
 	LOG_FUNC_ENTER;
 
-	add_bookmark('H', cfg.home_dir, "../");
-	add_bookmark('z', cfg.config_dir, "../");
-}
-
-int
-create_trash_dir(const char trash_dir[])
-{
-	LOG_FUNC_ENTER;
-
-	if(is_dir_writable(trash_dir))
-	{
-		return 0;
-	}
-
-	if(make_dir(trash_dir, 0777) != 0)
-	{
-		show_error_msgf("Error Setting Trash Directory",
-				"Could not set trash directory to %s: %s", trash_dir, strerror(errno));
-		return 1;
-	}
-
-	return 0;
+	set_user_bookmark('H', cfg.home_dir, "../");
+	set_user_bookmark('z', cfg.config_dir, "../");
 }
 
 void
@@ -853,23 +840,6 @@ set_fuse_home(const char new_value[])
 	return replace_string(&cfg.fuse_home, canonicalized);
 }
 
-int
-set_trash_dir(const char new_value[])
-{
-	if(!is_path_absolute(new_value))
-	{
-		show_error_msgf("Error Setting Trash Directory",
-				"The path is not absolute: %s", new_value);
-		return 1;
-	}
-	if(create_trash_dir(new_value) != 0)
-	{
-		return 1;
-	}
-	snprintf(cfg.trash_dir, sizeof(cfg.trash_dir), "%s", new_value);
-	return 0;
-}
-
 void
 set_use_term_multiplexer(int use_term_multiplexer)
 {
@@ -926,5 +896,14 @@ save_into_history(const char item[], hist_t *hist, int len)
 	}
 }
 
+void
+cfg_set_shell(const char shell[])
+{
+	if(replace_string(&cfg.shell, shell) == 0)
+	{
+		stats_update_shell_type(cfg.shell);
+	}
+}
+
 /* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab cinoptions-=(0 : */
 /* vim: set cinoptions+=t0 : */
