@@ -48,7 +48,7 @@ enum
 	DECORATION_SUFFIX, /* The symbol, which is appended to file name. */
 };
 
-typedef struct
+typedef struct config_t
 {
 	char home_dir[PATH_MAX]; /* ends with a slash */
 	char config_dir[PATH_MAX];
@@ -59,7 +59,6 @@ typedef struct
 	int vi_cmd_bg;
 	char *vi_x_command;
 	int vi_x_cmd_bg;
-	int num_bookmarks; /* Number of active bookmarks (set at the moment) */
 	int use_trash;
 	int vim_filter;
 
@@ -77,7 +76,7 @@ typedef struct
 	char *time_format;
 	char *fuse_home; /* This one should be set using set_fuse_home() function. */
 
-	/* History command-line commands. */
+	/* History of command-line commands. */
 	hist_t cmd_hist;
 	/* History of search patterns. */
 	hist_t search_hist;
@@ -143,17 +142,11 @@ int are_old_color_schemes(void);
 const char * get_vicmd(int *bg);
 /* Generates name of file inside tmp folder. */
 void generate_tmp_file_name(const char prefix[], char buf[], size_t buf_len);
-/* Ensures existence of trash directory.  Returns zero on success, otherwise
- * non-zero value is returned. */
-int create_trash_dir(const char trash_dir[]);
 /* Changes size of all histories. */
 void resize_history(size_t new_len);
 /* Sets value of cfg.fuse_home.  Returns non-zero in case of error, otherwise
  * zero is returned. */
 int set_fuse_home(const char new_value[]);
-/* Sets value of cfg.trash_dir.  Returns non-zero in case of error, otherwise
- * zero is returned. */
-int set_trash_dir(const char new_value[]);
 /* Sets whether support of terminal multiplexers is enabled. */
 void set_use_term_multiplexer(int use_term_multiplexer);
 /* Frees memory previously allocated for specified history items. */
@@ -167,6 +160,9 @@ void save_prompt_history(const char input[]);
 /* Saves input to local filter history. */
 void save_filter_history(const char pattern[]);
 
+/* Sets shell invocation command. */
+void cfg_set_shell(const char shell[]);
+
 #endif /* VIFM__CFG__CONFIG_H__ */
 
 /* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab cinoptions-=(0 : */
