@@ -21,8 +21,9 @@
 
 #include <assert.h> /* assert() */
 #include <ctype.h> /* isdigit() */
+#include <stddef.h> /* NULL size_t */
 #include <stdio.h> /* fscanf() fgets() fputc() snprintf() */
-#include <stdlib.h> /* realloc() */
+#include <stdlib.h> /* free() realloc() */
 #include <string.h> /* memset() strtol() strcmp() strchr() strlen() */
 
 #include "../engine/cmds.h"
@@ -31,6 +32,7 @@
 #include "../utils/fs.h"
 #include "../utils/fs_limits.h"
 #include "../utils/log.h"
+#include "../utils/path.h"
 #include "../utils/str.h"
 #include "../utils/string_array.h"
 #include "../utils/utils.h"
@@ -50,19 +52,21 @@
 static void get_sort_info(FileView *view, const char line[]);
 static void append_to_history(hist_t *hist, void (*saver)(const char[]),
 		const char item[]);
-static void ensure_history_not_full(hist_t *hist, int *len);
+static void ensure_history_not_full(hist_t *hist);
 static void get_history(FileView *view, int reread, const char *dir,
 		const char *file, int pos);
 static void set_view_property(FileView *view, char type, const char value[]);
 static int copy_file(const char src[], const char dst[]);
 static int copy_file_internal(FILE *const src, FILE *const dst);
 static void update_info_file(const char filename[]);
+static char * convert_old_trash_path(const char trash_path[]);
 static void write_options(FILE *const fp);
 static void write_assocs(FILE *fp, const char str[], char mark,
 		assoc_list_t *assocs, int prev_count, char *prev[]);
 static void write_commands(FILE *const fp, char *cmds_list[], char *cmds[],
 		int ncmds);
-static void write_bookmarks(FILE *const fp, char *marks[], int nmarks);
+static void write_bookmarks(FILE *const fp, const char non_conflicting_bmarks[],
+		char *marks[], const int timestamps[], int nmarks);
 static void write_tui_state(FILE *const fp);
 static void write_view_history(FILE *fp, FileView *view, const char str[],
 		char mark, int prev_count, char *prev[], int pos[]);
@@ -76,7 +80,7 @@ static char * read_vifminfo_line(FILE *fp, char buffer[]);
 static void remove_leading_whitespace(char line[]);
 static const char * escape_spaces(const char *str);
 static void put_sort_info(FILE *fp, char leading_char, const FileView *view);
-static int read_possible_pos(FILE *f);
+static int read_optional_number(FILE *f);
 static size_t add_to_int_array(int **array, size_t len, int what);
 
 void
@@ -160,7 +164,8 @@ read_info_file(int reread)
 			{
 				if((line3 = read_vifminfo_line(fp, line3)) != NULL)
 				{
-					add_bookmark(line_val[0], line2, line3);
+					const int timestamp = read_optional_number(fp);
+					setup_user_bookmark(line_val[0], line2, line3, timestamp);
 				}
 			}
 		}
@@ -169,8 +174,8 @@ read_info_file(int reread)
 			/* don't change active view on :restart command */
 			if(line_val[0] == 'r' && !reread)
 			{
-				update_view_title(&lwin);
-				update_view_title(&rwin);
+				ui_views_update_titles();
+
 				curr_view = &rwin;
 				other_view = &lwin;
 			}
@@ -215,7 +220,7 @@ read_info_file(int reread)
 			}
 			else if((line2 = read_vifminfo_line(fp, line2)) != NULL)
 			{
-				const int pos = read_possible_pos(fp);
+				const int pos = read_optional_number(fp);
 				get_history(view, reread, line_val, line2, pos);
 			}
 		}
@@ -252,9 +257,9 @@ read_info_file(int reread)
 		{
 			if((line2 = read_vifminfo_line(fp, line2)) != NULL)
 			{
-				if(!path_exists_at(cfg.trash_dir, line_val))
-					continue;
-				add_to_trash(line2, line_val);
+				char *const trash_name = convert_old_trash_path(line_val);
+				(void)add_to_trash(line2, trash_name);
+				free(trash_name);
 			}
 		}
 		else if(type == LINE_TYPE_REG)
@@ -263,13 +268,13 @@ read_info_file(int reread)
 		}
 		else if(type == LINE_TYPE_LWIN_FILT)
 		{
-			(void)replace_string(&lwin.prev_name_filter, line_val);
-			(void)filter_set(&lwin.name_filter, line_val);
+			(void)replace_string(&lwin.prev_manual_filter, line_val);
+			(void)filter_set(&lwin.manual_filter, line_val);
 		}
 		else if(type == LINE_TYPE_RWIN_FILT)
 		{
-			(void)replace_string(&rwin.prev_name_filter, line_val);
-			(void)filter_set(&rwin.name_filter, line_val);
+			(void)replace_string(&rwin.prev_manual_filter, line_val);
+			(void)filter_set(&rwin.manual_filter, line_val);
 		}
 		else if(type == LINE_TYPE_LWIN_FILT_INV)
 		{
@@ -325,10 +330,7 @@ get_sort_info(FileView *view, const char line[])
 		{
 			line++;
 		}
-		while(*line == ',')
-		{
-			line++;
-		}
+		line = skip_char(line, ',');
 	}
 	memset(&view->sort[j], NO_SORT_OPTION, sizeof(view->sort) - j);
 
@@ -340,26 +342,20 @@ static void
 append_to_history(hist_t *hist, void (*saver)(const char[]),
 		const char item[])
 {
-	ensure_history_not_full(hist, &cfg.history_len);
+	ensure_history_not_full(hist);
 	saver(item);
 }
 
 /* Checks that history has at least one more empty slot or extends history by
  * one more element. */
 static void
-ensure_history_not_full(hist_t *hist, int *len)
+ensure_history_not_full(hist_t *hist)
 {
-	void *p;
-
-	if(hist->pos != *len)
-		return;
-
-	p = realloc(hist->items, sizeof(char*)*(*len + 1));
-	if(p == NULL)
-		return;
-
-	++*len;
-	hist->items = p;
+	if(hist->pos + 1 == cfg.history_len)
+	{
+		resize_history(MAX(0, cfg.history_len + 1));
+		assert(hist->pos + 1 != cfg.history_len && "Failed to resize history.");
+	}
 }
 
 static void
@@ -485,14 +481,14 @@ update_info_file(const char filename[])
 	int ncmds_list = -1;
 	char **ft = NULL, **fx = NULL, **fv = NULL, **cmds = NULL, **marks = NULL;
 	char **lh = NULL, **rh = NULL, **cmdh = NULL, **srch = NULL, **regs = NULL;
-	int *lhp = NULL, *rhp = NULL;
-	size_t nlhp = 0, nrhp = 0;
+	int *lhp = NULL, *rhp = NULL, *bt = NULL;
+	size_t nlhp = 0UL, nrhp = 0UL, nbt = 0UL;
 	char **prompt = NULL, **filter = NULL, **trash = NULL;
 	int nft = 0, nfx = 0, nfv = 0, ncmds = 0, nmarks = 0, nlh = 0, nrh = 0;
 	int ncmdh = 0, nsrch = 0, nregs = 0, nprompt = 0, nfilter = 0, ntrash = 0;
 	char **dir_stack = NULL;
 	int ndir_stack = 0;
-	int i;
+	char *non_conflicting_bmarks;
 
 	if(cfg.vifm_info == 0)
 		return;
@@ -500,6 +496,8 @@ update_info_file(const char filename[])
 	cmds_list = list_udf();
 	while(cmds_list[++ncmds_list] != NULL);
 
+	non_conflicting_bmarks = strdup(valid_bookmarks);
+
 	if((fp = fopen(filename, "r")) != NULL)
 	{
 		char *line = NULL, *line2 = NULL, *line3 = NULL, *line4 = NULL;
@@ -561,6 +559,7 @@ update_info_file(const char filename[])
 					continue;
 				if((line2 = read_vifminfo_line(fp, line2)) != NULL)
 				{
+					int i;
 					const char *p = line_val;
 					for(i = 0; i < ncmds_list; i += 2)
 					{
@@ -589,7 +588,7 @@ update_info_file(const char filename[])
 					if(is_in_view_history(&lwin, line_val))
 						continue;
 
-					pos = read_possible_pos(fp);
+					pos = read_optional_number(fp);
 					nlh = add_to_string_array(&lh, nlh, 2, line_val, line2);
 					if(nlh/2 > nlhp)
 					{
@@ -611,7 +610,7 @@ update_info_file(const char filename[])
 					if(is_in_view_history(&rwin, line_val))
 						continue;
 
-					pos = read_possible_pos(fp);
+					pos = read_optional_number(fp);
 					nrh = add_to_string_array(&rh, nrh, 2, line_val, line2);
 					if(nrh/2 > nrhp)
 					{
@@ -631,13 +630,26 @@ update_info_file(const char filename[])
 				{
 					if((line3 = read_vifminfo_line(fp, line3)) != NULL)
 					{
+						const int timestamp = read_optional_number(fp);
 						const char mark_str[] = { mark, '\0' };
+
 						if(!char_is_one_of(valid_bookmarks, mark))
+						{
 							continue;
-						if(!is_bookmark_empty(mark2index(mark)))
-							continue;
-						nmarks = add_to_string_array(&marks, nmarks, 3, mark_str, line2,
-								line3);
+						}
+
+						if(is_bookmark_older(mark, timestamp))
+						{
+							char *const pos = strchr(non_conflicting_bmarks, mark);
+							if(pos != NULL)
+							{
+								nmarks = add_to_string_array(&marks, nmarks, 3, mark_str, line2,
+										line3);
+								nbt = add_to_int_array(&bt, nbt, timestamp);
+
+								*pos = '\xff';
+							}
+						}
 					}
 				}
 			}
@@ -645,11 +657,12 @@ update_info_file(const char filename[])
 			{
 				if((line2 = read_vifminfo_line(fp, line2)) != NULL)
 				{
-					if(!path_exists_at(cfg.trash_dir, line_val))
-						continue;
-					if(is_in_trash(line_val))
-						continue;
-					ntrash = add_to_string_array(&trash, ntrash, 2, line_val, line2);
+					char *const trash_name = convert_old_trash_path(line_val);
+					if(exists_in_trash(trash_name) && !is_in_trash(trash_name))
+					{
+						ntrash = add_to_string_array(&trash, ntrash, 2, trash_name, line2);
+					}
+					free(trash_name);
 				}
 			}
 			else if(type == LINE_TYPE_CMDLINE_HIST)
@@ -734,7 +747,7 @@ update_info_file(const char filename[])
 
 		if(cfg.vifm_info & VIFMINFO_BOOKMARKS)
 		{
-			write_bookmarks(fp, marks, nmarks);
+			write_bookmarks(fp, non_conflicting_bmarks, marks, bt, nmarks);
 		}
 
 		if(cfg.vifm_info & VIFMINFO_TUI)
@@ -808,12 +821,32 @@ update_info_file(const char filename[])
 	free_string_array(rh, nrh);
 	free(lhp);
 	free(rhp);
+	free(bt);
 	free_string_array(cmdh, ncmdh);
 	free_string_array(srch, nsrch);
 	free_string_array(regs, nregs);
 	free_string_array(prompt, nprompt);
 	free_string_array(trash, ntrash);
 	free_string_array(dir_stack, ndir_stack);
+	free(non_conflicting_bmarks);
+}
+
+/* Performs conversions on files in trash required for partial backward
+ * compatibility.  Returns newly allocated string that should be freed by the
+ * caller. */
+static char *
+convert_old_trash_path(const char trash_path[])
+{
+	if(!is_path_absolute(trash_path) && is_dir_writable(cfg.trash_dir))
+	{
+		char *const full_path = format_str("%s/%s", cfg.trash_dir, trash_path);
+		if(path_exists(full_path))
+		{
+			return full_path;
+		}
+		free(full_path);
+	}
+	return strdup(trash_path);
 }
 
 /* Writes current values of all options into vifminfo file. */
@@ -869,6 +902,12 @@ write_options(FILE *const fp)
 	fprintf(fp, "=]viewcolumns=%s\n", escape_spaces(rwin.view_columns));
 	fprintf(fp, "=[%slsview\n", lwin.ls_view ? "" : "no");
 	fprintf(fp, "=]%slsview\n", rwin.ls_view ? "" : "no");
+	fprintf(fp, "=[%snumber\n", (lwin.num_type & NT_SEQ) ? "" : "no");
+	fprintf(fp, "=]%snumber\n", (rwin.num_type & NT_SEQ) ? "" : "no");
+	fprintf(fp, "=[numberwidth=%d\n", lwin.num_width);
+	fprintf(fp, "=]numberwidth=%d\n", rwin.num_width);
+	fprintf(fp, "=[%srelativenumber\n", (lwin.num_type & NT_REL) ? "" : "no");
+	fprintf(fp, "=]%srelativenumber\n", (rwin.num_type & NT_REL) ? "" : "no");
 
 	fprintf(fp, "%s", "=dotdirs=");
 	if(cfg.dot_dirs & DD_ROOT_PARENT)
@@ -973,25 +1012,34 @@ write_commands(FILE *const fp, char *cmds_list[], char *cmds[], int ncmds)
 /* Writes bookmarks to vifminfo file.  marks is a list of length nmarks
  * bookmarks read from vifminfo. */
 static void
-write_bookmarks(FILE *const fp, char *marks[], int nmarks)
+write_bookmarks(FILE *const fp, const char non_conflicting_bmarks[],
+		char *marks[], const int timestamps[], int nmarks)
 {
-	const int len = init_active_bookmarks(valid_bookmarks);
+	int active_bookmarks[NUM_BOOKMARKS];
+	const int len = init_active_bookmarks(valid_bookmarks, active_bookmarks);
 	int i;
 
 	fputs("\n# Bookmarks:\n", fp);
 	for(i = 0; i < len; i++)
 	{
 		const int index = active_bookmarks[i];
-		if(!is_spec_bookmark(index))
+		const char mark = index2mark(index);
+		if(!is_spec_bookmark(index) && char_is_one_of(non_conflicting_bmarks, mark))
 		{
-			fprintf(fp, "'%c\n\t%s\n\t", index2mark(index),
-					bookmarks[index].directory);
-			fprintf(fp, "%s\n", bookmarks[index].file);
+			const bookmark_t *const bookmark = &bookmarks[index];
+
+			fprintf(fp, "'%c\n", mark);
+			fprintf(fp, "\t%s\n", bookmark->directory);
+			fprintf(fp, "\t%s\n", bookmark->file);
+			fprintf(fp, TIME_T "\n", bookmark->timestamp);
 		}
 	}
 	for(i = 0; i < nmarks; i += 3)
 	{
-		fprintf(fp, "'%c\n\t%s\n\t%s\n", marks[i][0], marks[i + 1], marks[i + 2]);
+		fprintf(fp, "'%c\n", marks[i][0]);
+		fprintf(fp, "\t%s\n", marks[i + 1]);
+		fprintf(fp, "\t%s\n", marks[i + 2]);
+		fprintf(fp, "%d\n", timestamps[i/3]);
 	}
 }
 
@@ -1127,11 +1175,11 @@ static void
 write_general_state(FILE *const fp)
 {
 	fputs("\n# State:\n", fp);
-	fprintf(fp, "f%s\n", lwin.name_filter.raw);
+	fprintf(fp, "f%s\n", lwin.manual_filter.raw);
 	fprintf(fp, "i%d\n", lwin.invert);
 	fprintf(fp, "[.%d\n", lwin.hide_dot);
 	fprintf(fp, "[F%s\n", lwin.auto_filter.raw);
-	fprintf(fp, "F%s\n", rwin.name_filter.raw);
+	fprintf(fp, "F%s\n", rwin.manual_filter.raw);
 	fprintf(fp, "I%d\n", rwin.invert);
 	fprintf(fp, "].%d\n", rwin.hide_dot);
 	fprintf(fp, "]F%s\n", rwin.auto_filter.raw);
@@ -1202,23 +1250,23 @@ put_sort_info(FILE *fp, char leading_char, const FileView *view)
 /* Ensures that the next character of the stream is a digit and reads a number.
  * Returns read number or -1 in case there is no digit. */
 static int
-read_possible_pos(FILE *f)
+read_optional_number(FILE *f)
 {
-	int pos = -1;
+	int num = -1;
 	const int c = getc(f);
 
 	if(c != EOF)
 	{
 		ungetc(c, f);
-		if(isdigit(c))
+		if(isdigit(c) || c == '-' || c == '+')
 		{
-			const int nread = fscanf(f, "%d\n", &pos);
+			const int nread = fscanf(f, "%d\n", &num);
 			assert(nread == 1 && "Wrong number of read numbers.");
 			(void)nread;
 		}
 	}
 
-	return pos;
+	return num;
 }
 
 static size_t
