@@ -37,7 +37,7 @@
 #include <stddef.h> /* NULL size_t */
 #include <stdlib.h> /* free() */
 #include <stdio.h> /* snprintf() */
-#include <string.h> /* strdup() strlen() strncasecmp() strncmp() */
+#include <string.h> /* strdup() strlen() strncasecmp() strncmp() strrchr() */
 
 #include "cfg/config.h"
 #include "engine/completion.h"
@@ -79,7 +79,7 @@ static void filename_completion_in_dir(const char *path, const char *str,
 		CompletionType type);
 static void filename_completion_internal(DIR * dir, const char * dirname,
 		const char * filename, CompletionType type);
-static void add_filename_completion(const char * filename, CompletionType type);
+static void add_filename_completion(const char filename[], CompletionType type);
 static int is_entry_dir(const struct dirent *d);
 static int is_entry_exec(const struct dirent *d);
 #ifdef _WIN32
@@ -98,16 +98,28 @@ complete_args(int id, const char args[], int argc, char *argv[], int arg_pos)
 	const char *start;
 	const char *slash;
 	const char *dollar;
+	const char *ampersand;
 
 	arg = after_last(args, ' ');
 	start = arg;
-	dollar = strrchr(arg, '$');
 	slash = strrchr(args + arg_pos, '/');
+	dollar = strrchr(arg, '$');
+	ampersand = strrchr(arg, '&');
 
 	if(id == COM_SET)
 		complete_options(args, &start);
-	else if(id == COM_LET || id == COM_ECHO || id == COM_EXE)
-		complete_variables((dollar > arg) ? dollar : arg, &start);
+	else if(command_accepts_expr(id))
+	{
+		if(ampersand > dollar)
+		{
+			start = ampersand + 1;
+			complete_real_option_names(ampersand + 1);
+		}
+		else
+		{
+			complete_variables((dollar > arg) ? dollar : arg, &start);
+		}
+	}
 	else if(id == COM_UNLET)
 		complete_variables(arg, &start);
 	else if(id == COM_HELP)
@@ -553,9 +565,10 @@ complete_command_name(const char beginning[])
 	paths = get_paths(&paths_count);
 	for(i = 0; i < paths_count; i++)
 	{
-		if(my_chdir(paths[i]) != 0)
-			continue;
-		filename_completion(beginning, CT_EXECONLY);
+		if(vifm_chdir(paths[i]) == 0)
+		{
+			filename_completion(beginning, CT_EXECONLY);
+		}
 	}
 	add_completion(beginning);
 }
@@ -650,14 +663,14 @@ filename_completion(const char *str, CompletionType type)
 
 	dir = opendir(dirname);
 
-	if(dir == NULL || my_chdir(dirname) != 0)
+	if(dir == NULL || vifm_chdir(dirname) != 0)
 	{
 		add_completion(filename);
 	}
 	else
 	{
 		filename_completion_internal(dir, dirname, filename, type);
-		(void)my_chdir(curr_view->curr_dir);
+		(void)vifm_chdir(curr_view->curr_dir);
 	}
 
 	free(filename);
@@ -716,16 +729,36 @@ filename_completion_internal(DIR * dir, const char * dirname,
 	}
 }
 
+/* Adds completion of a filename to the list of matches taking care of escaping
+ * (depends on the type parameter). */
 static void
-add_filename_completion(const char * filename, CompletionType type)
+add_filename_completion(const char filename[], CompletionType type)
 {
 #ifndef _WIN32
-	int woe = (type == CT_ALL_WOE || type == CT_FILE_WOE);
-	char * temp = woe ? strdup(filename) : escape_filename(filename, 1);
-	add_completion(temp);
-	free(temp);
+	char *escaped = NULL;
+#endif
+
+	const int woe = (type == CT_ALL_WOE || type == CT_FILE_WOE);
+	const char *completion;
+
+	if(woe)
+	{
+		completion = filename;
+	}
+	else
+	{
+#ifndef _WIN32
+		escaped = escape_filename(filename, 1);
+		completion = escaped;
 #else
-	add_completion(escape_for_cd(filename));
+		completion = escape_for_cd(filename);
+#endif
+	}
+
+	add_completion(completion);
+
+#ifndef _WIN32
+	free(escaped);
 #endif
 }
 
