@@ -39,7 +39,8 @@
 #endif
 
 #include <assert.h> /* assert() */
-#include <errno.h>
+#include <errno.h> /* errno */
+#include <math.h> /* abs() */
 #include <stddef.h> /* NULL size_t */
 #include <stdint.h> /* uint64_t */
 #include <stdio.h> /* snprintf() */
@@ -56,6 +57,7 @@
 #include "utils/fs.h"
 #include "utils/fs_limits.h"
 #include "utils/log.h"
+#include "utils/macros.h"
 #include "utils/path.h"
 #include "utils/str.h"
 #include "utils/string_array.h"
@@ -76,7 +78,6 @@
 #include "running.h"
 #include "sort.h"
 #include "status.h"
-#include "term_title.h"
 #include "types.h"
 #include "ui.h"
 
@@ -86,19 +87,27 @@
 #define CASE_SENSATIVE_FILTER 1
 #endif
 
+/* Mark for a cursor position of inactive pane. */
+#define INACTIVE_CURSOR_MARK "*"
+
 /* Packet set of parameters to pass as user data for processing columns. */
 typedef struct
 {
-	FileView *view;
-	size_t line;
-	int current;
-	size_t current_line;
-	size_t column_offset;
+	FileView *view;  /* View on which cell is being drawn. */
+	size_t line_pos; /* File position in the file list (the view). */
+	int is_current;  /* Whether this file is selected with the cursor. */
+
+	size_t current_line;  /* Line of the cell. */
+	size_t column_offset; /* Offset in characters of the column. */
 }
 column_data_t;
 
 static void column_line_print(const void *data, int column_id, const char *buf,
 		size_t offset);
+static int prepare_primary_col_color(FileView *view, int line_color,
+		int selected, int current);
+static int prepare_secondary_col_color(FileView *view, int selected,
+		int current);
 static void format_name(int id, const void *data, size_t buf_len, char *buf);
 static void format_size(int id, const void *data, size_t buf_len, char *buf);
 static void format_ext(int id, const void *data, size_t buf_len, char *buf);
@@ -120,10 +129,14 @@ static char * get_viewer_command(const char *viewer);
 static void consider_scroll_bind(FileView *view);
 static void correct_list_pos_down(FileView *view, size_t pos_delta);
 static void correct_list_pos_up(FileView *view, size_t pos_delta);
+static void draw_cell(const FileView *view, const column_data_t *cdt,
+		size_t col_width);
 static int calculate_top_position(FileView *view, int top);
 static size_t calculate_print_width(const FileView *view, int i,
 		size_t max_width);
 static void calculate_table_conf(FileView *view, size_t *count, size_t *width);
+static void calculate_number_width(FileView *view);
+static int count_digits(int num);
 size_t calculate_columns_count(FileView *view);
 static size_t calculate_column_width(FileView *view);
 static size_t get_effective_scroll_offset(const FileView *view);
@@ -186,65 +199,119 @@ static void
 column_line_print(const void *data, int column_id, const char *buf,
 		size_t offset)
 {
-	int line_color;
-	col_attr_t col;
+	int line_attrs;
 	char print_buf[strlen(buf) + 1];
 	size_t width_left;
 	size_t trim_pos;
 
-	const column_data_t *cdt = data;
-	size_t i = cdt->line;
+	const column_data_t *const cdt = data;
+	const size_t i = cdt->line_pos;
 	FileView *view = cdt->view;
-	dir_entry_t *entry = &view->dir_entry[cdt->line];
-	checked_wmove(view->win, cdt->current_line, 1 + cdt->column_offset + offset);
+	dir_entry_t *entry = &view->dir_entry[i];
+
+	const size_t prefix_len = view->real_num_width + 1;
+	const size_t final_offset = prefix_len + cdt->column_offset + offset;
 
-	col = view->cs.color[WIN_COLOR];
 	if(column_id == SORT_BY_NAME || column_id == SORT_BY_INAME)
 	{
-		line_color = get_line_color(view, i);
-		mix_colors(&col, &view->cs.color[line_color]);
-		if(entry->selected)
-			mix_colors(&col, &view->cs.color[SELECTED_COLOR]);
-		if(cdt->current)
-		{
-			mix_colors(&col, &view->cs.color[CURR_LINE_COLOR]);
-			line_color = CURRENT_COLOR;
-		}
-		else if(entry->selected)
-		{
-			line_color = SELECTED_COLOR;
-		}
-		init_pair(view->color_scheme + line_color, col.fg, col.bg);
+		line_attrs = prepare_primary_col_color(view, get_line_color(view, i),
+				entry->selected, cdt->is_current);
 	}
 	else
 	{
-		line_color = WIN_COLOR;
-
-		if(entry->selected)
-		{
-			mix_colors(&col, &view->cs.color[SELECTED_COLOR]);
-			line_color = SELECTED_COLOR;
-		}
-
-		if(cdt->current)
-		{
-			mix_colors(&col, &view->cs.color[CURR_LINE_COLOR]);
-			line_color = CURRENT_COLOR;
-		}
-		else
-		{
-			init_pair(view->color_scheme + line_color, col.fg, col.bg);
-		}
+		line_attrs = prepare_secondary_col_color(view, entry->selected,
+				cdt->is_current);
 	}
-	wattron(view->win, COLOR_PAIR(view->color_scheme + line_color) | col.attr);
+
+	if(offset == 0 && ui_view_displays_numbers(view))
+	{
+		char number[view->real_num_width + 1];
+		int mixed;
+		const char *format;
+		int line_number;
+		int is_current_line;
+
+		const int line_attrs = prepare_secondary_col_color(view, entry->selected,
+				cdt->is_current);
+
+		is_current_line = (i == view->list_pos);
+		mixed = is_current_line && view->num_type == NT_MIX;
+		format = mixed ? "%-*d" : "%*d";
+		line_number = ((view->num_type & NT_REL) && !mixed)
+		            ? abs(i - view->list_pos)
+		            : (i + 1);
+
+		snprintf(number, sizeof(number), format, view->real_num_width, line_number);
+
+		checked_wmove(view->win, cdt->current_line,
+				final_offset - 1 - view->real_num_width);
+		wprinta(view->win, number, line_attrs);
+	}
+
+	checked_wmove(view->win, cdt->current_line, final_offset);
 
 	strcpy(print_buf, buf);
 	width_left = view->window_width - (column_id != FILL_COLUMN_ID) - offset;
 	trim_pos = get_normal_utf8_string_widthn(buf, width_left);
 	print_buf[trim_pos] = '\0';
-	wprint(view->win, print_buf);
+	wprinta(view->win, print_buf, line_attrs);
+}
 
-	wattroff(view->win, COLOR_PAIR(view->color_scheme + line_color) | col.attr);
+/* Calculate color attributes for primary view column.  Returns attributes that
+ * can be used for drawing on a window. */
+static int
+prepare_primary_col_color(FileView *view, int line_color, int selected,
+		int current)
+{
+	col_attr_t col = view->cs.color[WIN_COLOR];
+
+	mix_colors(&col, &view->cs.color[line_color]);
+
+	if(selected)
+	{
+		mix_colors(&col, &view->cs.color[SELECTED_COLOR]);
+	}
+
+	if(current)
+	{
+		mix_colors(&col, &view->cs.color[CURR_LINE_COLOR]);
+		line_color = CURRENT_COLOR;
+	}
+	else if(selected)
+	{
+		line_color = SELECTED_COLOR;
+	}
+
+	init_pair(view->color_scheme + line_color, col.fg, col.bg);
+
+	return COLOR_PAIR(view->color_scheme + line_color) | col.attr;
+}
+
+/* Calculate color attributes for secondary view column.  Returns attributes
+ * that can be used for drawing on a window. */
+static int
+prepare_secondary_col_color(FileView *view, int selected, int current)
+{
+	col_attr_t col = view->cs.color[WIN_COLOR];
+	int line_color = WIN_COLOR;
+
+	if(selected)
+	{
+		mix_colors(&col, &view->cs.color[SELECTED_COLOR]);
+		line_color = SELECTED_COLOR;
+	}
+
+	if(current)
+	{
+		mix_colors(&col, &view->cs.color[CURR_LINE_COLOR]);
+		line_color = CURRENT_COLOR;
+	}
+	else
+	{
+		init_pair(view->color_scheme + line_color, col.fg, col.bg);
+	}
+
+	return COLOR_PAIR(view->color_scheme + line_color) | col.attr;
 }
 
 /* File name format callback for column_view unit. */
@@ -252,7 +319,7 @@ static void
 format_name(int id, const void *data, size_t buf_len, char *buf)
 {
 	const column_data_t *cdt = data;
-	format_entry_name(cdt->view, cdt->line, buf_len + 1, buf);
+	format_entry_name(cdt->view, cdt->line_pos, buf_len + 1, buf);
 }
 
 /* File size format callback for column_view unit. */
@@ -261,7 +328,7 @@ format_size(int id, const void *data, size_t buf_len, char *buf)
 {
 	char str[24] = "";
 	const column_data_t *cdt = data;
-	uint64_t size = get_file_size_by_entry(cdt->view, cdt->line);
+	uint64_t size = get_file_size_by_entry(cdt->view, cdt->line_pos);
 	friendly_size_notation(size, sizeof(str), str);
 	snprintf(buf, buf_len + 1, " %s", str);
 }
@@ -271,7 +338,7 @@ static void
 format_ext(int id, const void *data, size_t buf_len, char *buf)
 {
 	const column_data_t *cdt = data;
-	dir_entry_t *entry = &cdt->view->dir_entry[cdt->line];
+	dir_entry_t *entry = &cdt->view->dir_entry[cdt->line_pos];
 	const char *dot = strrchr(entry->name,  '.');
 	snprintf(buf, buf_len + 1, "%s", (dot == NULL) ? "" : (dot + 1));
 	chosp(buf);
@@ -284,7 +351,7 @@ static void
 format_group(int id, const void *data, size_t buf_len, char *buf)
 {
 	const column_data_t *cdt = data;
-	dir_entry_t *entry = &cdt->view->dir_entry[cdt->line];
+	dir_entry_t *entry = &cdt->view->dir_entry[cdt->line_pos];
 	if(id == SORT_BY_GROUP_NAME)
 	{
 		struct group *grp_buf;
@@ -303,7 +370,7 @@ static void
 format_owner(int id, const void *data, size_t buf_len, char *buf)
 {
 	const column_data_t *cdt = data;
-	dir_entry_t *entry = &cdt->view->dir_entry[cdt->line];
+	dir_entry_t *entry = &cdt->view->dir_entry[cdt->line_pos];
 	if(id == SORT_BY_OWNER_NAME)
 	{
 		struct passwd *pwd_buf;
@@ -322,7 +389,7 @@ static void
 format_mode(int id, const void *data, size_t buf_len, char *buf)
 {
 	const column_data_t *cdt = data;
-	dir_entry_t *entry = &cdt->view->dir_entry[cdt->line];
+	dir_entry_t *entry = &cdt->view->dir_entry[cdt->line_pos];
 	snprintf(buf, buf_len, " %s", get_mode_str(entry->mode));
 }
 
@@ -335,7 +402,7 @@ format_time(int id, const void *data, size_t buf_len, char *buf)
 	struct tm *tm_ptr;
 	const column_data_t *cdt = data;
 	FileView *view = cdt->view;
-	dir_entry_t *entry = &view->dir_entry[cdt->line];
+	dir_entry_t *entry = &view->dir_entry[cdt->line_pos];
 
 	switch(id)
 	{
@@ -372,7 +439,7 @@ format_perms(int id, const void *data, size_t buf_len, char buf[])
 {
 	const column_data_t *cdt = data;
 	FileView *view = cdt->view;
-	dir_entry_t *entry = &view->dir_entry[cdt->line];
+	dir_entry_t *entry = &view->dir_entry[cdt->line_pos];
 	get_perm_string(buf, buf_len, entry->mode);
 }
 #endif
@@ -420,8 +487,12 @@ reset_view(FileView *view)
 	view->max_filename_len = 0;
 	view->column_count = 1;
 
-	(void)replace_string(&view->prev_name_filter, "");
-	reset_filter(&view->name_filter);
+	view->num_type = NT_NONE;
+	view->num_width = 4;
+	view->real_num_width = 0;
+
+	(void)replace_string(&view->prev_manual_filter, "");
+	reset_filter(&view->manual_filter);
 	(void)replace_string(&view->prev_auto_filter, "");
 	reset_filter(&view->auto_filter);
 
@@ -432,8 +503,8 @@ reset_view(FileView *view)
 	view->local_filter.poshist = NULL;
 	view->local_filter.poshist_len = 0U;
 
-	view->sort[0] = DEFAULT_SORT_KEY;
-	memset(&view->sort[1], NO_SORT_OPTION, sizeof(view->sort) - 1);
+	memset(&view->sort[0], NO_SORT_OPTION, sizeof(view->sort));
+	ui_view_sort_list_ensure_well_formed(view->sort);
 }
 
 /* Resets filter to empty state (either initializes or clears it). */
@@ -770,90 +841,6 @@ find_file_pos_in_list(const FileView *const view, const char file[])
 	return -1;
 }
 
-void
-update_view_title(FileView *view)
-{
-	char *buf;
-	size_t len;
-	int gen_view = get_mode() == VIEW_MODE && !curr_view->explore_mode;
-	FileView *selected = gen_view ? other_view : curr_view;
-
-	if(gen_view && view == other_view)
-		return;
-
-	if(curr_stats.load_stage < 2)
-		return;
-
-	if(view == selected)
-	{
-		col_attr_t col;
-
-		col = cfg.cs.color[TOP_LINE_COLOR];
-		mix_colors(&col, &cfg.cs.color[TOP_LINE_SEL_COLOR]);
-		init_pair(DCOLOR_BASE + TOP_LINE_SEL_COLOR, col.fg, col.bg);
-
-		wbkgdset(view->title, COLOR_PAIR(DCOLOR_BASE + TOP_LINE_SEL_COLOR) |
-				(col.attr & A_REVERSE));
-		wattrset(view->title, col.attr & ~A_REVERSE);
-	}
-	else
-	{
-		wbkgdset(view->title, COLOR_PAIR(DCOLOR_BASE + TOP_LINE_COLOR) |
-				(cfg.cs.color[TOP_LINE_COLOR].attr & A_REVERSE));
-		wattrset(view->title, cfg.cs.color[TOP_LINE_COLOR].attr & ~A_REVERSE);
-		wbkgdset(top_line, COLOR_PAIR(DCOLOR_BASE + TOP_LINE_COLOR) |
-				(cfg.cs.color[TOP_LINE_COLOR].attr & A_REVERSE));
-		wattrset(top_line, cfg.cs.color[TOP_LINE_COLOR].attr & ~A_REVERSE);
-		werase(top_line);
-	}
-	werase(view->title);
-
-	if(curr_stats.load_stage < 2)
-		return;
-
-	buf = replace_home_part(view->curr_dir);
-	if(view == selected)
-	{
-		set_term_title(replace_home_part(view->curr_dir));
-	}
-
-	if(view->explore_mode)
-	{
-		if(!is_root_dir(buf))
-			strcat(buf, "/");
-		strcat(buf, get_current_file_name(view));
-	}
-
-	len = get_screen_string_length(buf);
-	if(len > view->window_width + 1 && view == selected)
-	{ /* Truncate long directory names */
-		const char *ptr;
-
-		ptr = buf;
-		while(len > view->window_width - 2)
-		{
-			len--;
-			ptr += get_char_width(ptr);
-		}
-
-		wprintw(view->title, "...");
-		wprint(view->title, ptr);
-	}
-	else if(len > view->window_width + 1 && view != selected)
-	{
-		size_t len = get_normal_utf8_string_widthn(buf, view->window_width - 3 + 1);
-		buf[len] = '\0';
-		wprint(view->title, buf);
-		wprintw(view->title, "...");
-	}
-	else
-	{
-		wprint(view->title, buf);
-	}
-
-	wnoutrefresh(view->title);
-}
-
 void
 reset_view_sort(FileView *view)
 {
@@ -889,22 +876,28 @@ draw_dir_list(FileView *view)
 {
 	int attr;
 	int x;
-	int y;
+	int cell;
 	size_t col_width;
 	size_t col_count;
 	int top = view->top_line;
 
 	if(curr_stats.load_stage < 2)
+	{
 		return;
+	}
 
 	calculate_table_conf(view, &col_count, &col_width);
 
 	if(top + view->window_rows > view->list_rows)
+	{
 		top = view->list_rows - view->window_rows;
+	}
 	if(top < 0)
+	{
 		top = 0;
+	}
 
-	update_view_title(view);
+	ui_view_title_update(view);
 
 	/* This is needed for reloading a list that has had files deleted */
 	while((view->list_rows - view->list_pos) <= 0)
@@ -924,16 +917,24 @@ draw_dir_list(FileView *view)
 	wbkgdset(view->win, COLOR_PAIR(WIN_COLOR + view->color_scheme) | attr);
 	werase(view->win);
 
-	y = 0;
+	cell = 0;
 	for(x = top; x < view->list_rows; x++)
 	{
-		checked_wmove(view->win, y, 1);
-		wclrtoeol(view->win);
-		column_data_t cdt = {view, x, 0, y/col_count, (y%col_count)*col_width};
-		columns_format_line(view->columns, &cdt, col_width);
-		y++;
-		if(y >= view->window_cells)
+		const column_data_t cdt =
+		{
+			.view = view,
+			.line_pos = x,
+			.is_current = 0,
+			.current_line = cell/col_count,
+			.column_offset = (cell%col_count)*col_width,
+		};
+
+		draw_cell(view, &cdt, col_width);
+
+		if(++cell >= view->window_cells)
+		{
 			break;
+		}
 	}
 
 	view->top_line = top;
@@ -1086,7 +1087,6 @@ erase_current_line_bar(FileView *view)
 	int old_pos = view->top_line + old_cursor;
 	size_t col_width;
 	size_t col_count;
-	size_t print_width;
 	column_data_t cdt = {view, old_pos, 0};
 
 	if(old_cursor < 0)
@@ -1102,14 +1102,11 @@ erase_current_line_bar(FileView *view)
 	}
 
 	calculate_table_conf(view, &col_count, &col_width);
-	print_width = calculate_print_width(view, old_pos, col_width);
 
 	cdt.current_line = old_cursor/col_count;
 	cdt.column_offset = (old_cursor%col_count)*col_width;
 
-	column_line_print(&cdt, FILL_COLUMN_ID, " ", -1);
-	columns_format_line(view->columns, &cdt, col_width);
-	column_line_print(&cdt, FILL_COLUMN_ID, " ", print_width);
+	draw_cell(view, &cdt, col_width);
 
 	if(view == other_view)
 	{
@@ -1117,6 +1114,18 @@ erase_current_line_bar(FileView *view)
 	}
 }
 
+/* Draws a full cell of the file list. */
+static void
+draw_cell(const FileView *view, const column_data_t *cdt, size_t col_width)
+{
+	const size_t print_width = calculate_print_width(view, cdt->line_pos,
+			col_width);
+
+	column_line_print(cdt, FILL_COLUMN_ID, " ", -1);
+	columns_format_line(view->columns, cdt, col_width);
+	column_line_print(cdt, FILL_COLUMN_ID, " ", print_width);
+}
+
 int
 move_curr_line(FileView *view)
 {
@@ -1148,17 +1157,20 @@ move_curr_line(FileView *view)
 	else if(pos > last)
 	{
 		scroll_down(view, pos - last);
-		redraw = 1;
+		redraw++;
 	}
 	else if(pos < view->top_line)
 	{
 		scroll_up(view, view->top_line - pos);
-		redraw = 1;
+		redraw++;
 	}
 
-	redraw = consider_scroll_offset(view) ? 1 : redraw;
+	if(consider_scroll_offset(view))
+	{
+		redraw++;
+	}
 
-	return redraw;
+	return redraw != 0 || (view->num_type & NT_REL);
 }
 
 /* Calculates top position basing on window and list size and trying to show as
@@ -1221,7 +1233,7 @@ move_to_list_pos(FileView *view, int pos)
 	calculate_table_conf(view, &col_count, &col_width);
 	print_width = calculate_print_width(view, view->list_pos, col_width);
 
-	cdt.line = pos;
+	cdt.line_pos = pos;
 	cdt.current_line = view->curr_line/col_count;
 	cdt.column_offset = (view->curr_line%col_count)*col_width;
 
@@ -1263,18 +1275,29 @@ put_inactive_mark(FileView *view)
 {
 	size_t col_width;
 	size_t col_count;
+	int is_selected;
+	int line_attrs;
+	int line, column;
 
 	calculate_table_conf(view, &col_count, &col_width);
 
-	mvwaddstr(view->win, view->curr_line/col_count,
-			(view->curr_line%col_count)*col_width, "*");
+	is_selected = view->dir_entry[view->list_pos].selected;
+	line_attrs = prepare_secondary_col_color(view, is_selected, 0);
+
+	line = view->curr_line/col_count;
+	column = view->real_num_width + (view->curr_line%col_count)*col_width;
+	checked_wmove(view->win, line, column);
+
+	wprinta(view->win, INACTIVE_CURSOR_MARK, line_attrs);
 }
 
 /* Calculates number of columns and maximum width of column in a view. */
 static void
 calculate_table_conf(FileView *view, size_t *count, size_t *width)
 {
-	*width = view->window_width - 1;
+	calculate_number_width(view);
+
+	*width = MAX(0, (int)view->window_width - 1 - view->real_num_width);
 	*count = 1;
 
 	if(view->ls_view)
@@ -1286,6 +1309,39 @@ calculate_table_conf(FileView *view, size_t *count, size_t *width)
 	view->window_cells = *count*(view->window_rows + 1);
 }
 
+/* Calculates real number of characters that should be allocated in view for
+ * numbers column. */
+static void
+calculate_number_width(FileView *view)
+{
+	if(ui_view_displays_numbers(view))
+	{
+		const int digit_count = count_digits(view->list_rows);
+		const int min = view->num_width;
+		const int max = view->window_width - 1;
+		view->real_num_width = MIN(MAX(1 + digit_count, min), max);
+	}
+	else
+	{
+		view->real_num_width = 0;
+	}
+}
+
+/* Counts number of digits in a number assuming that zero takes on digit.
+ * Returns the count. */
+static int
+count_digits(int num)
+{
+	int count = 0;
+	do
+	{
+		count++;
+		num /= 10;
+	}
+	while(num != 0);
+	return count;
+}
+
 size_t
 calculate_columns_count(FileView *view)
 {
@@ -1708,6 +1764,7 @@ invert_selection(FileView *view)
 			e->selected = !e->selected;
 		}
 	}
+	view->selected_files = view->list_rows - view->selected_files;
 }
 
 void
@@ -1904,6 +1961,7 @@ change_directory(FileView *view, const char *directory)
 {
 	char newdir[PATH_MAX];
 	char dir_dup[PATH_MAX];
+	int location_changed;
 
 	if(is_dir_list_loaded(view))
 	{
@@ -1939,12 +1997,17 @@ change_directory(FileView *view, const char *directory)
 	if(!is_valid_dir(dir_dup))
 	{
 		show_error_msgf("Directory Access Error", "Cannot open %s", dir_dup);
-		snprintf(view->curr_dir, sizeof(view->curr_dir), "%s", dir_dup);
+		copy_str(view->curr_dir, sizeof(view->curr_dir), dir_dup);
 		leave_invalid_dir(view);
-		snprintf(dir_dup, sizeof(dir_dup), "%s", view->curr_dir);
+		copy_str(dir_dup, sizeof(dir_dup), view->curr_dir);
 	}
 
-	snprintf(view->last_dir, sizeof(view->last_dir), "%s", view->curr_dir);
+	location_changed = stroscmp(dir_dup, view->curr_dir) != 0;
+
+	if(location_changed)
+	{
+		copy_str(view->last_dir, sizeof(view->last_dir), view->curr_dir);
+	}
 
 	/* Check if we're exiting from a FUSE mounted top level directory and the
 	 * other pane isn't in it or any of it subdirectories.
@@ -2002,14 +2065,14 @@ change_directory(FileView *view, const char *directory)
 		LOG_SERROR_MSG(errno, "Can't access(, R_OK) \"%s\"", dir_dup);
 		log_cwd();
 
-		if(stroscmp(view->curr_dir, dir_dup) != 0)
+		if(location_changed)
 		{
 			show_error_msgf("Directory Access Error",
 					"You do not have read access on %s", dir_dup);
 		}
 	}
 
-	if(my_chdir(dir_dup) == -1 && !is_unc_root(dir_dup))
+	if(vifm_chdir(dir_dup) != 0 && !is_unc_root(dir_dup))
 	{
 		LOG_SERROR_MSG(errno, "Can't chdir() \"%s\"", dir_dup);
 		log_cwd();
@@ -2021,7 +2084,7 @@ change_directory(FileView *view, const char *directory)
 	if(!is_root_dir(dir_dup))
 		chosp(dir_dup);
 
-	if(stroscmp(dir_dup, view->curr_dir) != 0)
+	if(location_changed)
 	{
 		filter_clear(&view->local_filter.filter);
 		free_saved_selection(view);
@@ -2035,7 +2098,7 @@ change_directory(FileView *view, const char *directory)
 	/* Need to use setenv instead of getcwd for a symlink directory */
 	env_set("PWD", dir_dup);
 
-	snprintf(view->curr_dir, sizeof(view->curr_dir), "%s", dir_dup);
+	copy_str(view->curr_dir, sizeof(view->curr_dir), dir_dup);
 
 	if(is_dir_list_loaded(view))
 	{
@@ -2271,12 +2334,7 @@ fill_dir_list(FileView *view)
 					dir_entry->name);
 			log_cwd();
 
-			dir_entry->type = type_from_dir_entry(d);
-			if(dir_entry->type == DIRECTORY)
-			{
-				strcat(dir_entry->name, "/");
-				name_len++;
-			}
+			dir_entry->type = UNKNOWN;
 			dir_entry->size = 0;
 			dir_entry->mode = 0;
 			dir_entry->uid = -1;
@@ -2284,39 +2342,46 @@ fill_dir_list(FileView *view)
 			dir_entry->mtime = 0;
 			dir_entry->atime = 0;
 			dir_entry->ctime = 0;
-
-			name_len += get_filetype_decoration_width(dir_entry->type);
-			view->max_filename_len = MAX(view->max_filename_len, name_len);
-			continue;
 		}
-
-		dir_entry->size = (uintmax_t)s.st_size;
-		dir_entry->mode = s.st_mode;
-		dir_entry->uid = s.st_uid;
-		dir_entry->gid = s.st_gid;
-		dir_entry->mtime = s.st_mtime;
-		dir_entry->atime = s.st_atime;
-		dir_entry->ctime = s.st_ctime;
-
-		if(s.st_ino)
+		else
 		{
 			dir_entry->type = get_type_from_mode(s.st_mode);
-			if(dir_entry->type == LINK)
-			{
-				struct stat st;
-				if(check_link_is_dir(dir_entry->name))
-					strcat(dir_entry->name, "/");
-				if(stat(dir_entry->name, &st) == 0)
-					dir_entry->mode = st.st_mode;
-			}
-			else if(dir_entry->type == DIRECTORY)
-			{
-					strcat(dir_entry->name, "/");
-					name_len++;
-			}
-			name_len += get_filetype_decoration_width(dir_entry->type);
-			view->max_filename_len = MAX(view->max_filename_len, name_len);
+			dir_entry->size = (uintmax_t)s.st_size;
+			dir_entry->mode = s.st_mode;
+			dir_entry->uid = s.st_uid;
+			dir_entry->gid = s.st_gid;
+			dir_entry->mtime = s.st_mtime;
+			dir_entry->atime = s.st_atime;
+			dir_entry->ctime = s.st_ctime;
 		}
+
+		if(dir_entry->type == UNKNOWN)
+		{
+			dir_entry->type = type_from_dir_entry(d);
+		}
+
+		if(dir_entry->type == LINK)
+		{
+			struct stat st;
+
+			if(check_link_is_dir(dir_entry->name))
+			{
+				strcat(dir_entry->name, "/");
+			}
+
+			if(stat(dir_entry->name, &st) == 0)
+			{
+				dir_entry->mode = st.st_mode;
+			}
+		}
+		else if(dir_entry->type == DIRECTORY)
+		{
+			strcat(dir_entry->name, "/");
+			name_len++;
+		}
+
+		name_len += get_filetype_decoration_width(dir_entry->type);
+		view->max_filename_len = MAX(view->max_filename_len, name_len);
 	}
 	closedir(dir);
 #else
@@ -2453,25 +2518,29 @@ fill_dir_list(FileView *view)
 TSTATIC int
 file_is_visible(FileView *view, const char filename[], int is_dir)
 {
-	char name_with_slash[strlen(filename) + 1 + 1];
-	sprintf(name_with_slash, "%s%c", filename, is_dir ? '/' : '\0');
+	char name_with_slash[NAME_MAX + 1 + 1];
+	if(is_dir)
+	{
+		sprintf(name_with_slash, "%s/", filename);
+		filename = name_with_slash;
+	}
 
-	if(filter_matches(&view->auto_filter, name_with_slash) > 0)
+	if(filter_matches(&view->auto_filter, filename) > 0)
 	{
 		return 0;
 	}
 
-	if(filter_matches(&view->local_filter.filter, name_with_slash) == 0)
+	if(filter_matches(&view->local_filter.filter, filename) == 0)
 	{
 		return 0;
 	}
 
-	if(filter_is_empty(&view->name_filter))
+	if(filter_is_empty(&view->manual_filter))
 	{
 		return 1;
 	}
 
-	if(filter_matches(&view->name_filter, name_with_slash) > 0)
+	if(filter_matches(&view->manual_filter, filename) > 0)
 	{
 		return !view->invert;
 	}
@@ -2540,17 +2609,21 @@ populate_dir_list_internal(FileView *view, int reload)
 	}
 
 	if(curr_stats.load_stage < 2)
+	{
 		update_all_windows();
+	}
 
 	/* this is needed for lstat() below */
-	if(my_chdir(view->curr_dir) != 0 && !is_unc_root(view->curr_dir))
+	if(vifm_chdir(view->curr_dir) != 0 && !is_unc_root(view->curr_dir))
 	{
 		LOG_SERROR_MSG(errno, "Can't chdir() into \"%s\"", view->curr_dir);
 		return 1;
 	}
 
 	if(reload && view->selected_files > 0 && view->selected_filelist == NULL)
+	{
 		get_all_selected_files(view);
+	}
 
 	if(view->dir_entry != NULL)
 	{
@@ -2677,13 +2750,13 @@ rescue_from_empty_filelist(FileView * view)
 	 * in the / directory.  All other directories will always show at least the
 	 * ../ file.  This resets the filter and reloads the directory.
 	 */
-	if(is_path_absolute(view->curr_dir) && !filter_is_empty(&view->name_filter))
+	if(is_path_absolute(view->curr_dir) && !filter_is_empty(&view->manual_filter))
 	{
 		show_error_msgf("Filter error",
 				"The %s\"%s\" pattern did not match any files. It was reset.",
-				view->invert ? "" : "inverted ", view->name_filter.raw);
+				view->invert ? "" : "inverted ", view->manual_filter.raw);
 		filter_clear(&view->auto_filter);
-		filter_clear(&view->name_filter);
+		filter_clear(&view->manual_filter);
 		view->invert = 1;
 
 		load_dir_list(view, 1);
@@ -2802,8 +2875,8 @@ toggle_dot_files(FileView *view)
 void
 remove_filename_filter(FileView *view)
 {
-	(void)replace_string(&view->prev_name_filter, view->name_filter.raw);
-	filter_clear(&view->name_filter);
+	(void)replace_string(&view->prev_manual_filter, view->manual_filter.raw);
+	filter_clear(&view->manual_filter);
 	(void)replace_string(&view->prev_auto_filter, view->auto_filter.raw);
 	filter_clear(&view->auto_filter);
 
@@ -2815,7 +2888,7 @@ remove_filename_filter(FileView *view)
 void
 restore_filename_filter(FileView *view)
 {
-	(void)filter_set(&view->name_filter, view->prev_name_filter);
+	(void)filter_set(&view->manual_filter, view->prev_manual_filter);
 	(void)filter_set(&view->auto_filter, view->prev_auto_filter);
 	view->invert = view->prev_invert;
 	load_saving_pos(view, 0);
@@ -2922,6 +2995,12 @@ local_filter_update_view(FileView *view, int rel_pos)
 
 	if(pos >= 0)
 	{
+		if(pos == 0 && is_parent_dir(view->dir_entry[0].name) &&
+				view->list_rows > 0)
+		{
+			pos++;
+		}
+
 		view->list_pos = pos;
 		view->top_line = pos - rel_pos;
 	}
