@@ -25,7 +25,7 @@
 
 #include <pthread.h>
 
-#include <dirent.h> /* DIR */
+#include <dirent.h> /* DIR dirent opendir() readdir() closedir() */
 #include <fcntl.h>
 #include <sys/stat.h> /* stat */
 #include <sys/types.h> /* waitpid() */
@@ -67,6 +67,7 @@
 #include "registers.h"
 #include "running.h"
 #include "status.h"
+#include "trash.h"
 #include "ui.h"
 #include "undo.h"
 
@@ -111,16 +112,22 @@ TSTATIC const char * add_to_name(const char filename[], int k);
 TSTATIC int check_file_rename(const char dir[], const char old[],
 		const char new[], SignalType signal_type);
 static int is_file_name_changed(const char old[], const char new[]);
-static void put_confirm_cb(const char *dest_name);
+static void put_confirm_cb(const char dest_name[]);
 static void prompt_what_to_do(const char src_name[]);
 TSTATIC const char * gen_clone_name(const char normal_name[]);
-static void put_decide_cb(const char *dest_name);
+static void put_decide_cb(const char dest_name[]);
 static int entry_is_dir(const char full_path[], const struct dirent* dentry);
 static int put_files_from_register_i(FileView *view, int start);
+static int mv_file(const char src[], const char src_path[], const char dst[],
+		const char path[], int tmpfile_num, int cancellable);
+static int cp_file(const char src_dir[], const char dst_dir[], const char src[],
+		const char dst[], int type, int cancellable);
 static int have_read_access(FileView *view);
 static char ** edit_list(size_t count, char **orig, int *nlines,
 		int ignore_change);
 static int edit_file(const char filepath[], int force_changed);
+static void general_prepare_for_bg_task(FileView *view, bg_args_t *args);
+static const char * get_cancellation_suffix(void);
 
 /* returns new value for save_msg */
 int
@@ -179,23 +186,6 @@ progress_msg(const char *text, int ready, int total)
 	curr_stats.save_msg = 2;
 }
 
-/* returns string that needs to be released by caller */
-static char *
-gen_trash_name(const char *name)
-{
-	struct stat st;
-	char buf[PATH_MAX];
-	int i = 0;
-
-	do
-	{
-		snprintf(buf, sizeof(buf), "%s/%03d_%s", cfg.trash_dir, i++, name);
-		chosp(buf);
-	}
-	while(lstat(buf, &st) == 0);
-	return strdup(buf);
-}
-
 /* buf should be at least COMMAND_GROUP_INFO_LEN characters length */
 static void
 get_group_file_list(char **list, int count, char *buf)
@@ -222,13 +212,12 @@ delete_file(FileView *view, int reg, int count, int *indexes, int use_trash)
 {
 	char buf[MAX(COMMAND_GROUP_INFO_LEN, 8 + PATH_MAX*2)];
 	int x, y;
-	int i;
+	int sel_len;
 
 	if(!check_if_dir_writable(DR_CURRENT, view->curr_dir))
 		return 0;
 
-	if(cfg.use_trash && use_trash &&
-			path_starts_with(view->curr_dir, cfg.trash_dir))
+	if(cfg.use_trash && use_trash && is_under_trash(view->curr_dir))
 	{
 		show_error_msg("Can't perform deletion",
 				"Current directory is under Trash directory");
@@ -237,24 +226,12 @@ delete_file(FileView *view, int reg, int count, int *indexes, int use_trash)
 
 	if(count > 0)
 	{
-		int j;
 		get_selected_files(view, count, indexes);
-		i = view->list_pos;
-		for(j = 0; j < count; j++)
-			if(indexes[j] == i && i < view->list_rows - 1)
-			{
-				i++;
-				j = -1;
-			}
 	}
 	else
 	{
 		get_all_selected_files(view);
-		i = view->list_pos;
-		while(i < view->list_rows - 1 && view->dir_entry[i].selected)
-			i++;
 	}
-	view->list_pos = i;
 
 	if(cfg.use_trash && use_trash)
 	{
@@ -272,78 +249,103 @@ delete_file(FileView *view, int reg, int count, int *indexes, int use_trash)
 	cmd_group_begin(buf);
 
 	y = 0;
-	if(my_chdir(curr_view->curr_dir) != 0)
+	if(vifm_chdir(curr_view->curr_dir) != 0)
 	{
 		show_error_msg("Directory return", "Can't chdir() to current directory");
 		return 1;
 	}
-	for(x = 0; x < view->selected_files; x++)
+
+	ui_cancellation_reset();
+
+	sel_len = view->selected_files;
+	for(x = 0; x < sel_len && !ui_cancellation_requested(); x++)
 	{
+		const char *const fname = view->selected_filelist[x];
 		char full_buf[PATH_MAX];
 		int result;
 
-		if(is_parent_dir(view->selected_filelist[x]))
+		if(is_parent_dir(fname))
 		{
-			show_error_msg("Background Process Error",
+			show_error_msg("Can't perform deletion",
 					"You cannot delete the ../ directory");
 			continue;
 		}
 
-		snprintf(full_buf, sizeof(full_buf), "%s/%s", view->curr_dir,
-				view->selected_filelist[x]);
+		snprintf(full_buf, sizeof(full_buf), "%s/%s", view->curr_dir, fname);
 		chosp(full_buf);
 
-		progress_msg("Deleting files", x + 1, view->selected_files);
+		progress_msg("Deleting files", x + 1, sel_len);
 		if(cfg.use_trash && use_trash)
 		{
-			if(stroscmp(full_buf, cfg.trash_dir) == 0)
+			if(!is_trash_directory(full_buf))
 			{
-				show_error_msg("Background Process Error",
-						"You cannot delete trash directory to trash");
-				result = -1;
+				char *const dest = gen_trash_name(view->curr_dir, fname);
+				if(dest != NULL)
+				{
+					result = perform_operation(OP_MOVE, NULL, full_buf, dest);
+					/* For some reason "rm" sometimes returns 0 on cancellation. */
+					if(path_exists(full_buf))
+					{
+						result = -1;
+					}
+					if(result == 0)
+					{
+						add_operation(OP_MOVE, NULL, NULL, full_buf, dest);
+						append_to_register(reg, dest);
+					}
+					free(dest);
+				}
+				else
+				{
+					show_error_msgf("No trash directory is available",
+							"Either correct trash directory paths or prune files.  "
+							"Deletion failed on: %s", fname);
+					result = -1;
+				}
 			}
 			else
 			{
-				char *dest;
-
-				dest = gen_trash_name(view->selected_filelist[x]);
-				result = perform_operation(OP_MOVE, NULL, full_buf, dest);
-				if(result == 0)
-				{
-					add_operation(OP_MOVE, NULL, NULL, full_buf, dest);
-					append_to_register(reg, dest);
-				}
-				free(dest);
+				show_error_msg("Can't perform deletion",
+						"You cannot delete trash directory to trash");
+				result = -1;
 			}
 		}
 		else
 		{
 			result = perform_operation(OP_REMOVE, NULL, full_buf, NULL);
+			/* For some reason "rm" sometimes returns 0 on cancellation. */
+			if(path_exists(full_buf))
+			{
+				result = -1;
+			}
 			if(result == 0)
+			{
 				add_operation(OP_REMOVE, NULL, NULL, full_buf, "");
+			}
 		}
 
 		if(result == 0)
 		{
+			if(strcmp(view->dir_entry[view->list_pos].name, fname) == 0)
+			{
+				if(view->list_pos + 1 < view->list_rows)
+				{
+					view->list_pos++;
+				}
+			}
 			y++;
 		}
-		else if(!view->dir_entry[view->list_pos].selected)
-		{
-			int pos = find_file_pos_in_list(view, view->selected_filelist[x]);
-			if(pos >= 0)
-				view->list_pos = pos;
-		}
 	}
 	free_selected_file_array(view);
-	clean_selected_files(view);
 
 	update_unnamed_reg(reg);
 
 	cmd_group_end();
 
-	load_saving_pos(view, 1);
+	ui_view_reset_selection_and_reload(view);
 
-	status_bar_messagef("%d %s deleted", y, y == 1 ? "file" : "files");
+	status_bar_messagef("%d %s deleted%s", y, y == 1 ? "file" : "files",
+			get_cancellation_suffix());
 	return 1;
 }
 
@@ -387,28 +389,30 @@ delete_file_bg_i(const char curr_dir[], char *list[], int count, int use_trash)
 	int i;
 	for(i = 0; i < count; i++)
 	{
+		const char *const fname = list[i];
 		char full_buf[PATH_MAX];
 
-		if(is_parent_dir(list[i]))
+		if(is_parent_dir(fname))
+		{
 			continue;
+		}
 
-		snprintf(full_buf, sizeof(full_buf), "%s/%s", curr_dir, list[i]);
+		snprintf(full_buf, sizeof(full_buf), "%s/%s", curr_dir, fname);
 		chosp(full_buf);
 
 		if(use_trash)
 		{
-			if(strcmp(full_buf, cfg.trash_dir) != 0)
+			if(!is_trash_directory(full_buf))
 			{
-				char *dest;
-
-				dest = gen_trash_name(list[i]);
-				(void)perform_operation(OP_MOVE, NULL, full_buf, dest);
-				free(dest);
+				char *const trash_name = gen_trash_name(curr_dir, fname);
+				const char *const dest = (trash_name != NULL) ? trash_name : fname;
+				(void)perform_operation(OP_MOVE, (void *)1, full_buf, dest);
+				free(trash_name);
 			}
 		}
 		else
 		{
-			(void)perform_operation(OP_REMOVE, NULL, full_buf, NULL);
+			(void)perform_operation(OP_REMOVE, (void *)1, full_buf, NULL);
 		}
 		inner_bg_next();
 	}
@@ -429,7 +433,7 @@ delete_file_bg(FileView *view, int use_trash)
 	args = malloc(sizeof(*args));
 	args->from_trash = cfg.use_trash && use_trash;
 
-	if(args->from_trash && path_starts_with(view->curr_dir, cfg.trash_dir))
+	if(args->from_trash && is_under_trash(view->curr_dir))
 	{
 		show_error_msg("Can't perform deletion",
 				"Current directory is under Trash directory");
@@ -450,15 +454,7 @@ delete_file_bg(FileView *view, int use_trash)
 
 	view->list_pos = i;
 
-	args->sel_list = view->selected_filelist;
-	args->sel_list_len = view->selected_files;
-
-	view->selected_filelist = NULL;
-	free_selected_file_array(view);
-	clean_selected_files(view);
-	load_saving_pos(view, 1);
-
-	strcpy(args->src, view->curr_dir);
+	general_prepare_for_bg_task(view, args);
 
 	if(args->from_trash)
 		snprintf(buf, sizeof(buf), "delete in %s: ",
@@ -469,11 +465,7 @@ delete_file_bg(FileView *view, int use_trash)
 
 	get_group_file_list(view->selected_filelist, view->selected_files, buf);
 
-#ifndef _WIN32
-	args->job = add_background_job(-1, buf, -1);
-#else
-	args->job = add_background_job(-1, buf, (HANDLE)-1);
-#endif
+	args->job = add_background_job(-1, buf, NO_JOB_ID);
 	if(args->job == NULL)
 	{
 		free_string_array(args->sel_list, args->sel_list_len);
@@ -488,45 +480,8 @@ delete_file_bg(FileView *view, int use_trash)
 	return 0;
 }
 
-static int
-mv_file(const char *src, const char *src_path, const char *dst,
-		const char *path, int tmpfile_num)
-{
-	char full_src[PATH_MAX], full_dst[PATH_MAX];
-	int op;
-	int result;
-
-	snprintf(full_src, sizeof(full_src), "%s/%s", src_path, src);
-	chosp(full_src);
-	snprintf(full_dst, sizeof(full_dst), "%s/%s", path, dst);
-	chosp(full_dst);
-
-	/* compare case sensitive strings even on Windows to let user rename file
-	 * changing only case of some characters */
-	if(strcmp(full_src, full_dst) == 0)
-		return 0;
-
-	if(tmpfile_num <= 0)
-		op = OP_MOVE;
-	else if(tmpfile_num == 1)
-		op = OP_MOVETMP1;
-	else if(tmpfile_num == 2)
-		op = OP_MOVETMP2;
-	else if(tmpfile_num == 3)
-		op = OP_MOVETMP3;
-	else if(tmpfile_num == 4)
-		op = OP_MOVETMP4;
-	else
-		op = OP_NONE;
-
-	result = perform_operation(op, NULL, full_src, full_dst);
-	if(result == 0 && tmpfile_num >= 0)
-		add_operation(op, NULL, NULL, full_src, full_dst);
-	return result;
-}
-
 static void
-rename_file_cb(const char *new_name)
+rename_file_cb(const char new_name[])
 {
 	char *filename = get_current_file_name(curr_view);
 	char buf[MAX(COMMAND_GROUP_INFO_LEN, 10 + NAME_MAX + 1)];
@@ -535,8 +490,10 @@ rename_file_cb(const char *new_name)
 	int mv_res;
 	char **filename_ptr;
 
-	if(new_name == NULL || new_name[0] == '\0')
+	if(is_null_or_empty(new_name))
+	{
 		return;
+	}
 
 	if(contains_slash(new_name))
 	{
@@ -558,7 +515,8 @@ rename_file_cb(const char *new_name)
 	snprintf(buf, sizeof(buf), "rename in %s: %s to %s",
 			replace_home_part(curr_view->curr_dir), filename, new);
 	cmd_group_begin(buf);
-	mv_res = mv_file(filename, curr_view->curr_dir, new, curr_view->curr_dir, 0);
+	mv_res = mv_file(filename, curr_view->curr_dir, new, curr_view->curr_dir, 0,
+			1);
 	cmd_group_end();
 	if(mv_res != 0)
 	{
@@ -599,8 +557,7 @@ rename_current_file(FileView *view, int name_only)
 
 	if(name_only)
 	{
-		snprintf(rename_file_ext, sizeof(rename_file_ext), "%s",
-				extract_extension(filename));
+		copy_str(rename_file_ext, sizeof(rename_file_ext), cut_extension(filename));
 	}
 	else
 	{
@@ -706,7 +663,8 @@ perform_renaming(FileView *view, char **files, int *is_dup, int len,
 			continue;
 
 		unique_name = make_name_unique(files[i]);
-		if(mv_file(files[i], view->curr_dir, unique_name, view->curr_dir, 2) != 0)
+		if(mv_file(files[i], view->curr_dir, unique_name, view->curr_dir, 2, 1)
+				!= 0)
 		{
 			cmd_group_end();
 			if(!last_cmd_group_empty())
@@ -726,7 +684,7 @@ perform_renaming(FileView *view, char **files, int *is_dup, int len,
 			continue;
 
 		if(mv_file(files[i], view->curr_dir, list[i], view->curr_dir,
-				is_dup[i] ? 1 : 0) == 0)
+				is_dup[i] ? 1 : 0, 1) == 0)
 		{
 			int pos;
 
@@ -795,15 +753,12 @@ add_files_to_list(const char *path, char **files, int *len)
 
 	while((dentry = readdir(dir)) != NULL)
 	{
-		char buf[PATH_MAX];
-
-		if(stroscmp(dentry->d_name, ".") == 0)
-			continue;
-		else if(stroscmp(dentry->d_name, "..") == 0)
-			continue;
-
-		snprintf(buf, sizeof(buf), "%s%s%s", path, slash, dentry->d_name);
-		files = add_files_to_list(buf, files, len);
+		if(!is_builtin_dir(dentry->d_name))
+		{
+			char buf[PATH_MAX];
+			snprintf(buf, sizeof(buf), "%s%s%s", path, slash, dentry->d_name);
+			files = add_files_to_list(buf, files, len);
+		}
 	}
 
 	closedir(dir);
@@ -1051,7 +1006,8 @@ incdec_names(FileView *view, int k)
 	cmd_group_begin(buf);
 	for(i = 0; i < names_len && !err; i++)
 	{
-		if(mv_file(names[i], view->curr_dir, tmp_names[i], view->curr_dir, 4) != 0)
+		if(mv_file(names[i], view->curr_dir, tmp_names[i], view->curr_dir, 4, 1)
+				!= 0)
 		{
 			err = 1;
 			break;
@@ -1061,7 +1017,7 @@ incdec_names(FileView *view, int k)
 	for(i = 0; i < names_len && !err; i++)
 	{
 		if(mv_file(tmp_names[i], view->curr_dir, add_to_name(names[i], k),
-				view->curr_dir, 3) != 0)
+				view->curr_dir, 3, 1) != 0)
 		{
 			err = 1;
 			break;
@@ -1156,6 +1112,9 @@ chown_files(int u, int g, uid_t uid, gid_t gid)
 {
 	char buf[COMMAND_GROUP_INFO_LEN + 1];
 	int i;
+	int sel_len;
+
+	ui_cancellation_reset();
 
 	snprintf(buf, sizeof(buf), "ch%s in %s: ", ((u && g) || u) ? "own" : "grp",
 			replace_home_part(curr_view->curr_dir));
@@ -1163,7 +1122,9 @@ chown_files(int u, int g, uid_t uid, gid_t gid)
 	get_group_file_list(curr_view->saved_selection, curr_view->nsaved_selection,
 			buf);
 	cmd_group_begin(buf);
-	for(i = 0; i < curr_view->nsaved_selection; i++)
+
+	sel_len = curr_view->nsaved_selection;
+	for(i = 0; i < sel_len && !ui_cancellation_requested(); i++)
 	{
 		char *filename = curr_view->saved_selection[i];
 		int pos = find_file_pos_in_list(curr_view, filename);
@@ -1183,13 +1144,15 @@ chown_files(int u, int g, uid_t uid, gid_t gid)
 #endif
 
 static void
-change_owner_cb(const char *new_owner)
+change_owner_cb(const char new_owner[])
 {
 #ifndef _WIN32
 	uid_t uid;
 
-	if(new_owner == NULL || new_owner[0] == '\0')
+	if(is_null_or_empty(new_owner))
+	{
 		return;
+	}
 
 	if(get_uid(new_owner, &uid) != 0)
 	{
@@ -1228,13 +1191,15 @@ change_owner(void)
 }
 
 static void
-change_group_cb(const char *new_group)
+change_group_cb(const char new_group[])
 {
 #ifndef _WIN32
 	gid_t gid;
 
-	if(new_group == NULL || new_group[0] == '\0')
+	if(is_null_or_empty(new_group))
+	{
 		return;
+	}
 
 	if(get_gid(new_group, &gid) != 0)
 	{
@@ -1273,14 +1238,16 @@ change_group(void)
 }
 
 static void
-change_link_cb(const char *new_target)
+change_link_cb(const char new_target[])
 {
 	char buf[MAX(COMMAND_GROUP_INFO_LEN, PATH_MAX)];
 	char linkto[PATH_MAX];
 	const char *filename;
 
-	if(new_target == NULL || new_target[0] == '\0')
+	if(is_null_or_empty(new_target))
+	{
 		return;
+	}
 
 	curr_stats.confirmed = 1;
 
@@ -1352,7 +1319,7 @@ prompt_dest_name(const char *src_name)
 {
 	wchar_t buf[256];
 
-	my_swprintf(buf, ARRAY_LEN(buf), L"New name for %" WPRINTF_MBSTR L": ",
+	vifm_swprintf(buf, ARRAY_LEN(buf), L"New name for %" WPRINTF_MBSTR L": ",
 			src_name);
 	enter_prompt_mode(buf, src_name, put_confirm_cb, NULL, 0);
 }
@@ -1367,9 +1334,15 @@ put_next(const char dest_name[], int override)
 	int from_trash;
 	int op;
 	int move;
+	int success;
 
 	/* TODO: refactor this function (put_next()) */
 
+	if(ui_cancellation_requested())
+	{
+		return -1;
+	}
+
 	override = override || put_confirm.overwrite_all;
 
 	filename = put_confirm.reg->files[put_confirm.x];
@@ -1377,54 +1350,59 @@ put_next(const char dest_name[], int override)
 	if(lstat(filename, &src_st) != 0)
 		return 0;
 
-	from_trash = strnoscmp(filename, cfg.trash_dir, strlen(cfg.trash_dir)) == 0;
+	from_trash = is_under_trash(filename);
 	move = from_trash || put_confirm.force_move;
 
-	if(dest_name[0] == '\0')
-		dest_name = find_slashr(filename) + 1;
-
 	copy_str(src_buf, sizeof(src_buf), filename);
 	chosp(src_buf);
 
-	if(from_trash)
+	if(dest_name[0] == '\0')
 	{
-		while(isdigit(*dest_name))
-			dest_name++;
-		dest_name++;
+		if(from_trash)
+		{
+			dest_name = get_real_name_from_trash_name(src_buf);
+		}
+		else
+		{
+			dest_name = find_slashr(src_buf) + 1;
+		}
 	}
 
 	snprintf(dst_buf, sizeof(dst_buf), "%s/%s", put_confirm.view->curr_dir,
 			dest_name);
 	chosp(dst_buf);
 
-	if(path_exists(dst_buf) && !override)
+	if(path_exists(dst_buf))
 	{
-		struct stat dst_st;
-		put_confirm.allow_merge = lstat(dst_buf, &dst_st) == 0 &&
-				S_ISDIR(dst_st.st_mode) && S_ISDIR(src_st.st_mode);
-		prompt_what_to_do(dest_name);
-		return 1;
-	}
-
-	if(override)
-	{
-		struct stat dst_st;
-		if(lstat(dst_buf, &dst_st) == 0 && (!put_confirm.merge ||
-				S_ISDIR(dst_st.st_mode) != S_ISDIR(src_st.st_mode)))
+		if(override)
 		{
-			if(perform_operation(OP_REMOVESL, NULL, dst_buf, NULL) != 0)
+			struct stat dst_st;
+			if(lstat(dst_buf, &dst_st) == 0 && (!put_confirm.merge ||
+					S_ISDIR(dst_st.st_mode) != S_ISDIR(src_st.st_mode)))
 			{
-				return 0;
+				if(perform_operation(OP_REMOVESL, NULL, dst_buf, NULL) != 0)
+				{
+					return 0;
+				}
+
+				/* Schedule view update to reflect changes in UI. */
+				request_view_update(put_confirm.view);
+			}
+			else
+			{
+#ifndef _WIN32
+				remove_last_path_component(dst_buf);
+#endif
 			}
 		}
 		else
 		{
-#ifndef _WIN32
-			remove_last_path_component(dst_buf);
-#endif
+			struct stat dst_st;
+			put_confirm.allow_merge = lstat(dst_buf, &dst_st) == 0 &&
+					S_ISDIR(dst_st.st_mode) && S_ISDIR(src_st.st_mode);
+			prompt_what_to_do(dest_name);
+			return 1;
 		}
-
-		request_view_update(put_confirm.view);
 	}
 
 	if(put_confirm.link)
@@ -1446,11 +1424,71 @@ put_next(const char dest_name[], int override)
 	}
 
 	progress_msg("Putting files", put_confirm.x + 1, put_confirm.reg->num_files);
-	if(perform_operation(op, NULL, src_buf, dst_buf) == 0)
+
+	/* Merging directory on move requires special handling as it can't be done by
+	 * "mv" itself. */
+	if(move && put_confirm.merge)
+	{
+		DIR *dir;
+
+		success = 1;
+
+		cmd_group_continue();
+
+		if((dir = opendir(src_buf)) != NULL)
+		{
+			struct dirent *d;
+			while((d = readdir(dir)) != NULL)
+			{
+				if(!is_builtin_dir(d->d_name))
+				{
+					char src_path[PATH_MAX];
+					char dst_path[PATH_MAX];
+					snprintf(src_path, sizeof(src_path), "%s/%s", src_buf, d->d_name);
+					snprintf(dst_path, sizeof(dst_path), "%s/%s/%s",
+							put_confirm.view->curr_dir, dest_name, d->d_name);
+					if(perform_operation(OP_MOVEF, NULL, src_path, dst_path) != 0)
+					{
+						success = 0;
+						break;
+					}
+					add_operation(OP_MOVEF, NULL, NULL, src_path, dst_path);
+				}
+			}
+			closedir(dir);
+		}
+		else
+		{
+			success = 0;
+		}
+
+		if(success)
+		{
+			success = (perform_operation(OP_RMDIR, NULL, src_buf, NULL) == 0);
+			if(success)
+			{
+				add_operation(OP_RMDIR, NULL, NULL, src_buf, "");
+			}
+		}
+
+		cmd_group_end();
+	}
+	else
+	{
+		success = (perform_operation(op, NULL, src_buf, dst_buf) == 0);
+	}
+
+	if(success)
 	{
 		char *msg, *p;
 		size_t len;
 
+		/* For some reason "mv" sometimes returns 0 on cancellation. */
+		if(!path_exists(dst_buf))
+		{
+			return -1;
+		}
+
 		cmd_group_continue();
 
 		msg = replace_group_msg(NULL);
@@ -1466,7 +1504,10 @@ put_next(const char dest_name[], int override)
 		replace_group_msg(msg);
 		free(msg);
 
-		add_operation(op, NULL, NULL, src_buf, dst_buf);
+		if(!(move && put_confirm.merge))
+		{
+			add_operation(op, NULL, NULL, src_buf, dst_buf);
+		}
 
 		cmd_group_end();
 		put_confirm.y++;
@@ -1481,12 +1522,9 @@ put_next(const char dest_name[], int override)
 }
 
 static void
-put_confirm_cb(const char *dest_name)
+put_confirm_cb(const char dest_name[])
 {
-	if(dest_name == NULL || dest_name[0] == '\0')
-		return;
-
-	if(put_next(dest_name, 0) == 0)
+	if(!is_null_or_empty(dest_name) && put_next(dest_name, 0) == 0)
 	{
 		put_confirm.x++;
 		curr_stats.save_msg = put_files_from_register_i(put_confirm.view, 0);
@@ -1494,9 +1532,9 @@ put_confirm_cb(const char *dest_name)
 }
 
 static void
-put_decide_cb(const char *choice)
+put_decide_cb(const char choice[])
 {
-	if(choice == NULL || choice[0] == '\0' || strcmp(choice, "r") == 0)
+	if(is_null_or_empty(choice) || strcmp(choice, "r") == 0)
 	{
 		prompt_dest_name(put_confirm.name);
 	}
@@ -1544,7 +1582,7 @@ prompt_what_to_do(const char src_name[])
 	wchar_t buf[NAME_MAX];
 
 	(void)replace_string(&put_confirm.name, src_name);
-	my_swprintf(buf, ARRAY_LEN(buf), L"Name conflict for %" WPRINTF_MBSTR
+	vifm_swprintf(buf, ARRAY_LEN(buf), L"Name conflict for %" WPRINTF_MBSTR
 			L". [r]ename/[s]kip/[o]verwrite/overwrite [a]ll%" WPRINTF_MBSTR ": ",
 			src_name, put_confirm.allow_merge ? "/[m]erge" : "");
 	enter_prompt_mode(buf, "", put_decide_cb, NULL, 0);
@@ -1589,10 +1627,10 @@ gen_clone_name(const char normal_name[])
 	size_t len;
 	char *p;
 
-	snprintf(result, sizeof(result), "%s", normal_name);
+	copy_str(result, sizeof(result), normal_name);
 	chosp(result);
 
-	snprintf(extension, sizeof(extension), "%s", extract_extension(result));
+	copy_str(extension, sizeof(extension), cut_extension(result));
 
 	len = strlen(result);
 	i = 1;
@@ -1750,8 +1788,10 @@ clone_files(FileView *view, char **list, int nlines, int force, int copies)
 		view->selected_files = 0;
 	}
 
+	ui_cancellation_reset();
+
 	cmd_group_begin(buf);
-	for(i = 0; i < sel_len; i++)
+	for(i = 0; i < sel_len && !ui_cancellation_requested(); i++)
 	{
 		int j;
 		const char * clone_name;
@@ -1787,8 +1827,7 @@ clone_files(FileView *view, char **list, int nlines, int force, int copies)
 	free_string_array(sel, sel_len);
 
 	clean_selected_files(view);
-	load_saving_pos(view, 1);
-	load_saving_pos(other_view, 1);
+	ui_views_reload_filelists();
 	if(from_file)
 		free_string_array(list, nlines);
 	return 0;
@@ -1824,10 +1863,10 @@ calc_dirsize(const char *path, int force_update)
 	{
 		char buf[PATH_MAX];
 
-		if(stroscmp(dentry->d_name, ".") == 0)
-			continue;
-		else if(stroscmp(dentry->d_name, "..") == 0)
+		if(is_builtin_dir(dentry->d_name))
+		{
 			continue;
+		}
 
 		snprintf(buf, sizeof(buf), "%s%s%s", path, slash, dentry->d_name);
 		if(entry_is_dir(buf, dentry))
@@ -1908,8 +1947,7 @@ put_files_from_register_i(FileView *view, int start)
 	{
 		char buf[MAX(COMMAND_GROUP_INFO_LEN, PATH_MAX + NAME_MAX*2 + 4)];
 		const char *op = "UNKNOWN";
-		int from_trash = strnoscmp(put_confirm.reg->files[0], cfg.trash_dir,
-				strlen(cfg.trash_dir)) == 0;
+		const int from_trash = is_under_trash(put_confirm.reg->files[0]);
 		if(put_confirm.link == 0)
 			op = (put_confirm.force_move || from_trash) ? "Put" : "put";
 		else if(put_confirm.link == 1)
@@ -1922,22 +1960,35 @@ put_files_from_register_i(FileView *view, int start)
 		cmd_group_end();
 	}
 
-	if(my_chdir(view->curr_dir) != 0)
+	if(vifm_chdir(view->curr_dir) != 0)
 	{
 		show_error_msg("Directory Return", "Can't chdir() to current directory");
 		return 1;
 	}
+
+	ui_cancellation_reset();
+
 	while(put_confirm.x < put_confirm.reg->num_files)
 	{
-		if(put_next("", 0) != 0)
+		const int put_result = put_next("", 0);
+		if(put_result > 0)
+		{
+			/* In this case put_next() takes care of interacting with a user. */
 			return 0;
+		}
+		else if(put_result < 0)
+		{
+			status_bar_messagef("%d file%s inserted%s", put_confirm.y,
+					(put_confirm.y == 1) ? "" : "s", get_cancellation_suffix());
+			return 1;
+		}
 		put_confirm.x++;
 	}
 
 	pack_register(put_confirm.reg->name);
 
-	status_bar_messagef("%d file%s inserted", put_confirm.y,
-			(put_confirm.y == 1) ? "" : "s");
+	status_bar_messagef("%d file%s inserted%s", put_confirm.y,
+			(put_confirm.y == 1) ? "" : "s", get_cancellation_suffix());
 
 	load_saving_pos(put_confirm.view, 1);
 
@@ -2082,12 +2133,12 @@ change_in_names(FileView *view, char c, const char *pattern, const char *sub,
 	{
 		char buf[NAME_MAX];
 
-		if(!view->dir_entry[i].selected)
-			continue;
-		if(is_parent_dir(view->dir_entry[i].name))
+		if(!view->dir_entry[i].selected || is_parent_dir(view->dir_entry[i].name))
+		{
 			continue;
+		}
 
-		strncpy(buf, view->dir_entry[i].name, sizeof(buf));
+		copy_str(buf, sizeof(buf), view->dir_entry[i].name);
 		chosp(buf);
 		j++;
 		if(strcmp(buf, dest[j]) == 0)
@@ -2098,8 +2149,10 @@ change_in_names(FileView *view, char c, const char *pattern, const char *sub,
 			(void)replace_string(&view->dir_entry[i].name, dest[j]);
 		}
 
-		if(mv_file(buf, view->curr_dir, dest[j], view->curr_dir, 0) == 0)
+		if(mv_file(buf, view->curr_dir, dest[j], view->curr_dir, 0, 1) == 0)
+		{
 			n++;
+		}
 	}
 	cmd_group_end();
 	free_string_array(dest, j + 1);
@@ -2148,12 +2201,12 @@ substitute_in_names(FileView *view, const char *pattern, const char *sub,
 		regmatch_t matches[10];
 		struct stat st;
 
-		if(!view->dir_entry[i].selected)
-			continue;
-		if(is_parent_dir(view->dir_entry[i].name))
+		if(!view->dir_entry[i].selected || is_parent_dir(view->dir_entry[i].name))
+		{
 			continue;
+		}
 
-		strncpy(buf, view->dir_entry[i].name, sizeof(buf));
+		copy_str(buf, sizeof(buf), view->dir_entry[i].name);
 		chosp(buf);
 		if(regexec(&re, buf, ARRAY_LEN(matches), matches, 0) != 0)
 		{
@@ -2246,12 +2299,12 @@ tr_in_names(FileView *view, const char *pattern, const char *sub)
 		const char *dst;
 		struct stat st;
 
-		if(!view->dir_entry[i].selected)
-			continue;
-		if(is_parent_dir(view->dir_entry[i].name))
+		if(!view->dir_entry[i].selected || is_parent_dir(view->dir_entry[i].name))
+		{
 			continue;
+		}
 
-		strncpy(buf, view->dir_entry[i].name, sizeof(buf));
+		copy_str(buf, sizeof(buf), view->dir_entry[i].name);
 		chosp(buf);
 		dst = substitute_tr(buf, pattern, sub);
 		if(strcmp(buf, dst) == 0)
@@ -2383,8 +2436,10 @@ change_case(FileView *view, int toupper, int count, int indexes[])
 			(void)replace_string(&view->dir_entry[pos].name, dest[i]);
 		}
 		if(mv_file(view->selected_filelist[i], view->curr_dir, dest[i],
-				view->curr_dir, 0) == 0)
+				view->curr_dir, 0, 1) == 0)
+		{
 			k++;
+		}
 	}
 	cmd_group_end();
 
@@ -2410,47 +2465,6 @@ is_copy_list_ok(const char *dst, int count, char **list)
 	return 1;
 }
 
-/* type:
- *  <= 0 - copy
- *  1 - absolute symbolic links
- *  2 - relative symbolic links
- */
-static int
-cp_file(const char *src_dir, const char *dst_dir, const char *src,
-		const char *dst, int type)
-{
-	char full_src[PATH_MAX], full_dst[PATH_MAX];
-	int op;
-	int result;
-
-	snprintf(full_src, sizeof(full_src), "%s/%s", src_dir, src);
-	chosp(full_src);
-	snprintf(full_dst, sizeof(full_dst), "%s/%s", dst_dir, dst);
-	chosp(full_dst);
-
-	if(strcmp(full_src, full_dst) == 0)
-		return 0;
-
-	if(type <= 0)
-	{
-		op = OP_COPY;
-	}
-	else
-	{
-		op = OP_SYMLINK;
-		if(type == 2)
-		{
-			snprintf(full_src, sizeof(full_src), "%s", make_rel_path(full_src,
-					dst_dir));
-		}
-	}
-
-	result = perform_operation(op, NULL, full_src, full_dst);
-	if(result == 0 && type >= 0)
-		add_operation(op, NULL, NULL, full_src, full_dst);
-	return result;
-}
-
 static int
 cpmv_prepare(FileView *view, char ***list, int *nlines, int move, int type,
 		int force, char *buf, size_t buf_len, char *path, int *from_file,
@@ -2526,7 +2540,7 @@ cpmv_prepare(FileView *view, char ***list, int *nlines, int move, int type,
 		view->list_pos = i;
 	}
 
-	*from_trash = path_starts_with(view->curr_dir, cfg.trash_dir);
+	*from_trash = is_under_trash(view->curr_dir);
 	return 0;
 }
 
@@ -2596,7 +2610,7 @@ edit_file(const char filepath[], int force_changed)
 {
 	struct stat st_before, st_after;
 
-	if(force_changed && stat(filepath, &st_before) != 0)
+	if(!force_changed && stat(filepath, &st_before) != 0)
 	{
 		show_error_msgf("Error Editing File",
 				"Could not stat file \"%s\" before edit: %s", filepath,
@@ -2611,7 +2625,7 @@ edit_file(const char filepath[], int force_changed)
 		return -1;
 	}
 
-	if(force_changed && stat(filepath, &st_after) != 0)
+	if(!force_changed && stat(filepath, &st_after) != 0)
 	{
 		show_error_msgf("Error Editing File",
 				"Could not stat file \"%s\" after edit: %s", filepath, strerror(errno));
@@ -2662,21 +2676,24 @@ cpmv_files(FileView *view, char **list, int nlines, int move, int type,
 		erase_selection(view);
 	}
 
+	ui_cancellation_reset();
+
 	processed = 0;
 	cmd_group_begin(buf);
-	for(i = 0; i < sel_len; i++)
+	for(i = 0; i < sel_len && !ui_cancellation_requested(); i++)
 	{
 		char dst_full[PATH_MAX];
 		const char *dst = (nlines > 0) ? list[i] : sel[i];
 		if(from_trash)
 		{
-			while(isdigit(*dst))
-				dst++;
-			dst++;
+			char src_full[PATH_MAX];
+			snprintf(src_full, sizeof(src_full), "%s/%s", view->curr_dir, dst);
+			chosp(src_full);
+			dst = get_real_name_from_trash_name(src_full);
 		}
 
 		snprintf(dst_full, sizeof(dst_full), "%s/%s", path, dst);
-		if(path_exists(dst_full))
+		if(path_exists(dst_full) && !from_trash)
 		{
 			(void)perform_operation(OP_REMOVESL, NULL, dst_full, NULL);
 		}
@@ -2685,17 +2702,23 @@ cpmv_files(FileView *view, char **list, int nlines, int move, int type,
 		{
 			progress_msg("Moving files", i + 1, sel_len);
 
-			if(mv_file(sel[i], view->curr_dir, dst, path, 0) != 0)
+			if(mv_file(sel[i], view->curr_dir, dst, path, 0, 1) != 0)
+			{
 				view->list_pos = find_file_pos_in_list(view, sel[i]);
+			}
 			else
+			{
 				processed++;
+			}
 		}
 		else
 		{
 			if(type == 0)
 				progress_msg("Copying files", i + 1, sel_len);
-			if(cp_file(view->curr_dir, path, sel[i], dst, type) == 0)
+			if(cp_file(view->curr_dir, path, sel[i], dst, type, 1) == 0)
+			{
 				processed++;
+			}
 		}
 	}
 	cmd_group_end();
@@ -2703,13 +2726,12 @@ cpmv_files(FileView *view, char **list, int nlines, int move, int type,
 	free_string_array(sel, sel_len);
 	free_selected_file_array(view);
 	clean_selected_files(view);
-	load_saving_pos(view, 1);
-  load_saving_pos(other_view, 1);
+	ui_views_reload_filelists();
 	if(from_file)
 		free_string_array(list, nlines);
 
-	status_bar_messagef("%d file%s successfully processed", processed,
-			(processed == 1) ? "" : "s");
+	status_bar_messagef("%d file%s successfully processed%s", processed,
+			(processed == 1) ? "" : "s", get_cancellation_suffix());
 
 	return 1;
 }
@@ -2725,27 +2747,112 @@ cpmv_files_bg_i(char **list, int nlines, int move, int force, char **sel_list,
 		const char *dst = (nlines > 0) ? list[i] : sel_list[i];
 		if(from_trash)
 		{
-			while(isdigit(*dst))
-				dst++;
-			dst++;
+			char src_full[PATH_MAX];
+			snprintf(src_full, sizeof(src_full), "%s/%s", src, dst);
+			chosp(src_full);
+			dst = get_real_name_from_trash_name(src_full);
 		}
 
 		snprintf(dst_full, sizeof(dst_full), "%s/%s", path, dst);
-		if(path_exists(dst_full))
+		if(path_exists(dst_full) && !from_trash)
 		{
-			perform_operation(OP_REMOVESL, NULL, dst_full, NULL);
+			perform_operation(OP_REMOVESL, (void *)1, dst_full, NULL);
 		}
 
 		if(move)
-			(void)mv_file(sel_list[i], src, dst, path, -1);
+		{
+			(void)mv_file(sel_list[i], src, dst, path, -1, 0);
+		}
 		else
-			(void)cp_file(src, path, sel_list[i], dst, -1);
+		{
+			(void)cp_file(src, path, sel_list[i], dst, -1, 0);
+		}
 
 		inner_bg_next();
 	}
 	return 0;
 }
 
+static int
+mv_file(const char src[], const char src_path[], const char dst[],
+		const char path[], int tmpfile_num, int cancellable)
+{
+	char full_src[PATH_MAX], full_dst[PATH_MAX];
+	int op;
+	int result;
+
+	snprintf(full_src, sizeof(full_src), "%s/%s", src_path, src);
+	chosp(full_src);
+	snprintf(full_dst, sizeof(full_dst), "%s/%s", path, dst);
+	chosp(full_dst);
+
+	/* compare case sensitive strings even on Windows to let user rename file
+	 * changing only case of some characters */
+	if(strcmp(full_src, full_dst) == 0)
+		return 0;
+
+	if(tmpfile_num <= 0)
+		op = OP_MOVE;
+	else if(tmpfile_num == 1)
+		op = OP_MOVETMP1;
+	else if(tmpfile_num == 2)
+		op = OP_MOVETMP2;
+	else if(tmpfile_num == 3)
+		op = OP_MOVETMP3;
+	else if(tmpfile_num == 4)
+		op = OP_MOVETMP4;
+	else
+		op = OP_NONE;
+
+	result = perform_operation(op, cancellable ? NULL : (void *)1, full_src,
+			full_dst);
+	if(result == 0 && tmpfile_num >= 0)
+		add_operation(op, NULL, NULL, full_src, full_dst);
+	return result;
+}
+
+/* type:
+ *  <= 0 - copy
+ *  1 - absolute symbolic links
+ *  2 - relative symbolic links
+ */
+static int
+cp_file(const char src_dir[], const char dst_dir[], const char src[],
+		const char dst[], int type, int cancellable)
+{
+	char full_src[PATH_MAX], full_dst[PATH_MAX];
+	int op;
+	int result;
+
+	snprintf(full_src, sizeof(full_src), "%s/%s", src_dir, src);
+	chosp(full_src);
+	snprintf(full_dst, sizeof(full_dst), "%s/%s", dst_dir, dst);
+	chosp(full_dst);
+
+	if(strcmp(full_src, full_dst) == 0)
+		return 0;
+
+	if(type <= 0)
+	{
+		op = OP_COPY;
+	}
+	else
+	{
+		op = OP_SYMLINK;
+		if(type == 2)
+		{
+			snprintf(full_src, sizeof(full_src), "%s", make_rel_path(full_src,
+					dst_dir));
+		}
+	}
+
+	result = perform_operation(op, cancellable ? NULL : (void *)1, full_src,
+			full_dst);
+	if(result == 0 && type >= 0)
+		add_operation(op, NULL, NULL, full_src, full_dst);
+	return result;
+}
+
 static void *
 cpmv_stub(void *arg)
 {
@@ -2791,21 +2898,9 @@ cpmv_files_bg(FileView *view, char **list, int nlines, int move, int force)
 	else
 		args->list = copy_string_array(list, nlines);
 
-	args->sel_list = view->selected_filelist;
-	args->sel_list_len = view->selected_files;
+	general_prepare_for_bg_task(view, args);
 
-	view->selected_filelist = NULL;
-	free_selected_file_array(view);
-	clean_selected_files(view);
-	load_saving_pos(view, 1);
-
-	strcpy(args->src, view->curr_dir);
-
-#ifndef _WIN32
-	args->job = add_background_job(-1, buf, -1);
-#else
-	args->job = add_background_job(-1, buf, (HANDLE)-1);
-#endif
+	args->job = add_background_job(-1, buf, NO_JOB_ID);
 	if(args->job == NULL)
 	{
 		free_string_array(args->list, args->nlines);
@@ -2821,6 +2916,20 @@ cpmv_files_bg(FileView *view, char **list, int nlines, int move, int force)
 	return 0;
 }
 
+/* Fills basic fields of the args structure. */
+static void
+general_prepare_for_bg_task(FileView *view, bg_args_t *args)
+{
+	args->sel_list = view->selected_filelist;
+	args->sel_list_len = view->selected_files;
+
+	view->selected_filelist = NULL;
+	free_selected_file_array(view);
+	ui_view_reset_selection_and_reload(view);
+
+	copy_str(args->src, sizeof(args->src), view->curr_dir);
+}
+
 static void
 go_to_first_file(FileView *view, char **names, int count)
 {
@@ -2875,13 +2984,15 @@ make_dirs(FileView *view, char **names, int count, int create_parent)
 		}
 	}
 
+	ui_cancellation_reset();
+
 	snprintf(buf, sizeof(buf), "mkdir in %s: ",
 			replace_home_part(view->curr_dir));
 
 	get_group_file_list(names, count, buf);
 	cmd_group_begin(buf);
 	n = 0;
-	for(i = 0; i < count; i++)
+	for(i = 0; i < count && !ui_cancellation_requested(); i++)
 	{
 		char full[PATH_MAX];
 		snprintf(full, sizeof(full), "%s/%s", view->curr_dir, names[i]);
@@ -2911,10 +3022,8 @@ make_dirs(FileView *view, char **names, int count, int create_parent)
 		go_to_first_file(view, names, count);
 	}
 
-	if(n == 1)
-		status_bar_message("1 directory created");
-	else
-		status_bar_messagef("%d directories created", n);
+	status_bar_messagef("%d director%s created", n, (n == 1) ? "y" : "ies",
+			get_cancellation_suffix());
 }
 
 int
@@ -2958,7 +3067,7 @@ make_files(FileView *view, char **names, int count)
 	get_group_file_list(names, count, buf);
 	cmd_group_begin(buf);
 	n = 0;
-	for(i = 0; i < count; i++)
+	for(i = 0; i < count && !ui_cancellation_requested(); i++)
 	{
 		char full[PATH_MAX];
 		snprintf(full, sizeof(full), "%s/%s", view->curr_dir, names[i]);
@@ -2973,10 +3082,59 @@ make_files(FileView *view, char **names, int count)
 	if(n > 0)
 		go_to_first_file(view, names, count);
 
-	status_bar_messagef("%d file%s created", n, (n == 1) ? "" : "s");
+	status_bar_messagef("%d file%s created%s", n, (n == 1) ? "" : "s",
+			get_cancellation_suffix());
 	return 1;
 }
 
+int
+restore_files(FileView *view)
+{
+	int i;
+	int m = 0;
+	int n = view->selected_files;
+
+	i = view->list_pos;
+	while(i < view->list_rows - 1 && view->dir_entry[i].selected)
+	{
+		i++;
+	}
+	view->list_pos = i;
+
+	ui_cancellation_reset();
+
+	cmd_group_begin("restore: ");
+	cmd_group_end();
+	for(i = 0; i < view->list_rows && !ui_cancellation_requested(); i++)
+	{
+		if(view->dir_entry[i].selected)
+		{
+			char full_path[PATH_MAX];
+			snprintf(full_path, sizeof(full_path), "%s/%s", view->curr_dir,
+					view->dir_entry[i].name);
+			chosp(full_path);
+
+			if(restore_from_trash(full_path) == 0)
+			{
+				m++;
+			}
+		}
+	}
+
+	load_saving_pos(view, 1);
+
+	status_bar_messagef("Restored %d of %d%s", m, n, get_cancellation_suffix());
+	return 1;
+}
+
+/* Provides different suffixes depending on whether cancellation was requested
+ * or not.  Returns pointer to a string literal. */
+static const char *
+get_cancellation_suffix(void)
+{
+	return ui_cancellation_requested() ? " (cancelled)" : "";
+}
+
 int
 check_if_dir_writable(DirRole dir_role, const char *path)
 {
