@@ -32,6 +32,7 @@ typedef enum
 	MACRO_STATUSBAR_OUTPUT, /* redirect output to the status bar */
 	MACRO_SPLIT, /* run command in a new screen region */
 	MACRO_IGNORE, /* completely ignore command output */
+	MACRO_NO_TERM_MUX, /* Forbid using of terminal multiplexer, even if active. */
 }
 MacroFlags;
 
@@ -40,7 +41,8 @@ typedef struct
 {
 	char letter;       /* Macro identifier in the pattern. */
 	const char *value; /* A value to replace macro with. */
-	int uses_left;     /* Number of mandatory uses of the macro. */
+	int uses_left;     /* Number of mandatory uses of the macro for group head. */
+	int group;         /* Index of macro group head or -1. */
 }
 custom_macro_t;
 
