@@ -25,44 +25,46 @@
 #include "../cfg/config.h"
 #include "../modes/menu.h"
 #include "../utils/string_array.h"
+#include "../commands.h"
 #include "../ui.h"
 #include "menus.h"
 
 static int show_history(FileView *view, int type, hist_t *hist,
 		const char title[]);
+static int execute_history_cb(FileView *view, menu_info *m);
 
 int
 show_cmdhistory_menu(FileView *view)
 {
-	return show_history(view, CMDHISTORY, &cfg.cmd_hist,
+	return show_history(view, CMDHISTORY_MENU, &cfg.cmd_hist,
 			" Command Line History ");
 }
 
 int
 show_fsearchhistory_menu(FileView *view)
 {
-	return show_history(view, FSEARCHHISTORY, &cfg.search_hist,
+	return show_history(view, FSEARCHHISTORY_MENU, &cfg.search_hist,
 			" Search History ");
 }
 
 int
 show_bsearchhistory_menu(FileView *view)
 {
-	return show_history(view, BSEARCHHISTORY, &cfg.search_hist,
+	return show_history(view, BSEARCHHISTORY_MENU, &cfg.search_hist,
 			" Search History ");
 }
 
 int
 show_prompthistory_menu(FileView *view)
 {
-	return show_history(view, PROMPTHISTORY, &cfg.prompt_hist,
+	return show_history(view, PROMPTHISTORY_MENU, &cfg.prompt_hist,
 			" Prompt History ");
 }
 
 int
 show_filterhistory_menu(FileView *view)
 {
-	return show_history(view, FILTERHISTORY, &cfg.filter_hist,
+	return show_history(view, FILTERHISTORY_MENU, &cfg.filter_hist,
 			" Filter History ");
 }
 
@@ -75,6 +77,7 @@ show_history(FileView *view, int type, hist_t *hist, const char title[])
 
 	init_menu_info(&m, type, strdup("History disabled or empty"));
 	m.title = strdup(title);
+	m.execute_handler = &execute_history_cb;
 
 	for(i = 0; i <= hist->pos; i++)
 	{
@@ -84,5 +87,27 @@ show_history(FileView *view, int type, hist_t *hist, const char title[])
 	return display_menu(&m, view);
 }
 
+static int
+execute_history_cb(FileView *view, menu_info *m)
+{
+	switch(m->type)
+	{
+		case CMDHISTORY_MENU:
+			save_command_history(m->items[m->pos]);
+			exec_commands(m->items[m->pos], view, GET_COMMAND);
+			break;
+		case FSEARCHHISTORY_MENU:
+			save_search_history(m->items[m->pos]);
+			exec_commands(m->items[m->pos], view, GET_FSEARCH_PATTERN);
+			break;
+		case BSEARCHHISTORY_MENU:
+			save_search_history(m->items[m->pos]);
+			exec_commands(m->items[m->pos], view, GET_BSEARCH_PATTERN);
+			break;
+	}
+
+	return 0;
+}
+
 /* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab cinoptions-=(0 : */
 /* vim: set cinoptions+=t0 : */
