@@ -31,6 +31,8 @@
 #include "../ui.h"
 #include "menus.h"
 
+static int execute_locate_cb(FileView *view, menu_info *m);
+
 int
 show_locate_menu(FileView *view, const char args[])
 {
@@ -38,13 +40,15 @@ show_locate_menu(FileView *view, const char args[])
 	int save_msg;
 	custom_macro_t macros[] =
 	{
-		{ .letter = 'a', .value = args, .uses_left = 1 },
+		{ .letter = 'a', .value = args, .uses_left = 1, .group = -1 },
 	};
 
 	static menu_info m;
-	init_menu_info(&m, LOCATE, strdup("No files found"));
+	init_menu_info(&m, LOCATE_MENU, strdup("No files found"));
 	m.args = (args[0] == '-') ? strdup(args) : escape_filename(args, 0);
 	m.title = format_str(" Locate %s ", m.args);
+	m.execute_handler = &execute_locate_cb;
+	m.key_handler = &filelist_khandler;
 
 	status_bar_message("locate...");
 
@@ -55,5 +59,14 @@ show_locate_menu(FileView *view, const char args[])
 	return save_msg;
 }
 
+/* Callback that is called when menu item is selected.  Should return non-zero
+ * to stay in menu mode. */
+static int
+execute_locate_cb(FileView *view, menu_info *m)
+{
+	goto_selected_file(view, m->items[m->pos], 0);
+	return 0;
+}
+
 /* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab cinoptions-=(0 : */
 /* vim: set cinoptions+=t0 : */
