@@ -27,15 +27,19 @@
 
 #include <dirent.h> /* DIR */
 #include <sys/stat.h>
-#include <sys/types.h>
-#include <unistd.h> /* access() */
+#include <sys/types.h> /* pid_t */
+#include <unistd.h> /* access() F_OK R_OK */
 
 #include <assert.h> /* assert() */
 #include <ctype.h> /* isspace() */
+#include <errno.h> /* errno */
 #include <stddef.h> /* NULL size_t */
-#include <string.h> /* memset() strdup() strchr() strlen() strrchr() */
+#include <stdlib.h> /* free() malloc() */
+#include <string.h> /* memmove() memset() strdup() strcat() strncat() strchr()
+                       strlen() strrchr() */
 #include <stdarg.h>
 #include <signal.h>
+#include <wchar.h> /* wchar_t wcscmp() */
 
 #include "../cfg/config.h"
 #include "../modes/cmdline.h"
@@ -43,25 +47,31 @@
 #include "../modes/modes.h"
 #include "../utils/file_streams.h"
 #include "../utils/fs.h"
+#include "../utils/log.h"
+#include "../utils/path.h"
 #include "../utils/str.h"
 #include "../utils/string_array.h"
+#include "../utils/test_helpers.h"
 #include "../utils/utf8.h"
+#include "../utils/utils.h"
 #include "../background.h"
 #include "../bookmarks.h"
-#include "../commands.h"
 #include "../filelist.h"
 #include "../macros.h"
 #include "../running.h"
 #include "../search.h"
 #include "../status.h"
 #include "../ui.h"
-#include "all.h"
 
 static int prompt_error_msg_internalv(const char title[], const char format[],
 		int prompt_skip, va_list pa);
 static int prompt_error_msg_internal(const char title[], const char message[],
 		int prompt_skip);
+TSTATIC char * parse_spec(const char spec[], int *line_num);
+static void open_selected_file(const char path[], int line_num);
+static void navigate_to_selected_file(FileView *view, const char path[]);
 static void normalize_top(menu_info *m);
+static void append_to_string(char **str, const char suffix[]);
 static char * expand_tabulation_a(const char line[], size_t tab_stops);
 static size_t chars_in_str(const char s[], char c);
 static void redraw_error_msg(const char title_arg[], const char message_arg[],
@@ -76,6 +86,25 @@ show_position_in_menu(menu_info *m)
 	ui_pos_window_set(pos_buf);
 }
 
+void
+remove_current_item(menu_info *m)
+{
+	clean_menu_position(m);
+
+	remove_from_string_array(m->items, m->len, m->pos);
+	if(m->matches != NULL)
+	{
+		if(m->matches[m->pos])
+			m->matching_entries--;
+		memmove(m->matches + m->pos, m->matches + m->pos + 1,
+				sizeof(int)*((m->len - 1) - m->pos));
+	}
+	m->len--;
+	draw_menu(m);
+
+	move_to_menu_pos(m->pos, m);
+}
+
 void
 clean_menu_position(menu_info *m)
 {
@@ -195,9 +224,11 @@ prompt_error_msg_internalv(const char title[], const char format[],
 	return prompt_error_msg_internal(title, buf, prompt_skip);
 }
 
-/* Internal function for displaying messages to a user.  When the prompt_skip
+/* Internal function for displaying error messages to a user.  Automatically
+ * skips whitespace in front of the message and does nothing for empty messages
+ * (due to skipping whitespace-only are counted as empty). When the prompt_skip
  * isn't zero, asks user about successive messages.  Returns non-zero if all
- * successive messages should be skipped. */
+ * successive messages should be skipped, otherwise zero is returned. */
 static int
 prompt_error_msg_internal(const char title[], const char message[],
 		int prompt_skip)
@@ -210,6 +241,12 @@ prompt_error_msg_internal(const char title[], const char message[],
 	if(curr_stats.load_stage < 2 && skip_until_started)
 		return 1;
 
+	message = skip_whitespace(message);
+	if(*message == '\0')
+	{
+		return 0;
+	}
+
 	curr_stats.errmsg_shown = 1;
 
 	redraw_error_msg(title, message, prompt_skip);
@@ -416,166 +453,162 @@ redraw_menu(menu_info *m)
 	wrefresh(menu_win);
 }
 
-static void
-goto_selected_file(FileView *view, menu_info *m)
+void
+goto_selected_file(FileView *view, const char spec[], int try_open)
 {
-	char *dir;
-	char *file;
-	char *free_this;
-	char *num = NULL;
-	char *p = NULL;
+	char *path_buf;
+	int line_num;
 
-	free_this = file = dir = malloc(2 + strlen(m->items[m->pos]) + 1 + 1);
-	if(free_this == NULL)
+	path_buf = parse_spec(spec, &line_num);
+	if(path_buf == NULL)
 	{
 		show_error_msg("Memory Error", "Unable to allocate enough memory");
 		return;
 	}
 
-	if(m->items[m->pos][0] != '/')
-		strcpy(dir, "./");
+	if(access(path_buf, F_OK) == 0)
+	{
+		if(try_open)
+		{
+			open_selected_file(path_buf, line_num);
+		}
+		else
+		{
+			navigate_to_selected_file(view, path_buf);
+		}
+	}
 	else
-		dir[0] = '\0';
-	if(m->type == GREP)
 	{
-		p = strchr(m->items[m->pos], ':');
-		if(p != NULL)
-		{
-			*p = '\0';
-			num = p + 1;
-		}
-		else
-		{
-			num = NULL;
-		}
-	}
-	strcat(dir, m->items[m->pos]);
-	chomp(file);
-	if(m->type == GREP && p != NULL)
-	{
-		int n = 1;
-
-		*p = ':';
-
-		if(num != NULL)
-			n = atoi(num);
-
-		if(access(file, R_OK) == 0)
-		{
-			curr_stats.auto_redraws = 1;
-			(void)view_file(file, n, -1, 1);
-			curr_stats.auto_redraws = 0;
-		}
-		free(free_this);
-		return;
+		show_error_msgf("Missing file", "File \"%s\" doesn't exist", path_buf);
 	}
 
-	if(access(file, R_OK) == 0)
-	{
-		int isdir = 0;
-		char *last_slash;
-
-		if(is_dir(file))
-			isdir = 1;
-
-		if((last_slash = strrchr(dir, '/')) != NULL)
-		{
-			*last_slash = '\0';
-			file = last_slash + 1;
-		}
-
-		if(change_directory(view, dir) >= 0)
-		{
-			status_bar_message("Finding the correct directory...");
-
-			wrefresh(status_bar);
-			load_dir_list(view, 0);
-			if(isdir)
-				strcat(file, "/");
-
-			(void)ensure_file_is_selected(view, file);
-		}
-		else
-		{
-			show_error_msgf("Invalid path", "Cannot change dir to \"%s\"", dir);
-		}
-	}
-	else if(m->type == LOCATE || m->type == USER_NAVIGATE)
-	{
-		show_error_msgf("Missing file", "File \"%s\" doesn't exist", file);
-	}
-
-	free(free_this);
+	free(path_buf);
 }
 
-int
-execute_menu_cb(FileView *view, menu_info *m)
+/* Extracts path and line number from the spec (1 when absent from the spec).
+ * Returns path and sets *line_num to line number, otherwise NULL is
+ * returned. */
+TSTATIC char *
+parse_spec(const char spec[], int *line_num)
 {
-	switch(m->type)
+	char *path_buf;
+	const char *colon;
+	int colon_lookup_offset = 0;
+	const size_t bufs_len = 2 + strlen(spec) + 1 + 1;
+
+	path_buf = malloc(bufs_len);
+	if(path_buf == NULL)
 	{
-		case APROPOS:
-			execute_apropos_cb(m);
-			return 1;
-		case BOOKMARK:
-			move_to_bookmark(view, index2mark(active_bookmarks[m->pos]));
-			break;
-		case CMDHISTORY:
-			save_command_history(m->items[m->pos]);
-			exec_commands(m->items[m->pos], view, GET_COMMAND);
-			break;
-		case FSEARCHHISTORY:
-			save_search_history(m->items[m->pos]);
-			exec_commands(m->items[m->pos], view, GET_FSEARCH_PATTERN);
-			break;
-		case BSEARCHHISTORY:
-			save_search_history(m->items[m->pos]);
-			exec_commands(m->items[m->pos], view, GET_BSEARCH_PATTERN);
-			break;
-		case COLORSCHEME:
-			load_color_scheme(m->items[m->pos]);
-			break;
-		case COMMAND:
-			break_at(m->items[m->pos], ' ');
-			exec_command(m->items[m->pos], view, GET_COMMAND);
-			break;
-		case FILETYPE:
-			execute_filetype_cb(view, m);
-			break;
-		case DIRHISTORY:
-			if(!cfg.auto_ch_pos)
-			{
-				clean_positions_in_history(curr_view);
-				curr_stats.ch_pos = 0;
-			}
-			navigate_to(view, m->items[m->pos]);
-			if(!cfg.auto_ch_pos)
-				curr_stats.ch_pos = 1;
-			break;
-		case DIRSTACK:
-			execute_dirstack_cb(view, m);
-			break;
-		case JOBS:
-			execute_jobs_cb(view, m);
-			break;
-		case LOCATE:
-		case FIND:
-		case USER_NAVIGATE:
-			goto_selected_file(view, m);
-			break;
-		case GREP:
-			goto_selected_file(view, m);
-			return 1;
-		case VIFM:
-			break;
+		return NULL;
+	}
+
+	if(is_path_absolute(spec))
+	{
+		path_buf[0] = '\0';
+	}
+	else
+	{
+		copy_str(path_buf, bufs_len, "./");
+	}
+
 #ifdef _WIN32
-		case VOLUMES:
-			execute_volumes_cb(view, m);
-			break;
+	if(is_path_absolute(spec))
+	{
+		colon_lookup_offset = 2;
+	}
 #endif
-		default:
-			break;
+
+	colon = strchr(spec + colon_lookup_offset, ':');
+	if(colon != NULL)
+	{
+		strncat(path_buf, spec, colon - spec);
+		*line_num = atoi(colon + 1);
+	}
+	else
+	{
+		strcat(path_buf, spec);
+		*line_num = 1;
+	}
+
+	chomp(path_buf);
+
+#ifdef _WIN32
+	to_forward_slash(path_buf);
+#endif
+
+	return path_buf;
+}
+
+/* Opens file specified by its path on the given line number. */
+static void
+open_selected_file(const char path[], int line_num)
+{
+	if(access(path, R_OK) == 0)
+	{
+		(void)view_file(path, line_num, -1, 1);
+	}
+	else
+	{
+		show_error_msgf("Can't read file", "File \"%s\" is not readable", path);
+	}
+}
+
+/* Navigates the view to a given dir/file combination specified by the path. */
+static void
+navigate_to_selected_file(FileView *view, const char path[])
+{
+	/* Check whether target path is directory while we don't change current
+	 * working directory by invoking change_directory() function below. */
+	const int dst_is_dir = is_dir(path);
+
+	char name[NAME_MAX];
+	char *dir = strdup(path);
+	char *const last_slash = find_slashr(dir);
+
+	if(last_slash == NULL)
+	{
+		copy_str(name, sizeof(name), dir);
+	}
+	else
+	{
+		*last_slash = '\0';
+		copy_str(name, sizeof(name), last_slash + 1);
+	}
+
+	if(change_directory(view, dir) >= 0)
+	{
+		status_bar_message("Finding the correct directory...");
+		wrefresh(status_bar);
+
+		load_dir_list(view, 0);
+
+		if(dst_is_dir)
+		{
+			strcat(name, "/");
+		}
+		(void)ensure_file_is_selected(view, name);
+	}
+	else
+	{
+		show_error_msgf("Invalid path", "Cannot change dir to \"%s\"", dir);
+	}
+
+	free(dir);
+}
+
+void
+goto_selected_directory(FileView *view, menu_info *m)
+{
+	if(!cfg.auto_ch_pos)
+	{
+		clean_positions_in_history(curr_view);
+		curr_stats.ch_pos = 0;
+	}
+	navigate_to(view, m->items[m->pos]);
+	if(!cfg.auto_ch_pos)
+	{
+		curr_stats.ch_pos = 1;
 	}
-	return 0;
 }
 
 void
@@ -674,15 +707,25 @@ capture_output_to_menu(FileView *view, const char cmd[], menu_info *m)
 	FILE *file, *err;
 	char *line = NULL;
 	int x;
+	pid_t pid;
 
-	if(background_and_capture((char *)cmd, &file, &err) != 0)
+	LOG_INFO_MSG("Capturing output of the command to a menu: %s", cmd);
+
+	pid = background_and_capture((char *)cmd, &file, &err);
+	if(pid == (pid_t)-1)
 	{
 		show_error_msgf("Trouble running command", "Unable to run: %s", cmd);
 		return 0;
 	}
 
-	x = 0;
 	show_progress("", 0);
+
+	ui_cancellation_reset();
+	ui_cancellation_enable();
+
+	wait_for_data_from(pid, file, 0);
+
+	x = 0;
 	while((line = read_line(file, line)) != NULL)
 	{
 		char *expanded_line;
@@ -693,15 +736,40 @@ capture_output_to_menu(FileView *view, const char cmd[], menu_info *m)
 		{
 			m->items[x++] = expanded_line;
 		}
+
+		wait_for_data_from(pid, file, 0);
 	}
 	m->len = x;
 
+	ui_cancellation_disable();
+
 	fclose(file);
 	print_errors(err);
 
+	if(ui_cancellation_requested())
+	{
+		append_to_string(&m->title, "(cancelled) ");
+		append_to_string(&m->empty_msg, " (cancelled)");
+	}
+
 	return display_menu(m, view);
 }
 
+/* Replaces *str with a copy of the with string extended by the suffix.  *str
+ * can be NULL in which case it's treated as empty string. equal to the with (then function does nothing).  Returns non-zero if memory allocation
+ * failed. */
+static void
+append_to_string(char **str, const char suffix[])
+{
+	const char *const non_null_str = (*str == NULL) ? "" : *str;
+	char *const appended_str = format_str("%s%s", non_null_str, suffix);
+	if(appended_str != NULL)
+	{
+		free(*str);
+		*str = appended_str;
+	}
+}
+
 /* Clones the line replacing all occurrences of horizontal tabulation character
  * with appropriate number of spaces.  The tab_stops parameter shows how many
  * character position are taken by one tabulation.  Returns newly allocated
@@ -930,5 +998,21 @@ get_cmd_target(void)
 		expand_macros("%f", NULL, NULL, 1) : strdup(".");
 }
 
+KHandlerResponse
+filelist_khandler(menu_info *m, const wchar_t keys[])
+{
+	if(wcscmp(keys, L"gf") == 0)
+	{
+		goto_selected_file(curr_view, m->items[m->pos], 0);
+		return KHR_CLOSE_MENU;
+	}
+	else if(wcscmp(keys, L"e") == 0)
+	{
+		goto_selected_file(curr_view, m->items[m->pos], 1);
+		return KHR_REFRESH_WINDOW;
+	}
+	return KHR_UNHANDLED;
+}
+
 /* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab cinoptions-=(0 : */
 /* vim: set cinoptions+=t0 : */
