@@ -27,7 +27,7 @@
 #include <ctype.h>
 #include <stddef.h> /* NULL size_t */
 #include <stdlib.h> /* free() */
-#include <string.h>
+#include <string.h> /* strdup() */
 #include <wchar.h> /* wcswidth() */
 #include <wctype.h>
 
@@ -38,6 +38,7 @@
 #include "../engine/options.h"
 #include "../menus/menus.h"
 #include "../utils/fs_limits.h"
+#include "../utils/path.h"
 #include "../utils/str.h"
 #include "../utils/test_helpers.h"
 #include "../utils/utils.h"
@@ -88,6 +89,7 @@ typedef struct
 	int old_top;              /* for search_mode */
 	int old_pos;              /* for search_mode */
 	int line_edited;          /* Cache for whether input line changed flag. */
+	int entered_by_mapping;   /* The mode was entered by a mapping. */
 }
 line_stats_t;
 
@@ -139,6 +141,23 @@ static void search_next(void);
 static void complete_next(const hist_t *hist, size_t len);
 static void cmd_ctrl_u(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_ctrl_w(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_ctrl_xa(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_ctrl_xc(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_ctrl_xxc(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_ctrl_xd(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_ctrl_xxd(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_ctrl_xe(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_ctrl_xxe(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_ctrl_xm(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_ctrl_xr(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_ctrl_xxr(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_ctrl_xt(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_ctrl_xxt(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_ctrl_xequals(key_info_t key_info, keys_info_t *keys_info);
+static void paste_name_part(const char name[], int root);
+static void paste_str(const char str[], int allow_escaping);
+static char * escape_cmd_for_pasting(const char str[]);
+static const wchar_t * get_nonnull_input(void);
 static void cmd_ctrl_underscore(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_meta_b(key_info_t key_info, keys_info_t *keys_info);
 static void find_prev_word(void);
@@ -148,6 +167,7 @@ static void cmd_meta_dot(key_info_t key_info, keys_info_t *keys_info);
 static void remove_previous_dot_completion(void);
 static wchar_t * next_dot_completion(void);
 static int insert_dot_completion(const wchar_t completion[]);
+static int insert_str(const wchar_t str[]);
 static void find_next_word(void);
 static void cmd_left(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_right(key_info_t key_info, keys_info_t *keys_info);
@@ -212,6 +232,19 @@ static keys_add_info_t builtin_cmds[] = {
 	{L"\x04",         {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_delete}}},
 	{L"\x15",         {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_u}}},
 	{L"\x17",         {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_w}}},
+	{L"\x18"L"a",     {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_xa}}},
+	{L"\x18"L"c",     {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_xc}}},
+	{L"\x18\x18"L"c", {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_xxc}}},
+	{L"\x18"L"d",     {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_xd}}},
+	{L"\x18\x18"L"d", {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_xxd}}},
+	{L"\x18"L"e",     {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_xe}}},
+	{L"\x18\x18"L"e", {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_xxe}}},
+	{L"\x18"L"m",     {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_xm}}},
+	{L"\x18"L"r",     {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_xr}}},
+	{L"\x18\x18"L"r", {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_xxr}}},
+	{L"\x18"L"t",     {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_xt}}},
+	{L"\x18\x18"L"t", {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_xxt}}},
+	{L"\x18"L"=",     {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_xequals}}},
 #ifndef __PDCURSES__
 	{L"\x1b"L"b",     {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_meta_b}}},
 	{L"\x1b"L"d",     {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_meta_d}}},
@@ -373,7 +406,7 @@ input_line_changed(void)
 		char *mbinput;
 
 		free(previous);
-		previous = my_wcsdup(input_stat.line);
+		previous = vifm_wcsdup(input_stat.line);
 
 		mbinput = to_multibyte(input_stat.line);
 
@@ -458,6 +491,7 @@ void
 enter_cmdline_mode(CMD_LINE_SUBMODES cl_sub_mode, const wchar_t *cmd, void *ptr)
 {
 	const wchar_t *prompt;
+	complete_cmd_func complete_func;
 
 	sub_mode_ptr = ptr;
 	sub_mode = cl_sub_mode;
@@ -485,7 +519,8 @@ enter_cmdline_mode(CMD_LINE_SUBMODES cl_sub_mode, const wchar_t *cmd, void *ptr)
 		prompt = L"E";
 	}
 
-	prepare_cmdline_mode(prompt, cmd, complete_cmd);
+	complete_func = (sub_mode == FILTER_SUBMODE) ? NULL : complete_cmd;
+	prepare_cmdline_mode(prompt, cmd, complete_func);
 }
 
 void
@@ -553,6 +588,7 @@ prepare_cmdline_mode(const wchar_t *prompt, const wchar_t *cmd,
 	input_stat.search_mode = 0;
 	input_stat.dot_pos = -1;
 	input_stat.line_edited = 0;
+	input_stat.entered_by_mapping = is_inside_mapping();
 
 	if(sub_mode == SEARCH_FORWARD_SUBMODE
 			|| sub_mode == VSEARCH_FORWARD_SUBMODE
@@ -586,7 +622,7 @@ prepare_cmdline_mode(const wchar_t *prompt, const wchar_t *cmd,
 		{
 			wcscpy(input_stat.line, cmd);
 			input_stat.curs_pos += wcswidth(input_stat.line, (size_t)-1);
-			input_stat.initial_line = my_wcsdup(input_stat.line);
+			input_stat.initial_line = vifm_wcsdup(input_stat.line);
 		}
 	}
 
@@ -759,8 +795,7 @@ cmd_ctrl_g(key_info_t key_info, keys_info_t *keys_info)
 	const int prompt_ee = sub_mode == PROMPT_SUBMODE && sub_mode_allows_ee;
 	if(type != -1 || prompt_ee)
 	{
-		char *const mbstr = (input_stat.line == NULL) ?
-			strdup("") : to_multibyte(input_stat.line);
+		char *const mbstr = to_multibyte(get_nonnull_input());
 		leave_cmdline_mode();
 
 		if(sub_mode == FILTER_SUBMODE)
@@ -917,13 +952,17 @@ static void
 do_completion(void)
 {
 	if(input_stat.complete == NULL)
+	{
 		return;
+	}
 
 	if(input_stat.line == NULL)
 	{
-		input_stat.line = my_wcsdup(L"");
+		input_stat.line = vifm_wcsdup(L"");
 		if(input_stat.line == NULL)
+		{
 			return;
+		}
 	}
 
 	line_completion(&input_stat);
@@ -1051,7 +1090,8 @@ static void
 cmd_ctrl_m(key_info_t key_info, keys_info_t *keys_info)
 {
 	/* TODO: refactor this cmd_ctrl_m() function. */
-	char* p;
+	char *p;
+	const char *nonnull_input;
 
 	stop_completion();
 	werase(status_bar);
@@ -1064,6 +1104,7 @@ cmd_ctrl_m(key_info_t key_info, keys_info_t *keys_info)
 	}
 
 	p = input_stat.line ? to_multibyte(input_stat.line) : NULL;
+	nonnull_input = (p != NULL) ? p : "";
 
 	leave_cmdline_mode();
 
@@ -1075,13 +1116,17 @@ cmd_ctrl_m(key_info_t key_info, keys_info_t *keys_info)
 
 	if(sub_mode == CMD_SUBMODE || sub_mode == MENU_CMD_SUBMODE)
 	{
-		char* s = (p != NULL) ? p : "";
-		while(*s == ' ' || *s == ':')
-			s++;
-		if(sub_mode == CMD_SUBMODE)
-			curr_stats.save_msg = exec_commands(s, curr_view, GET_COMMAND);
-		else
-			curr_stats.save_msg = exec_commands(s, curr_view, GET_MENU_COMMAND);
+		const CMD_LINE_SUBMODES cmd_type = (sub_mode == CMD_SUBMODE)
+		                                 ? GET_COMMAND
+		                                 : GET_MENU_COMMAND;
+
+		const char *real_start = nonnull_input;
+		while(*real_start == ' ' || *real_start == ':')
+		{
+			real_start++;
+		}
+
+		curr_stats.save_msg = exec_commands(real_start, curr_view, cmd_type);
 	}
 	else if(sub_mode == PROMPT_SUBMODE)
 	{
@@ -1095,7 +1140,7 @@ cmd_ctrl_m(key_info_t key_info, keys_info_t *keys_info)
 		}
 		else
 		{
-			local_filter_apply(curr_view, p);
+			local_filter_apply(curr_view, nonnull_input);
 			load_saving_pos(curr_view, 1);
 		}
 	}
@@ -1134,13 +1179,21 @@ cmd_ctrl_m(key_info_t key_info, keys_info_t *keys_info)
 	}
 	else if(cfg.inc_search && input_stat.search_mode)
 	{
-		/* In case of successful search and 'hlsearch' option set, a message like
-		 * "n files selected" is printed automatically. */
-		if(curr_view->matches == 0 || !cfg.hl_search)
+		if(prev_mode == MENU_MODE)
 		{
-			print_search_msg(curr_view, is_backward_search(sub_mode));
+			menu_print_search_msg(sub_mode_ptr);
 			curr_stats.save_msg = 1;
 		}
+		else
+		{
+			/* In case of successful search and 'hlsearch' option set, a message like
+			* "n files selected" is printed automatically. */
+			if(curr_view->matches == 0 || !cfg.hl_search)
+			{
+				print_search_msg(curr_view, is_backward_search(sub_mode));
+				curr_stats.save_msg = 1;
+			}
+		}
 	}
 
 	free(p);
@@ -1161,7 +1214,9 @@ save_input_to_history(const keys_info_t *keys_info, const char input[])
 	}
 	else if(sub_mode == CMD_SUBMODE)
 	{
-		if(!keys_info->mapped && !keys_info->recursive)
+		const int mapped_input = input_stat.entered_by_mapping && keys_info->mapped;
+		const int ignore_input = mapped_input || keys_info->recursive;
+		if(!ignore_input)
 		{
 			save_command_history(input);
 		}
@@ -1210,7 +1265,7 @@ static void
 save_users_input(void)
 {
 	free(input_stat.line_buf);
-	input_stat.line_buf = my_wcsdup((input_stat.line != NULL) ?
+	input_stat.line_buf = vifm_wcsdup((input_stat.line != NULL) ?
 			input_stat.line : L"");
 }
 
@@ -1219,7 +1274,7 @@ restore_user_input(void)
 {
 	input_stat.cmd_pos = -1;
 	free(input_stat.line);
-	input_stat.line = my_wcsdup(input_stat.line_buf);
+	input_stat.line = vifm_wcsdup(input_stat.line_buf);
 	input_stat.len = wcslen(input_stat.line);
 	update_cmdline();
 }
@@ -1397,6 +1452,185 @@ cmd_ctrl_w(key_info_t key_info, keys_info_t *keys_info)
 	update_cmdline_text();
 }
 
+/* Inserts value of automatic filter of active pane into current cursor
+ * position. */
+static void
+cmd_ctrl_xa(key_info_t key_info, keys_info_t *keys_info)
+{
+	paste_str(curr_view->auto_filter.raw, 0);
+}
+
+/* Inserts name of the current file of active pane into current cursor
+ * position. */
+static void
+cmd_ctrl_xc(key_info_t key_info, keys_info_t *keys_info)
+{
+	paste_str(get_current_file_name(curr_view), 1);
+}
+
+/* Inserts name of the current file of inactive pane into current cursor
+ * position. */
+static void
+cmd_ctrl_xxc(key_info_t key_info, keys_info_t *keys_info)
+{
+	paste_str(get_current_file_name(other_view), 1);
+}
+
+/* Inserts path to the current directory of active pane into current cursor
+ * position. */
+static void
+cmd_ctrl_xd(key_info_t key_info, keys_info_t *keys_info)
+{
+	paste_str(curr_view->curr_dir, 1);
+}
+
+/* Inserts path to the current directory of inactive pane into current cursor
+ * position. */
+static void
+cmd_ctrl_xxd(key_info_t key_info, keys_info_t *keys_info)
+{
+	paste_str(other_view->curr_dir, 1);
+}
+
+/* Inserts extension of the current file of active pane into current cursor
+ * position. */
+static void
+cmd_ctrl_xe(key_info_t key_info, keys_info_t *keys_info)
+{
+	paste_name_part(get_current_file_name(curr_view), 0);
+}
+
+/* Inserts extension of the current file of inactive pane into current cursor
+ * position. */
+static void
+cmd_ctrl_xxe(key_info_t key_info, keys_info_t *keys_info)
+{
+	paste_name_part(get_current_file_name(other_view), 0);
+}
+
+/* Inserts value of manual filter of active pane into current cursor
+ * position. */
+static void
+cmd_ctrl_xm(key_info_t key_info, keys_info_t *keys_info)
+{
+	paste_str(curr_view->manual_filter.raw, 0);
+}
+
+/* Inserts name root of current file of active pane into current cursor
+ * position. */
+static void
+cmd_ctrl_xr(key_info_t key_info, keys_info_t *keys_info)
+{
+	paste_name_part(get_current_file_name(curr_view), 1);
+}
+
+/* Inserts name root of current file of inactive pane into current cursor
+ * position. */
+static void
+cmd_ctrl_xxr(key_info_t key_info, keys_info_t *keys_info)
+{
+	paste_name_part(get_current_file_name(other_view), 1);
+}
+
+/* Inserts last component of path to the current directory of active pane into
+ * current cursor position. */
+static void
+cmd_ctrl_xt(key_info_t key_info, keys_info_t *keys_info)
+{
+	paste_str(get_last_path_component(curr_view->curr_dir), 1);
+}
+
+/* Inserts last component of path to the current directory of inactive pane into
+ * current cursor position. */
+static void
+cmd_ctrl_xxt(key_info_t key_info, keys_info_t *keys_info)
+{
+	paste_str(get_last_path_component(other_view->curr_dir), 1);
+}
+
+/* Inserts value of local filter of active pane into current cursor position. */
+static void
+cmd_ctrl_xequals(key_info_t key_info, keys_info_t *keys_info)
+{
+	paste_str(curr_view->local_filter.filter.raw, 0);
+}
+
+/* Inserts root/extension of the name file into current cursor position. */
+static void
+paste_name_part(const char name[], int root)
+{
+	int root_len;
+	const char *ext;
+	char *const name_copy = strdup(name);
+
+	split_ext(name_copy, &root_len, &ext);
+	paste_str(root ? name_copy : ext, 1);
+
+	free(name_copy);
+}
+
+/* Inserts string into current cursor position and updates command-line on the
+ * screen. */
+static void
+paste_str(const char str[], int allow_escaping)
+{
+	char *escaped;
+	wchar_t *wide;
+
+	if(prev_mode == MENU_MODE)
+	{
+		return;
+	}
+
+	stop_completion();
+
+	escaped = (allow_escaping && sub_mode == CMD_SUBMODE)
+	        ? escape_cmd_for_pasting(str)
+	        : NULL;
+	if(escaped != NULL)
+	{
+		str = escaped;
+	}
+
+	wide = to_wide(str);
+
+	if(insert_str(wide) == 0)
+	{
+		update_cmdline_text();
+	}
+
+	free(wide);
+	free(escaped);
+}
+
+/* Escapes str for inserting into current position of command-line command when
+ * needed.  Returns escaped string or NULL when no escaping is needed. */
+static char *
+escape_cmd_for_pasting(const char str[])
+{
+	wchar_t *const wide_input = vifm_wcsdup(get_nonnull_input());
+	char *mb_input;
+	char *escaped;
+
+	wide_input[input_stat.index] = L'\0';
+	mb_input = to_multibyte(wide_input);
+
+	escaped = commands_escape_for_insertion(mb_input, strlen(mb_input), str);
+
+	free(mb_input);
+	free(wide_input);
+
+	return escaped;
+}
+
+/* Gets current input in a safe form.  Returning empty string on NULL input,
+ * otherwise input line is returned. */
+static const wchar_t *
+get_nonnull_input(void)
+{
+	return (input_stat.line == NULL) ? L"" : input_stat.line;
+}
+
 static void
 cmd_ctrl_underscore(key_info_t key_info, keys_info_t *keys_info)
 {
@@ -1553,27 +1787,41 @@ next_dot_completion(void)
 static int
 insert_dot_completion(const wchar_t completion[])
 {
-	wchar_t *ptr;
-	size_t len = wcslen(completion);
+	if(insert_str(completion) == 0)
+	{
+		input_stat.dot_index = input_stat.index;
+		input_stat.dot_len = wcslen(completion);
+		return 0;
+	}
+	return 1;
+}
 
-	ptr = realloc(input_stat.line, (input_stat.len + len + 1)*sizeof(wchar_t));
+/* Inserts wide string into current cursor position.  Returns zero on success,
+ * otherwise non-zero is returned. */
+static int
+insert_str(const wchar_t str[])
+{
+	const size_t len = wcslen(str);
+	const size_t new_len = input_stat.len + len + 1;
+
+	wchar_t *const ptr = realloc(input_stat.line, new_len*sizeof(wchar_t));
 	if(ptr == NULL)
 	{
 		return 1;
 	}
 
-	input_stat.dot_index = input_stat.index;
-	input_stat.dot_len = len;
-
 	if(input_stat.line == NULL)
 	{
 		ptr[0] = L'\0';
 	}
 	input_stat.line = ptr;
-	wcsins(input_stat.line, completion, input_stat.index + 1);
+
+	wcsins(input_stat.line, str, input_stat.index + 1);
+
 	input_stat.index += len;
 	input_stat.curs_pos += len;
 	input_stat.len += len;
+
 	return 0;
 }
 
@@ -1836,9 +2084,11 @@ line_part_complete(line_stats_t *stat, const char *line_mb, const char *p,
 	const size_t new_len = (p - line_mb) + mbstowcs(NULL, completed, 0)
 			+ (stat->len - stat->index) + 1;
 
-	line_ending = my_wcsdup(stat->line + stat->index);
+	line_ending = vifm_wcsdup(stat->line + stat->index);
 	if(line_ending == NULL)
+	{
 		return -1;
+	}
 
 	if((t = realloc(stat->line, new_len * sizeof(wchar_t))) == NULL)
 	{
@@ -1847,7 +2097,7 @@ line_part_complete(line_stats_t *stat, const char *line_mb, const char *p,
 	}
 	stat->line = t;
 
-	my_swprintf(stat->line + (p - line_mb), new_len,
+	vifm_swprintf(stat->line + (p - line_mb), new_len,
 			L"%" WPRINTF_MBSTR L"%" WPRINTF_WSTR, completed, line_ending);
 	free(line_ending);
 
