@@ -36,6 +36,7 @@
 #include "../../filelist.h"
 #include "../../fileops.h"
 #include "../../status.h"
+#include "../../ui.h"
 #include "../../undo.h"
 #include "../modes.h"
 
@@ -467,6 +468,8 @@ files_chmod(FileView *view, const char *mode, int recurse_dirs)
 {
 	int i;
 
+	ui_cancellation_reset();
+
 	i = 0;
 	while(i < view->list_rows && !view->dir_entry[i].selected)
 		i++;
@@ -505,8 +508,9 @@ files_chmod(FileView *view, const char *mode, int recurse_dirs)
 			}
 			i++;
 		}
+
 		cmd_group_begin(buf);
-		while(j < view->list_rows)
+		while(j < view->list_rows && !ui_cancellation_requested())
 		{
 			if(view->dir_entry[j].selected)
 			{
