@@ -152,8 +152,7 @@ leave_change_mode(int clean_selection)
 
 	if(clean_selection)
 	{
-		clean_selected_files(view);
-		load_saving_pos(view, 1);
+		ui_view_reset_selection_and_reload(view);
 	}
 
 	update_all_windows();
