@@ -56,20 +56,21 @@ static int indexes[] = {
 	0,               /* SORT_BY_EXTENSION */
 	1,               /* SORT_BY_NAME */
 #ifndef _WIN32
-	3,               /* SORT_BY_GROUP_ID */
-	4,               /* SORT_BY_GROUP_NAME */
-	5,               /* SORT_BY_MODE */
-	7,               /* SORT_BY_OWNER_ID */
-	8,               /* SORT_BY_OWNER_NAME */
+	4,               /* SORT_BY_GROUP_ID */
+	5,               /* SORT_BY_GROUP_NAME */
+	6,               /* SORT_BY_MODE */
+	8,               /* SORT_BY_OWNER_ID */
+	9,               /* SORT_BY_OWNER_NAME */
 #endif
-	9 + CORRECTION,  /* SORT_BY_SIZE */
-	10 + CORRECTION, /* SORT_BY_TIME_ACCESSED */
-	11 + CORRECTION, /* SORT_BY_TIME_CHANGED */
-	12 + CORRECTION, /* SORT_BY_TIME_MODIFIED */
+	10 + CORRECTION, /* SORT_BY_SIZE */
+	11 + CORRECTION, /* SORT_BY_TIME_ACCESSED */
+	12 + CORRECTION, /* SORT_BY_TIME_CHANGED */
+	13 + CORRECTION, /* SORT_BY_TIME_MODIFIED */
 	2,               /* SORT_BY_INAME */
 #ifndef _WIN32
-	6,               /* SORT_BY_PERMISSIONS */
+	7,               /* SORT_BY_PERMISSIONS */
 #endif
+	3,               /* SORT_BY_TYPE */
 };
 ARRAY_GUARD(indexes, 1 + SORT_OPTION_COUNT);
 
@@ -140,7 +141,7 @@ enter_sort_mode(FileView *active_view)
 	curs_set(FALSE);
 	update_all_windows();
 
-	top = 3;
+	top = 4;
 	bottom = top + SORT_OPTION_COUNT - 1;
 	curr = top + indexes[abs(view->sort[0])];
 	col = 6;
@@ -153,18 +154,19 @@ redraw_sort_dialog(void)
 {
 	int x, y, cy;
 
-	wresize(sort_win, SORT_OPTION_COUNT + 5, SORT_WIN_WIDTH);
+	wresize(sort_win, SORT_OPTION_COUNT + 6, SORT_WIN_WIDTH);
 
 	werase(sort_win);
 	box(sort_win, ACS_VLINE, ACS_HLINE);
 
 	getmaxyx(sort_win, y, x);
 	mvwaddstr(sort_win, 0, (x - 6)/2, " Sort ");
-	mvwaddstr(sort_win, top - 1, 2, " Sort files by:");
+	mvwaddstr(sort_win, top - 2, 2, " Sort files by:");
 	cy = top;
 	mvwaddstr(sort_win, cy++, 4, " [   ] File Extenstion");
 	mvwaddstr(sort_win, cy++, 4, " [   ] Name");
 	mvwaddstr(sort_win, cy++, 4, " [   ] Name (ignore case)");
+	mvwaddstr(sort_win, cy++, 4, " [   ] Type");
 #ifndef _WIN32
 	mvwaddstr(sort_win, cy++, 4, " [   ] Group ID");
 	mvwaddstr(sort_win, cy++, 4, " [   ] Group Name");
@@ -195,8 +197,7 @@ leave_sort_mode(void)
 {
 	*mode = NORMAL_MODE;
 
-	clean_selected_files(view);
-	load_saving_pos(view, 1);
+	ui_view_reset_selection_and_reload(view);
 
 	update_all_windows();
 }
