@@ -97,6 +97,7 @@ static void cmd_ctrl_wt(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_ctrl_wv(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_ctrl_ww(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_ctrl_wx(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_ctrl_wz(key_info_t key_info, keys_info_t *keys_info);
 static FileView * get_view(void);
 static void move_splitter(key_info_t key_info, int fact);
 static void cmd_ctrl_x(key_info_t key_info, keys_info_t *keys_info);
@@ -255,6 +256,8 @@ static keys_add_info_t builtin_cmds[] = {
 	{L"\x17w", {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_ww}}},
 	{L"\x17\x18", {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_wx}}},
 	{L"\x17x", {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_wx}}},
+	{L"\x17\x1a", {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_wz}}},
+	{L"\x17z", {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_wz}}},
 	{L"\x17=", {BUILTIN_NIM_KEYS, FOLLOWED_BY_NONE, {.handler = normal_cmd_ctrl_wequal}}},
 	{L"\x17<", {BUILTIN_NIM_KEYS, FOLLOWED_BY_NONE, {.handler = normal_cmd_ctrl_wless}}},
 	{L"\x17>", {BUILTIN_NIM_KEYS, FOLLOWED_BY_NONE, {.handler = normal_cmd_ctrl_wgreater}}},
@@ -520,7 +523,7 @@ cmd_emarkemark(key_info_t key_info, keys_info_t *keys_info)
 		}
 		else
 		{
-			my_swprintf(buf, ARRAY_LEN(buf), L".,.+%d!", key_info.count - 1);
+			vifm_swprintf(buf, ARRAY_LEN(buf), L".,.+%d!", key_info.count - 1);
 		}
 	}
 	enter_cmdline_mode(CMD_SUBMODE, buf, NULL);
@@ -602,32 +605,20 @@ cmd_ctrl_r(key_info_t key_info, keys_info_t *keys_info)
 	int ret;
 
 	curr_stats.confirmed = 0;
+	ui_cancellation_reset();
+
+	status_bar_message("Redoing...");
 
 	ret = redo_group();
+
 	if(ret == 0)
 	{
-		if(curr_stats.view)
-		{
-			load_saving_pos(curr_view, 1);
-		}
-		else
-		{
-			load_saving_pos(&lwin, 1);
-			load_saving_pos(&rwin, 1);
-		}
+		ui_views_reload_visible_filelists();
 		status_bar_message("Redone one group");
 	}
 	else if(ret == -2)
 	{
-		if(curr_stats.view)
-		{
-			load_saving_pos(curr_view, 1);
-		}
-		else
-		{
-			load_saving_pos(&lwin, 1);
-			load_saving_pos(&rwin, 1);
-		}
+		ui_views_reload_visible_filelists();
 		status_bar_error("Redone one group with errors");
 	}
 	else if(ret == -1)
@@ -646,6 +637,11 @@ cmd_ctrl_r(key_info_t key_info, keys_info_t *keys_info)
 	{
 		status_bar_message("Group redo skipped by user");
 	}
+	else if(ret == -7)
+	{
+		ui_views_reload_visible_filelists();
+		status_bar_message("Redoing was cancelled");
+	}
 	else if(ret == 1)
 	{
 		status_bar_error("Redo operation was skipped due to previous errors");
@@ -839,7 +835,7 @@ move_splitter(key_info_t key_info, int fact)
 	update_screen(UT_REDRAW);
 }
 
-/* Switch views. */
+/* Switches views. */
 static void
 cmd_ctrl_wx(key_info_t key_info, keys_info_t *keys_info)
 {
@@ -851,6 +847,13 @@ cmd_ctrl_wx(key_info_t key_info, keys_info_t *keys_info)
 	}
 }
 
+/* Quits preview pane or view modes. */
+static void
+cmd_ctrl_wz(key_info_t key_info, keys_info_t *keys_info)
+{
+	preview_close();
+}
+
 static void
 cmd_ctrl_x(key_info_t key_info, keys_info_t *keys_info)
 {
@@ -1270,16 +1273,19 @@ cmd_quote(key_info_t key_info, keys_info_t *keys_info)
 {
 	if(keys_info->selector)
 	{
-		int pos = check_mark_directory(curr_view, key_info.multi);
-		if(pos < 0)
-			return;
-		pick_files(curr_view, pos, keys_info);
+		const int pos = check_mark_directory(curr_view, key_info.multi);
+		if(pos >= 0)
+		{
+			pick_files(curr_view, pos, keys_info);
+		}
 	}
 	else
 	{
-		curr_stats.save_msg = get_bookmark(curr_view, key_info.multi);
+		curr_stats.save_msg = goto_bookmark(curr_view, key_info.multi);
 		if(!cfg.auto_ch_pos)
+		{
 			move_to_list_pos(curr_view, 0);
+		}
 	}
 }
 
@@ -1351,7 +1357,9 @@ cmd_colon(key_info_t key_info, keys_info_t *keys_info)
 {
 	wchar_t buf[16] = L"";
 	if(key_info.count != NO_COUNT_GIVEN)
-		my_swprintf(buf, ARRAY_LEN(buf), L".,.+%d", key_info.count - 1);
+	{
+		vifm_swprintf(buf, ARRAY_LEN(buf), L".,.+%d", key_info.count - 1);
+	}
 	enter_cmdline_mode(CMD_SUBMODE, buf, NULL);
 }
 
@@ -1388,8 +1396,7 @@ cmd_al(key_info_t key_info, keys_info_t *keys_info)
 	if(key_info.reg == NO_REG_GIVEN)
 		key_info.reg = DEFAULT_REG_NAME;
 	curr_stats.save_msg = put_links(curr_view, key_info.reg, 0);
-	load_saving_pos(&lwin, 1);
-	load_saving_pos(&rwin, 1);
+	ui_views_reload_filelists();
 }
 
 /* Change word (rename file without extension). */
@@ -1634,7 +1641,7 @@ go_to_next(key_info_t key_info, keys_info_t *keys_info, int def, int step)
 static void
 cmd_m(key_info_t key_info, keys_info_t *keys_info)
 {
-	add_bookmark(key_info.multi, curr_view->curr_dir,
+	curr_stats.save_msg = set_user_bookmark(key_info.multi, curr_view->curr_dir,
 			get_current_file_name(curr_view));
 }
 
@@ -1700,8 +1707,7 @@ put_files(key_info_t key_info, int move)
 	if(key_info.reg == NO_REG_GIVEN)
 		key_info.reg = DEFAULT_REG_NAME;
 	curr_stats.save_msg = put_files_from_register(curr_view, key_info.reg, move);
-	load_saving_pos(&lwin, 1);
-	load_saving_pos(&rwin, 1);
+	ui_views_reload_filelists();
 }
 
 /* Create link with absolute path */
@@ -1711,8 +1717,7 @@ cmd_rl(key_info_t key_info, keys_info_t *keys_info)
 	if(key_info.reg == NO_REG_GIVEN)
 		key_info.reg = DEFAULT_REG_NAME;
 	curr_stats.save_msg = put_links(curr_view, key_info.reg, 1);
-	load_saving_pos(&lwin, 1);
-	load_saving_pos(&rwin, 1);
+	ui_views_reload_filelists();
 }
 
 /* Runs external editor to get command-line command and then executes it. */
@@ -1790,32 +1795,20 @@ cmd_u(key_info_t key_info, keys_info_t *keys_info)
 	int ret;
 
 	curr_stats.confirmed = 0;
+	ui_cancellation_reset();
+
+	status_bar_message("Undoing...");
 
 	ret = undo_group();
+
 	if(ret == 0)
 	{
-		if(curr_stats.view)
-		{
-			load_saving_pos(curr_view, 1);
-		}
-		else
-		{
-			load_saving_pos(&lwin, 1);
-			load_saving_pos(&rwin, 1);
-		}
+		ui_views_reload_visible_filelists();
 		status_bar_message("Undone one group");
 	}
 	else if(ret == -2)
 	{
-		if(curr_stats.view)
-		{
-			load_saving_pos(curr_view, 1);
-		}
-		else
-		{
-			load_saving_pos(&lwin, 1);
-			load_saving_pos(&rwin, 1);
-		}
+		ui_views_reload_visible_filelists();
 		status_bar_error("Undone one group with errors");
 	}
 	else if(ret == -1)
@@ -1838,6 +1831,11 @@ cmd_u(key_info_t key_info, keys_info_t *keys_info)
 	{
 		status_bar_message("Group undo skipped by user");
 	}
+	else if(ret == -7)
+	{
+		ui_views_reload_visible_filelists();
+		status_bar_message("Undoing was cancelled");
+	}
 	else if(ret == 1)
 	{
 		status_bar_error("Undo operation was skipped due to previous errors");
