@@ -85,6 +85,7 @@ enum
 };
 
 static int try_ressurect_abandoned(const char full_path[], int explore);
+static void try_redraw_explore_view(const FileView *const view, int vi_index);
 static void reset_view_info(view_info_t *vi);
 static void init_view_info(view_info_t *vi);
 static void free_view_info(view_info_t *vi);
@@ -114,6 +115,7 @@ static int is_top_or_left(void);
 static void cmd_ctrl_wv(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_ctrl_ww(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_ctrl_wx(key_info_t key_info, keys_info_t *keys_info);
+static void cmd_ctrl_wz(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_meta_space(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_percent(key_info_t key_info, keys_info_t *keys_info);
 static void cmd_tab(key_info_t key_info, keys_info_t *keys_info);
@@ -192,6 +194,8 @@ static keys_add_info_t builtin_cmds[] = {
 	{L"\x17w", {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_ww}}},
 	{L"\x17\x18", {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_wx}}},
 	{L"\x17x", {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_wx}}},
+	{L"\x17\x1a", {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_wz}}},
+	{L"\x17z", {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_wz}}},
 	{L"\x17\x13", {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_ws}}},
 	{L"\x17s", {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_ws}}},
 	{L"\x17\x16", {BUILTIN_KEYS, FOLLOWED_BY_NONE, {.handler = cmd_ctrl_wv}}},
@@ -286,6 +290,7 @@ enter_view_mode(int explore)
 	/* Either make use of abandoned view or prune it. */
 	if(try_ressurect_abandoned(full_path, explore) == 0)
 	{
+		ui_views_update_titles();
 		return;
 	}
 
@@ -299,8 +304,6 @@ enter_view_mode(int explore)
 	vi->filename = strdup(full_path);
 
 	*mode = VIEW_MODE;
-	update_view_title(&lwin);
-	update_view_title(&rwin);
 
 	if(explore)
 	{
@@ -312,6 +315,8 @@ enter_view_mode(int explore)
 		vi->view = other_view;
 	}
 
+	ui_views_update_titles();
+
 	view_redraw();
 }
 
@@ -385,16 +390,9 @@ view_redraw(void)
 
 	colmgr_reset();
 
-	if(lwin.explore_mode)
-	{
-		vi = &view_info[VI_LWIN];
-		redraw();
-	}
-	if(rwin.explore_mode)
-	{
-		vi = &view_info[VI_RWIN];
-		redraw();
-	}
+	try_redraw_explore_view(&lwin, VI_LWIN);
+	try_redraw_explore_view(&rwin, VI_RWIN);
+
 	if(!lwin.explore_mode && !rwin.explore_mode)
 	{
 		redraw();
@@ -403,6 +401,18 @@ view_redraw(void)
 	vi = saved_vi;
 }
 
+/* Redraws view in explore mode if view is really in explore mode and is visible
+ * on the screen. */
+static void
+try_redraw_explore_view(const FileView *const view, int vi_index)
+{
+	if(view->explore_mode && ui_view_is_visible(view))
+	{
+		vi = &view_info[vi_index];
+		redraw();
+	}
+}
+
 void
 leave_view_mode(void)
 {
@@ -418,7 +428,7 @@ leave_view_mode(void)
 		quick_view_file(curr_view);
 	}
 
-	update_view_title(curr_view);
+	ui_view_title_update(curr_view);
 
 	reset_view_info(vi);
 
@@ -442,7 +452,7 @@ view_explore_mode_quit(FileView *view)
 	reset_view_info(&view_info[(view == &lwin) ? VI_LWIN : VI_RWIN]);
 
 	redraw_view(view);
-	update_view_title(view);
+	ui_view_title_update(view);
 }
 
 /* Frees and initializes anew view_into_t structure instance. */
@@ -481,7 +491,9 @@ free_view_info(view_info_t *vi)
 	free_string_array(vi->lines, vi->nlines);
 	free(vi->widths);
 	if(vi->last_search_backward != -1)
+	{
 		regfree(&vi->re);
+	}
 	free(vi->filename);
 }
 
@@ -489,7 +501,7 @@ free_view_info(view_info_t *vi)
 static void
 redraw(void)
 {
-	update_view_title(vi->view);
+	ui_view_title_update(vi->view);
 	calc_vlines();
 	draw();
 }
@@ -792,8 +804,11 @@ cmd_ctrl_ww(key_info_t key_info, keys_info_t *keys_info)
 	{
 		go_to_other_pane();
 	}
+
+	ui_views_update_titles();
 }
 
+/* Switches views. */
 static void
 cmd_ctrl_wx(key_info_t key_info, keys_info_t *keys_info)
 {
@@ -806,6 +821,14 @@ cmd_ctrl_wx(key_info_t key_info, keys_info_t *keys_info)
 	}
 }
 
+/* Quits preview pane or view modes. */
+static void
+cmd_ctrl_wz(key_info_t key_info, keys_info_t *keys_info)
+{
+	leave_view_mode();
+	preview_close();
+}
+
 static void
 cmd_meta_space(key_info_t key_info, keys_info_t *keys_info)
 {
@@ -840,11 +863,12 @@ cmd_tab(key_info_t key_info, keys_info_t *keys_info)
 
 	change_window();
 	if(!curr_view->explore_mode)
+	{
 		*mode = NORMAL_MODE;
+	}
 	pick_vi(curr_view->explore_mode);
 
-	update_view_title(&lwin);
-	update_view_title(&rwin);
+	ui_views_update_titles();
 }
 
 /* Updates value of the vi variable and points it to the correct element of the
