@@ -756,7 +756,7 @@ go_to_next(key_info_t key_info, keys_info_t *keys_info, int def, int step)
 static void
 cmd_m(key_info_t key_info, keys_info_t *keys_info)
 {
-	add_bookmark(key_info.multi, view->curr_dir,
+	set_user_bookmark(key_info.multi, view->curr_dir,
 			get_current_file_name(view));
 }
 
@@ -870,19 +870,20 @@ leave_clearing_selection(int go_to_top, int save_msg)
 static void
 update_marks(FileView *view)
 {
+	char start_mark, end_mark;
+
 	if(start_pos >= view->list_rows)
+	{
 		start_pos = view->list_rows - 1;
+	}
+
 	upwards_range = view->list_pos < start_pos;
-	if(upwards_range)
-	{
-		set_specmark('<', view->curr_dir, get_current_file_name(view));
-		set_specmark('>', view->curr_dir, view->dir_entry[start_pos].name);
-	}
-	else
-	{
-		set_specmark('<', view->curr_dir, view->dir_entry[start_pos].name);
-		set_specmark('>', view->curr_dir, get_current_file_name(view));
-	}
+
+	start_mark = upwards_range ? '<' : '>';
+	end_mark = upwards_range ? '>' : '<';
+
+	set_spec_bookmark(start_mark, view->curr_dir, get_current_file_name(view));
+	set_spec_bookmark(end_mark, view->curr_dir, view->dir_entry[start_pos].name);
 }
 
 static void
