@@ -24,7 +24,7 @@
 #include <stddef.h> /* NULL */
 #include <stdio.h> /* snprintf() */
 #include <stdlib.h>
-#include <string.h> /* memcpy() memset() strstr() */
+#include <string.h> /* memcpy() memmove() strstr() */
 
 #include "cfg/config.h"
 #include "engine/options.h"
@@ -42,6 +42,7 @@
 #include "quickview.h"
 #include "sort.h"
 #include "status.h"
+#include "trash.h"
 #include "ui.h"
 #include "viewcolumns_parser.h"
 
@@ -73,12 +74,15 @@ static void init_timefmt(optval_t *val);
 static void init_trash_dir(optval_t *val);
 static void init_lsview(optval_t *val);
 static void init_shortmess(optval_t *val);
+static void init_number(optval_t *val);
+static void init_numberwidth(optval_t *val);
+static void init_relativenumber(optval_t *val);
 static void init_sort(optval_t *val);
 static void init_sortorder(optval_t *val);
 static void init_viewcolumns(optval_t *val);
 static void load_options_defaults(void);
 static void add_options(void);
-static void apropos_handler(OPT_OP op, optval_t val);
+static void aproposprg_handler(OPT_OP op, optval_t val);
 static void autochpos_handler(OPT_OP op, optval_t val);
 static void classify_handler(OPT_OP op, optval_t val);
 static int str_to_classify(const char str[],
@@ -94,7 +98,7 @@ static void findprg_handler(OPT_OP op, optval_t val);
 static void followlinks_handler(OPT_OP op, optval_t val);
 static void fusehome_handler(OPT_OP op, optval_t val);
 static void gdefault_handler(OPT_OP op, optval_t val);
-static void grep_handler(OPT_OP op, optval_t val);
+static void grepprg_handler(OPT_OP op, optval_t val);
 static void history_handler(OPT_OP op, optval_t val);
 static void hlsearch_handler(OPT_OP op, optval_t val);
 static void iec_handler(OPT_OP op, optval_t val);
@@ -102,7 +106,7 @@ static void ignorecase_handler(OPT_OP op, optval_t val);
 static void incsearch_handler(OPT_OP op, optval_t val);
 static void laststatus_handler(OPT_OP op, optval_t val);
 static void lines_handler(OPT_OP op, optval_t val);
-static void locate_handler(OPT_OP op, optval_t val);
+static void locateprg_handler(OPT_OP op, optval_t val);
 static void scroll_line_down(FileView *view);
 static void rulerformat_handler(OPT_OP op, optval_t val);
 static void runexec_handler(OPT_OP op, optval_t val);
@@ -116,6 +120,10 @@ static void slowfs_handler(OPT_OP op, optval_t val);
 static void smartcase_handler(OPT_OP op, optval_t val);
 static void sortnumbers_handler(OPT_OP op, optval_t val);
 static void lsview_handler(OPT_OP op, optval_t val);
+static void number_handler(OPT_OP op, optval_t val);
+static void numberwidth_handler(OPT_OP op, optval_t val);
+static void relativenumber_handler(OPT_OP op, optval_t val);
+static void update_num_type(NumberingType num_type, int enable);
 static void sort_handler(OPT_OP op, optval_t val);
 static void sortorder_handler(OPT_OP op, optval_t val);
 static void viewcolumns_handler(OPT_OP op, optval_t val);
@@ -156,6 +164,7 @@ static const char * sort_enum[] = {
 #ifndef _WIN32
 	"perms",
 #endif
+	"type",
 };
 ARRAY_GUARD(sort_enum, SORT_OPTION_COUNT);
 
@@ -191,6 +200,7 @@ static const char * sort_types[] = {
 #ifndef _WIN32
 	"perms", "+perms", "-perms",
 #endif
+	"type", "+type", "-type",
 };
 ARRAY_GUARD(sort_types, SORT_OPTION_COUNT*3);
 
@@ -227,107 +237,219 @@ static struct
 	opt_handler handler;
 	optinit_t initializer;
 	optval_t val;
-}options[] = {
+}
+options[] =
+{
 	/* global options */
-	{ "aproposprg",   "",    OPT_STR,     0,                          NULL,            &apropos_handler,
-		{ .ref.str_val = &cfg.apropos_prg }                                                                   },
-	{ "autochpos",   "",     OPT_BOOL,    0,                          NULL,            &autochpos_handler,
-		{ .ref.bool_val = &cfg.auto_ch_pos }                                                                   },
-	{ "classify",    "",     OPT_STRLIST, 0,                          NULL,            &classify_handler,
-		{ .init = &init_classify }                                                                             },
-	{ "columns",     "co",   OPT_INT,     0,                          NULL,            &columns_handler,
-		{ .ref.int_val = &cfg.columns }                                                                        },
-	{ "confirm",     "cf",   OPT_BOOL,    0,                          NULL,            &confirm_handler,
-		{ .ref.bool_val = &cfg.confirm }                                                                       },
-	{ "cpoptions",   "cpo",  OPT_CHARSET, cpoptions_count,            &cpoptions_vals, &cpoptions_handler,
-		{ .init = &init_cpoptions }                                                                            },
-	{ "dotdirs",     "",     OPT_SET,     ARRAY_LEN(dotdirs_vals),    dotdirs_vals,    &dotdirs_handler,
-		{ .ref.set_items = &cfg.dot_dirs }                                                                     },
-	{ "fastrun",     "",     OPT_BOOL,    0,                          NULL,            &fastrun_handler,
-		{ .ref.bool_val = &cfg.fast_run }                                                                      },
-	{ "findprg",     "",     OPT_STR,     0,                          NULL,            &findprg_handler,
-		{ .ref.str_val = &cfg.find_prg }                                                                       },
-	{ "followlinks", "",     OPT_BOOL,    0,                          NULL,            &followlinks_handler,
-		{ .ref.bool_val = &cfg.follow_links }                                                                  },
-	{ "fusehome",    "",     OPT_STR,     0,                          NULL,            &fusehome_handler,
-		{ .ref.str_val = &cfg.fuse_home }                                                                      },
-	{ "gdefault",    "gd",   OPT_BOOL,    0,                          NULL,            &gdefault_handler,
-		{ .ref.bool_val = &cfg.gdefault }                                                                      },
-	{ "grepprg",     "",     OPT_STR,     0,                          NULL,            &grep_handler,
-		{ .ref.str_val = &cfg.grep_prg }                                                                       },
-	{ "history",     "hi",   OPT_INT,     0,                          NULL,            &history_handler,
-		{ .ref.int_val = &cfg.history_len }                                                                    },
-	{ "hlsearch",    "hls",  OPT_BOOL,    0,                          NULL,            &hlsearch_handler,
-		{ .ref.bool_val = &cfg.hl_search }                                                                     },
-	{ "iec",         "",     OPT_BOOL,    0,                          NULL,            &iec_handler,
-		{ .ref.bool_val = &cfg.use_iec_prefixes }                                                              },
-	{ "ignorecase",  "ic",   OPT_BOOL,    0,                          NULL,            &ignorecase_handler,
-		{ .ref.bool_val = &cfg.ignore_case }                                                                   },
-	{ "incsearch",   "is",   OPT_BOOL,    0,                          NULL,            &incsearch_handler ,
-		{ .ref.bool_val = &cfg.inc_search }                                                                    },
-	{ "laststatus",  "ls",   OPT_BOOL,    0,                          NULL,            &laststatus_handler,
-		{ .ref.bool_val = &cfg.last_status }                                                                   },
-	{ "lines",       "",     OPT_INT,     0,                          NULL,            &lines_handler,
-		{ .ref.int_val = &cfg.lines }                                                                          },
-	{ "locateprg",   "",     OPT_STR,     0,                          NULL,            &locate_handler,
-		{ .ref.str_val = &cfg.locate_prg }                                                                     },
-	{ "rulerformat", "ruf",  OPT_STR,     0,                          NULL,            &rulerformat_handler,
-		{ .ref.str_val = &cfg.ruler_format }                                                                   },
-	{ "runexec",     "",     OPT_BOOL,    0,                          NULL,            &runexec_handler,
-		{ .ref.bool_val = &cfg.auto_execute }                                                                  },
-	{ "scrollbind",  "scb",  OPT_BOOL,    0,                          NULL,            &scrollbind_handler,
-		{ .ref.bool_val = &cfg.scroll_bind }                                                                   },
-	{ "scrolloff",   "so",   OPT_INT,     0,                          NULL,            &scrolloff_handler,
-		{ .ref.int_val = &cfg.scroll_off }                                                                     },
-	{ "shell",       "sh",   OPT_STR,     0,                          NULL,            &shell_handler,
-		{ .ref.str_val = &cfg.shell }                                                                          },
-	{ "shortmess",   "shm",  OPT_CHARSET, shortmess_count,            &shortmess_vals, &shortmess_handler,
-		{ .init = &init_shortmess }                                                                            },
+	{ "aproposprg", "",
+	  OPT_STR, 0, NULL, &aproposprg_handler,
+	  { .ref.str_val = &cfg.apropos_prg },
+	},
+	{ "autochpos", "",
+	  OPT_BOOL, 0, NULL, &autochpos_handler,
+	  { .ref.bool_val = &cfg.auto_ch_pos },
+	},
+	{ "classify", "",
+	  OPT_STRLIST, 0, NULL, &classify_handler,
+	  { .init = &init_classify },
+	},
+	{ "columns", "co",
+	  OPT_INT, 0, NULL, &columns_handler,
+	  { .ref.int_val = &cfg.columns },
+	},
+	{ "confirm", "cf",
+	  OPT_BOOL, 0, NULL, &confirm_handler,
+	  { .ref.bool_val = &cfg.confirm },
+	},
+	{ "cpoptions", "cpo",
+	  OPT_CHARSET, cpoptions_count, &cpoptions_vals, &cpoptions_handler,
+	  { .init = &init_cpoptions },
+	},
+	{ "dotdirs", "",
+	  OPT_SET, ARRAY_LEN(dotdirs_vals), dotdirs_vals, &dotdirs_handler,
+	  { .ref.set_items = &cfg.dot_dirs },
+	},
+	{ "fastrun", "",
+	  OPT_BOOL, 0, NULL, &fastrun_handler,
+	  { .ref.bool_val = &cfg.fast_run },
+	},
+	{ "findprg", "",
+	  OPT_STR, 0, NULL, &findprg_handler,
+	  { .ref.str_val = &cfg.find_prg },
+	},
+	{ "followlinks", "",
+	  OPT_BOOL, 0, NULL, &followlinks_handler,
+	  { .ref.bool_val = &cfg.follow_links },
+	},
+	{ "fusehome", "",
+	  OPT_STR, 0, NULL, &fusehome_handler,
+	  { .ref.str_val = &cfg.fuse_home },
+	},
+	{ "gdefault", "gd",
+	  OPT_BOOL, 0, NULL, &gdefault_handler,
+	  { .ref.bool_val = &cfg.gdefault },
+	},
+	{ "grepprg", "",
+	  OPT_STR, 0, NULL, &grepprg_handler,
+	  { .ref.str_val = &cfg.grep_prg },
+	},
+	{ "history", "hi",
+	  OPT_INT, 0, NULL, &history_handler,
+	  { .ref.int_val = &cfg.history_len },
+	},
+	{ "hlsearch", "hls",
+	  OPT_BOOL, 0, NULL, &hlsearch_handler,
+	  { .ref.bool_val = &cfg.hl_search },
+	},
+	{ "iec", "",
+	  OPT_BOOL, 0, NULL, &iec_handler,
+	  { .ref.bool_val = &cfg.use_iec_prefixes },
+	},
+	{ "ignorecase", "ic",
+	  OPT_BOOL, 0, NULL, &ignorecase_handler,
+	  { .ref.bool_val = &cfg.ignore_case },
+	},
+	{ "incsearch", "is",
+	  OPT_BOOL, 0, NULL, &incsearch_handler ,
+	  { .ref.bool_val = &cfg.inc_search },
+	},
+	{ "laststatus", "ls",
+	  OPT_BOOL, 0, NULL, &laststatus_handler,
+	  { .ref.bool_val = &cfg.last_status },
+	},
+	{ "lines", "",
+	  OPT_INT, 0, NULL, &lines_handler,
+	  { .ref.int_val = &cfg.lines },
+	},
+	{ "locateprg", "",
+	  OPT_STR, 0, NULL, &locateprg_handler,
+	  { .ref.str_val = &cfg.locate_prg },
+	},
+	{ "rulerformat", "ruf",
+	  OPT_STR, 0, NULL, &rulerformat_handler,
+	  { .ref.str_val = &cfg.ruler_format },
+	},
+	{ "runexec", "",
+	  OPT_BOOL, 0, NULL, &runexec_handler,
+	  { .ref.bool_val = &cfg.auto_execute },
+	},
+	{ "scrollbind", "scb",
+	  OPT_BOOL, 0, NULL, &scrollbind_handler,
+	  { .ref.bool_val = &cfg.scroll_bind },
+	},
+	{ "scrolloff", "so",
+	  OPT_INT, 0, NULL, &scrolloff_handler,
+	  { .ref.int_val = &cfg.scroll_off },
+	},
+	{ "shell", "sh",
+	  OPT_STR, 0, NULL, &shell_handler,
+	  { .ref.str_val = &cfg.shell },
+	},
+	{ "shortmess", "shm",
+	  OPT_CHARSET, shortmess_count, &shortmess_vals, &shortmess_handler,
+	  { .init = &init_shortmess },
+	},
 #ifndef _WIN32
-	{ "slowfs",      "",     OPT_STRLIST, 0,                          NULL,            &slowfs_handler,
-		{ .ref.str_val = &cfg.slow_fs_list }                                                                   },
+	{ "slowfs", "",
+	  OPT_STRLIST, 0, NULL, &slowfs_handler,
+	  { .ref.str_val = &cfg.slow_fs_list },
+	},
 #endif
-	{ "smartcase",   "scs",  OPT_BOOL,    0,                          NULL,            &smartcase_handler,
-		{ .ref.bool_val = &cfg.smart_case }                                                                    },
-	{ "sortnumbers", "",     OPT_BOOL,    0,                          NULL,            &sortnumbers_handler,
-		{ .ref.bool_val = &cfg.sort_numbers }                                                                  },
-	{ "statusline",  "stl",  OPT_STR,     0,                          NULL,            &statusline_handler,
-		{ .ref.str_val = &cfg.status_line }                                                                    },
-	{ "tabstop",     "ts",   OPT_INT,     0,                          NULL,            &tabstop_handler,
-		{ .ref.int_val = &cfg.tab_stop }                                                                       },
-	{ "timefmt",     "",     OPT_STR,     0,                          NULL,            &timefmt_handler,
-		{ .init = &init_timefmt }                                                                              },
-	{ "timeoutlen",  "tm",   OPT_INT,     0,                          NULL,            &timeoutlen_handler,
-		{ .ref.int_val = &cfg.timeout_len }                                                                    },
-	{ "trash",       "",     OPT_BOOL,    0,                          NULL,            &trash_handler,
-		{ .ref.bool_val = &cfg.use_trash }                                                                     },
-	{ "trashdir",    "",     OPT_STR,     0,                          NULL,            &trashdir_handler,
-		{ .init = &init_trash_dir }                                                                            },
-	{ "undolevels",  "ul",   OPT_INT,     0,                          NULL,            &undolevels_handler,
-		{ .ref.int_val = &cfg.undo_levels }                                                                    },
-	{ "vicmd",       "",     OPT_STR,     0,                          NULL,            &vicmd_handler,
-		{ .ref.str_val = &cfg.vi_command }                                                                     },
-	{ "vixcmd",      "",     OPT_STR,     0,                          NULL,            &vixcmd_handler,
-		{ .ref.str_val = &cfg.vi_x_command }                                                                   },
-	{ "vifminfo",    "",     OPT_SET,     ARRAY_LEN(vifminfo_set),    vifminfo_set,    &vifminfo_handler,
-		{ .ref.set_items = &cfg.vifm_info }                                                                    },
-	{ "vimhelp",     "",     OPT_BOOL,    0,                          NULL,            &vimhelp_handler,
-		{ .ref.bool_val = &cfg.use_vim_help }                                                                  },
-	{ "wildmenu",    "wmnu", OPT_BOOL,    0,                          NULL,            &wildmenu_handler,
-		{ .ref.bool_val = &cfg.wild_menu }                                                                     },
-	{ "wrap",        "",     OPT_BOOL,    0,                          NULL,            &wrap_handler,
-		{ .ref.bool_val = &cfg.wrap_quick_view }                                                               },
-	{ "wrapscan",    "ws",   OPT_BOOL,    0,                          NULL,            &wrapscan_handler,
-		{ .ref.bool_val = &cfg.wrap_scan }                                                                     },
-	/* local options */
-	{ "lsview",      "",     OPT_BOOL,    0,                          NULL,            &lsview_handler,
-		{ .init = &init_lsview }                                                                               },
-	{ "sort",        "",     OPT_STRLIST, ARRAY_LEN(sort_types),      sort_types,      &sort_handler,
-		{ .init = &init_sort }                                                                                 },
-	{ "sortorder",   "",     OPT_ENUM,    ARRAY_LEN(sortorder_enum),  sortorder_enum,  &sortorder_handler,
-		{ .init = &init_sortorder }                                                                            },
-	{ "viewcolumns", "",     OPT_STRLIST, 0,                          NULL,            &viewcolumns_handler,
-		{ .init = &init_viewcolumns }                                                                          },
+	{ "smartcase", "scs",
+	  OPT_BOOL, 0, NULL, &smartcase_handler,
+	  { .ref.bool_val = &cfg.smart_case },
+	},
+	{ "sortnumbers", "",
+	  OPT_BOOL, 0, NULL, &sortnumbers_handler,
+	  { .ref.bool_val = &cfg.sort_numbers },
+	},
+	{ "statusline", "stl",
+	  OPT_STR, 0, NULL, &statusline_handler,
+	  { .ref.str_val = &cfg.status_line },
+	},
+	{ "tabstop", "ts",
+	  OPT_INT, 0, NULL, &tabstop_handler,
+	  { .ref.int_val = &cfg.tab_stop },
+	},
+	{ "timefmt", "",
+	  OPT_STR, 0, NULL, &timefmt_handler,
+	  { .init = &init_timefmt },
+	},
+	{ "timeoutlen", "tm",
+	  OPT_INT, 0, NULL, &timeoutlen_handler,
+	  { .ref.int_val = &cfg.timeout_len },
+	},
+	{ "trash", "",
+	  OPT_BOOL, 0, NULL, &trash_handler,
+	  { .ref.bool_val = &cfg.use_trash },
+	},
+	{ "trashdir", "",
+	  OPT_STRLIST, 0, NULL, &trashdir_handler,
+	  { .init = &init_trash_dir },
+	},
+	{ "undolevels", "ul",
+	  OPT_INT, 0, NULL, &undolevels_handler,
+	  { .ref.int_val = &cfg.undo_levels },
+	},
+	{ "vicmd", "",
+	  OPT_STR, 0, NULL, &vicmd_handler,
+	  { .ref.str_val = &cfg.vi_command },
+	},
+	{ "vixcmd", "",
+	  OPT_STR, 0, NULL, &vixcmd_handler,
+	  { .ref.str_val = &cfg.vi_x_command },
+	},
+	{ "vifminfo", "",
+	  OPT_SET, ARRAY_LEN(vifminfo_set), vifminfo_set, &vifminfo_handler,
+	  { .ref.set_items = &cfg.vifm_info },
+	},
+	{ "vimhelp", "",
+	  OPT_BOOL, 0, NULL, &vimhelp_handler,
+	  { .ref.bool_val = &cfg.use_vim_help },
+	},
+	{ "wildmenu", "wmnu",
+	  OPT_BOOL, 0, NULL, &wildmenu_handler,
+	  { .ref.bool_val = &cfg.wild_menu },
+	},
+	{ "wrap", "",
+	  OPT_BOOL, 0, NULL, &wrap_handler,
+	  { .ref.bool_val = &cfg.wrap_quick_view },
+	},
+	{ "wrapscan", "ws",
+	  OPT_BOOL, 0, NULL, &wrapscan_handler,
+	  { .ref.bool_val = &cfg.wrap_scan },
+	},
+
+	/* Local options. */
+
+	{ "lsview", "",
+	  OPT_BOOL, 0, NULL, &lsview_handler,
+	  { .init = &init_lsview },
+	},
+	{ "number", "nu",
+	  OPT_BOOL, 0, NULL, &number_handler,
+	  { .init = &init_number },
+	},
+	{ "numberwidth", "nuw",
+	  OPT_INT, 0, NULL, &numberwidth_handler,
+	  { .init = &init_numberwidth },
+	},
+	{ "relativenumber", "rnu",
+	  OPT_BOOL, 0, NULL, &relativenumber_handler,
+	  { .init = &init_relativenumber },
+	},
+	{ "sort", "",
+	  OPT_STRLIST, ARRAY_LEN(sort_types), sort_types, &sort_handler,
+	  { .init = &init_sort },
+	},
+	{ "sortorder", "",
+	  OPT_ENUM, ARRAY_LEN(sortorder_enum), sortorder_enum, &sortorder_handler,
+	  { .init = &init_sortorder },
+	},
+	{ "viewcolumns", "",
+	  OPT_STRLIST, 0, NULL, &viewcolumns_handler,
+	  { .init = &init_viewcolumns },
+	},
 };
 
 static int error;
@@ -410,6 +532,27 @@ init_shortmess(optval_t *val)
 	val->str_val = buf;
 }
 
+/* Default-initializes whether to display file numbers. */
+static void
+init_number(optval_t *val)
+{
+	val->bool_val = 0;
+}
+
+/* Default-initializes minimum width of file number field. */
+static void
+init_numberwidth(optval_t *val)
+{
+	val->int_val = 4;
+}
+
+/* Default-initializes whether to display relative file numbers. */
+static void
+init_relativenumber(optval_t *val)
+{
+	val->bool_val = 0;
+}
+
 static void
 init_sort(optval_t *val)
 {
@@ -469,6 +612,15 @@ load_local_options(FileView *view)
 
 	val.bool_val = view->ls_view;
 	set_option("lsview", val);
+
+	val.bool_val = view->num_type & NT_SEQ;
+	set_option("number", val);
+
+	val.int_val = view->num_width;
+	set_option("numberwidth", val);
+
+	val.int_val = view->num_type & NT_REL;
+	set_option("relativenumber", val);
 }
 
 void
@@ -478,25 +630,13 @@ load_sort_option(FileView *view)
 	 * comma (",") between items. */
 	enum { MAX_SORT_OPTION_NAME_LEN = 16 };
 
+	int i;
+
 	optval_t val;
 	char opt_val[MAX_SORT_OPTION_NAME_LEN*SORT_OPTION_COUNT] = "";
 	size_t opt_val_len = 0U;
-	int j, i;
 
-	/* Ensure that list of sorting keys contains either "name" or "iname". */
-	j = -1;
-	while(++j < SORT_OPTION_COUNT && abs(view->sort[j]) <= LAST_SORT_OPTION)
-	{
-		const int sort_option = abs(view->sort[j]);
-		if(sort_option == SORT_BY_NAME || sort_option == SORT_BY_INAME)
-		{
-			break;
-		}
-	}
-	if(j < SORT_OPTION_COUNT && abs(view->sort[j]) > LAST_SORT_OPTION)
-	{
-		view->sort[j++] = DEFAULT_SORT_KEY;
-	}
+	ui_view_sort_list_ensure_well_formed(view->sort);
 
 	/* Produce a string, which represents a list of sorting keys. */
 	i = -1;
@@ -547,7 +687,7 @@ process_set_args(const char *args)
 }
 
 static void
-apropos_handler(OPT_OP op, optval_t val)
+aproposprg_handler(OPT_OP op, optval_t val)
 {
 	(void)replace_string(&cfg.apropos_prg, val.str_val);
 }
@@ -772,7 +912,7 @@ gdefault_handler(OPT_OP op, optval_t val)
 }
 
 static void
-grep_handler(OPT_OP op, optval_t val)
+grepprg_handler(OPT_OP op, optval_t val)
 {
 	(void)replace_string(&cfg.grep_prg, val.str_val);
 }
@@ -880,7 +1020,7 @@ lines_handler(OPT_OP op, optval_t val)
 }
 
 static void
-locate_handler(OPT_OP op, optval_t val)
+locateprg_handler(OPT_OP op, optval_t val)
 {
 	(void)replace_string(&cfg.locate_prg, val.str_val);
 }
@@ -936,7 +1076,7 @@ scrolloff_handler(OPT_OP op, optval_t val)
 static void
 shell_handler(OPT_OP op, optval_t val)
 {
-	(void)replace_string(&cfg.shell, val.str_val);
+	cfg_set_shell(val.str_val);
 }
 
 static void
@@ -1003,11 +1143,54 @@ lsview_handler(OPT_OP op, optval_t val)
 	}
 }
 
+/* Handles file numbers displaying toggle. */
+static void
+number_handler(OPT_OP op, optval_t val)
+{
+	update_num_type(NT_SEQ, val.bool_val);
+}
+
+/* Handles changes of minimum width of file number field. */
+static void
+numberwidth_handler(OPT_OP op, optval_t val)
+{
+	curr_view->num_width = val.int_val;
+
+	if(ui_view_displays_numbers(curr_view))
+	{
+		redraw_current_view();
+	}
+}
+
+/* Handles relative file numbers displaying toggle. */
+static void
+relativenumber_handler(OPT_OP op, optval_t val)
+{
+	update_num_type(NT_REL, val.bool_val);
+}
+
+/* Handles toggling of boolean number related option and updates current view if
+ * needed. */
+static void
+update_num_type(NumberingType num_type, int enable)
+{
+	const int old_num_type = curr_view->num_type;
+
+	curr_view->num_type = enable
+	                    ? (old_num_type | num_type)
+	                    : (old_num_type & ~num_type);
+
+	if(curr_view->num_type != old_num_type)
+	{
+		redraw_current_view();
+	}
+}
+
 static void
 sort_handler(OPT_OP op, optval_t val)
 {
 	char *p = val.str_val - 1;
-	int i, j;
+	int i;
 
 	i = 0;
 	do
@@ -1016,6 +1199,7 @@ sort_handler(OPT_OP op, optval_t val)
 		char *t;
 		int minus;
 		int pos;
+		int j;
 
 		t = p + 1;
 		p = strchr(t, ',');
@@ -1046,17 +1230,11 @@ sort_handler(OPT_OP op, optval_t val)
 	}
 	while(*p != '\0');
 
-	for(j = 0; j < i; j++)
+	if(i < SORT_OPTION_COUNT)
 	{
-		int sort_key = abs(curr_view->sort[j]);
-		if(sort_key == SORT_BY_NAME || sort_key == SORT_BY_INAME)
-			break;
+		curr_view->sort[i] = NO_SORT_OPTION;
 	}
-	if(j == i)
-	{
-		curr_view->sort[i++] = DEFAULT_SORT_KEY;
-	}
-	memset(&curr_view->sort[i], NO_SORT_OPTION, sizeof(curr_view->sort) - i);
+	ui_view_sort_list_ensure_well_formed(curr_view->sort);
 
 	reset_view_sort(curr_view);
 	resort_view(curr_view);
