@@ -71,7 +71,7 @@ static void run_win_executable(char full_path[]);
 static int run_win_executable_as_evaluated(const char full_path[]);
 #endif
 static int selection_is_consistent(const FileView *const view);
-static void execute_file(char full_path[]);
+static void execute_file(const char full_path[]);
 static void run_selection(FileView *view, int dont_execute);
 static void run_file(FileView *view, int dont_execute);
 static int multi_run_compat(FileView *view, const char *program);
@@ -297,15 +297,19 @@ selection_is_consistent(const FileView *const view)
 /* Executes file, specified by the full_path.  Changes type of slashes on
  * Windows. */
 static void
-execute_file(char full_path[])
+execute_file(const char full_path[])
 {
 #ifndef _WIN32
 	char *const escaped = escape_filename(full_path, 0);
 	shellout(escaped, 1, 1);
 	free(escaped);
 #else
-	to_back_slash(full_path);
-	run_win_executable(full_path);
+	char *const dquoted_full_path = strdup(enclose_in_dquotes(full_path));
+
+	to_back_slash(dquoted_full_path);
+	run_win_executable(dquoted_full_path);
+
+	free(dquoted_full_path);
 #endif
 }
 
@@ -458,11 +462,11 @@ multi_run_compat(FileView *view, const char *program)
 }
 
 int
-view_file(const char filename[], int line, int column, int do_fork)
+view_file(const char filename[], int line, int column, int allow_forking)
 {
 	char vicmd[PATH_MAX];
 	char command[PATH_MAX + 5] = "";
-	const char *fork_str = do_fork ? "" : "--nofork";
+	const char *fork_str = allow_forking ? "" : "--nofork";
 	char *escaped;
 	int bg;
 	int result;
@@ -488,7 +492,7 @@ view_file(const char filename[], int line, int column, int do_fork)
 
 	snprintf(vicmd, sizeof(vicmd), "%s", get_vicmd(&bg));
 	(void)trim_right(vicmd);
-	if(!do_fork)
+	if(!allow_forking)
 	{
 		char *p = strrchr(vicmd, ' ');
 		if(p != NULL && strstr(p, "remote"))
@@ -510,10 +514,14 @@ view_file(const char filename[], int line, int column, int do_fork)
 	free(escaped);
 #endif
 
-	if(bg && do_fork)
+	if(bg && allow_forking)
+	{
 		result = start_background_job(command, 0);
+	}
 	else
-		result = shellout(command, -1, 1);
+	{
+		result = shellout(command, -1, allow_forking);
+	}
 	curs_set(FALSE);
 
 	return result;
@@ -590,18 +598,33 @@ run_using_prog(FileView *view, const char *program, int dont_execute,
 		else if(flags == MACRO_IGNORE)
 			output_to_nowhere(command);
 		else
-			shellout(command, pause ? 1 : -1, 1);
+			shellout(command, pause ? 1 : -1, flags != MACRO_NO_TERM_MUX);
 
 		free(command);
 	}
 	else
 	{
 		char buf[NAME_MAX + 1 + NAME_MAX + 1];
-		char *temp = escape_filename(view->dir_entry[view->list_pos].name, 0);
+		const char *name_macro;
+		char *file_name;
 
-		snprintf(buf, sizeof(buf), "%s %s", program, temp);
+#ifdef _WIN32
+		if(curr_stats.shell_type == ST_CMD)
+		{
+			name_macro = (view == curr_view) ? "%\"c" : "%\"C";
+		}
+		else
+#endif
+		{
+			name_macro = (view == curr_view) ? "%c" : "%C";
+		}
+
+		file_name = expand_macros(name_macro, NULL, NULL, 1);
+
+		snprintf(buf, sizeof(buf), "%s %s", program, file_name);
 		shellout(buf, pause ? 1 : -1, 1);
-		free(temp);
+
+		free(file_name);
 	}
 }
 
@@ -817,24 +840,33 @@ store_for_external(const FileView *view, FILE *fp, int argc, char *argv[])
  *  < 0 - pause on error
  */
 int
-shellout(const char *command, int pause, int allow_screen)
+shellout(const char *command, int pause, int use_term_multiplexer)
 {
-	char buf[cfg.max_args];
+	/* cfg.max_args can be an enormous huge number and we want to set an upper
+	 * bound for the length of command line buffer.  Picket upper bound based on
+	 * command-line formatting done by gen_shell_cmd() and still should left some
+	 * space unused just for safety reasons (to take possible future changes into
+	 * account).  4096 here stands for minimum value of _POSIX_ARG_MAX parameter,
+	 * which is not used directly for portability reasons. */
+	const size_t command_len = (command != NULL) ? strlen(command) : 0UL;
+	const size_t max_composed_cmd_len = (4096 + command_len)*4;
+	char buf[MIN(cfg.max_args, max_composed_cmd_len)];
+
 	int result;
 	int ec;
 
 	if(pause > 0 && command != NULL && ends_with(command, "&"))
 		pause = -1;
 
-	gen_shell_cmd(command, pause > 0, allow_screen, sizeof(buf), buf);
+	gen_shell_cmd(command, pause > 0, use_term_multiplexer, sizeof(buf), buf);
 
 	endwin();
 
 	/* Need to use setenv instead of getcwd for a symlink directory */
 	env_set("PWD", curr_view->curr_dir);
 
-	ec = my_system(buf);
-	/* No WIFEXITED(ec) check here, since my_system(...) shouldn't return until
+	ec = vifm_system(buf);
+	/* No WIFEXITED(ec) check here, since vifm_system(...) shouldn't return until
 	 * subprocess exited. */
 	result = WEXITSTATUS(ec);
 
@@ -851,9 +883,11 @@ shellout(const char *command, int pause, int allow_screen)
 
 	recover_after_shellout();
 
-	/* always redraw to handle resizing of terminal */
-	if(!curr_stats.auto_redraws)
+	if(!curr_stats.skip_shellout_redraw)
+	{
+		/* Redraw to handle resizing of terminal that we could have missed. */
 		curr_stats.need_update = UT_FULL;
+	}
 
 	curs_set(FALSE);
 
@@ -935,7 +969,7 @@ gen_term_multiplexer_cmd(const char cmd[], int pause, size_t shell_cmd_len,
 
 		/* Needed for symlink directories and sshfs mounts. */
 		snprintf(shell_cmd, shell_cmd_len, "screen -X setenv PWD %s", escaped_dir);
-		my_system(shell_cmd);
+		(void)vifm_system(shell_cmd);
 
 		snprintf(shell_cmd, shell_cmd_len, "screen %s %s -c %s", title_arg_buffer,
 				escaped_sh, escaped);
@@ -994,12 +1028,20 @@ gen_normal_cmd(const char cmd[], int pause, size_t shell_cmd_len,
 {
 	if(pause)
 	{
+		const char *cmd_with_pause_fmt;
+
 #ifdef _WIN32
-		if(stroscmp(cfg.shell, "cmd") == 0)
-			snprintf(shell_cmd, shell_cmd_len, "%s" PAUSE_STR, cmd);
+		if(curr_stats.shell_type == ST_CMD)
+		{
+			cmd_with_pause_fmt = "%s" PAUSE_STR;
+		}
 		else
 #endif
-			snprintf(shell_cmd, shell_cmd_len, "%s; " PAUSE_CMD, cmd);
+		{
+			cmd_with_pause_fmt = "%s; " PAUSE_CMD;
+		}
+
+		snprintf(shell_cmd, shell_cmd_len, cmd_with_pause_fmt, cmd);
 	}
 	else
 	{
@@ -1018,7 +1060,7 @@ gen_term_multiplexer_run_cmd(size_t shell_cmd_len, char shell_cmd[])
 	{
 		/* Needed for symlink directories and sshfs mounts. */
 		snprintf(shell_cmd, shell_cmd_len, "screen -X setenv PWD %s", escaped_dir);
-		my_system(shell_cmd);
+		(void)vifm_system(shell_cmd);
 
 		snprintf(shell_cmd, shell_cmd_len, "screen");
 	}
@@ -1039,7 +1081,7 @@ output_to_nowhere(const char *cmd)
 {
 	FILE *file, *err;
 
-	if(background_and_capture((char *)cmd, &file, &err) != 0)
+	if(background_and_capture((char *)cmd, &file, &err) == (pid_t)-1)
 	{
 		show_error_msgf("Trouble running command", "Unable to run: %s", cmd);
 		return;
