@@ -99,6 +99,9 @@ handle_signal(int sig)
 {
 	switch(sig)
 	{
+		case SIGINT:
+			ui_cancellation_request();
+			break;
 		case SIGCHLD:
 			received_sigchld();
 			break;
@@ -118,6 +121,7 @@ handle_signal(int sig)
 			break;
 	}
 }
+
 #else
 BOOL WINAPI
 ctrl_handler(DWORD dwCtrlType)
@@ -128,6 +132,7 @@ ctrl_handler(DWORD dwCtrlType)
 	{
 		case CTRL_C_EVENT:
 		case CTRL_BREAK_EVENT:
+			ui_cancellation_request();
 			break;
 		case CTRL_CLOSE_EVENT:
 		case CTRL_LOGOFF_EVENT:
@@ -167,6 +172,7 @@ setup_signals(void)
 
 	sigaction(SIGCHLD, &handle_signal_action, NULL);
 	sigaction(SIGHUP, &handle_signal_action, NULL);
+	sigaction(SIGINT, &handle_signal_action, NULL);
 	sigaction(SIGQUIT, &handle_signal_action, NULL);
 	sigaction(SIGCONT, &handle_signal_action, NULL);
 	sigaction(SIGTERM, &handle_signal_action, NULL);
@@ -180,9 +186,8 @@ setup_signals(void)
 	{
 		LOG_WERROR(GetLastError());
 	}
-#endif
-
 	signal(SIGINT, SIG_IGN);
+#endif
 }
 
 /* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab cinoptions-=(0 : */
