@@ -17,6 +17,7 @@ const char *tags[] = {
 	"vifm-%f",
 	"vifm-%i",
 	"vifm-%m",
+	"vifm-%n",
 	"vifm-%r",
 	"vifm-%s",
 	"vifm-'",
@@ -50,6 +51,12 @@ const char *tags[] = {
 	"vifm-'locateprg'",
 	"vifm-'ls'",
 	"vifm-'lsview'",
+	"vifm-'nu'",
+	"vifm-'number'",
+	"vifm-'numberwidth'",
+	"vifm-'nuw'",
+	"vifm-'relativenumber'",
+	"vifm-'rnu'",
 	"vifm-'ruf'",
 	"vifm-'rulerformat'",
 	"vifm-'runexec'",
@@ -180,6 +187,7 @@ const char *tags[] = {
 	"vifm-:let",
 	"vifm-:locate",
 	"vifm-:ls",
+	"vifm-:lstrash",
 	"vifm-:m",
 	"vifm-:ma",
 	"vifm-:map",
@@ -244,6 +252,7 @@ const char *tags[] = {
 	"vifm-:sync",
 	"vifm-:touch",
 	"vifm-:tr",
+	"vifm-:trashes",
 	"vifm-:undol",
 	"vifm-:undolist",
 	"vifm-:unl",
@@ -279,11 +288,13 @@ const char *tags[] = {
 	"vifm-C",
 	"vifm-CTRL-A",
 	"vifm-CTRL-B",
+	"vifm-CTRL-C",
 	"vifm-CTRL-D",
 	"vifm-CTRL-E",
 	"vifm-CTRL-F",
 	"vifm-CTRL-G",
 	"vifm-CTRL-I",
+	"vifm-CTRL-L",
 	"vifm-CTRL-N",
 	"vifm-CTRL-O",
 	"vifm-CTRL-P",
@@ -311,11 +322,13 @@ const char *tags[] = {
 	"vifm-CTRL-W_v",
 	"vifm-CTRL-W_w",
 	"vifm-CTRL-W_x",
+	"vifm-CTRL-W_z",
 	"vifm-CTRL-X",
 	"vifm-CTRL-Y",
 	"vifm-D",
 	"vifm-DD",
 	"vifm-Enter",
+	"vifm-Escape",
 	"vifm-F",
 	"vifm-FUSE_MOUNT",
 	"vifm-FUSE_MOUNT2",
@@ -359,6 +372,19 @@ const char *tags[] = {
 	"vifm-c_CTRL-T",
 	"vifm-c_CTRL-U",
 	"vifm-c_CTRL-W",
+	"vifm-c_CTRL-X_=",
+	"vifm-c_CTRL-X_CTRL-X_c",
+	"vifm-c_CTRL-X_CTRL-X_d",
+	"vifm-c_CTRL-X_CTRL-X_e",
+	"vifm-c_CTRL-X_CTRL-X_r",
+	"vifm-c_CTRL-X_CTRL-X_t",
+	"vifm-c_CTRL-X_a",
+	"vifm-c_CTRL-X_c",
+	"vifm-c_CTRL-X_d",
+	"vifm-c_CTRL-X_e",
+	"vifm-c_CTRL-X_m",
+	"vifm-c_CTRL-X_r",
+	"vifm-c_CTRL-X_t",
 	"vifm-c_CTRL-_",
 	"vifm-c_Delete",
 	"vifm-c_Down",
@@ -371,6 +397,7 @@ const char *tags[] = {
 	"vifm-c_SHIFT-Tab",
 	"vifm-c_Tab",
 	"vifm-c_Up",
+	"vifm-cancellation",
 	"vifm-cg",
 	"vifm-cl",
 	"vifm-clientserver",
@@ -381,6 +408,7 @@ const char *tags[] = {
 	"vifm-command-line",
 	"vifm-command-line-edit",
 	"vifm-commands",
+	"vifm-commands-and-selection",
 	"vifm-commands-bg",
 	"vifm-configure",
 	"vifm-count",
@@ -396,14 +424,23 @@ const char *tags[] = {
 	"vifm-expr-!=",
 	"vifm-expr-'",
 	"vifm-expr-.",
+	"vifm-expr-<",
+	"vifm-expr-<=",
 	"vifm-expr-==",
+	"vifm-expr->",
+	"vifm-expr->=",
 	"vifm-expr-env",
 	"vifm-expr-function",
+	"vifm-expr-number",
+	"vifm-expr-option",
 	"vifm-expr-quote",
 	"vifm-expr-string",
+	"vifm-expr-unary-+",
+	"vifm-expr-unary--",
 	"vifm-expr1",
 	"vifm-expr2",
 	"vifm-expr3",
+	"vifm-expr4",
 	"vifm-expression-syntax",
 	"vifm-f",
 	"vifm-filetype()",
@@ -460,6 +497,8 @@ const char *tags[] = {
 	"vifm-m_N",
 	"vifm-m_ZQ",
 	"vifm-m_ZZ",
+	"vifm-m_e",
+	"vifm-m_gf",
 	"vifm-m_gg",
 	"vifm-m_j",
 	"vifm-m_k",
@@ -526,6 +565,7 @@ const char *tags[] = {
 	"vifm-q_CTRL-W_v",
 	"vifm-q_CTRL-W_w",
 	"vifm-q_CTRL-W_x",
+	"vifm-q_CTRL-W_z",
 	"vifm-q_CTRL-Y",
 	"vifm-q_Enter",
 	"vifm-q_G",
@@ -566,7 +606,9 @@ const char *tags[] = {
 	"vifm-u",
 	"vifm-v",
 	"vifm-v_:",
+	"vifm-v_CTRL-C",
 	"vifm-v_Enter",
+	"vifm-v_Escape",
 	"vifm-v_O",
 	"vifm-v_U",
 	"vifm-v_V",
