@@ -29,13 +29,13 @@
 #endif
 #include <unistd.h>
 
-#include <assert.h>
+#include <assert.h> /* assert() */
 #include <ctype.h>
 #include <signal.h> /* signal() */
 #include <stdarg.h> /* va_list va_start() va_end() */
 #include <stdlib.h> /* malloc() free() */
 #include <stdio.h> /* snprintf() vsnprintf() */
-#include <string.h> /* strcpy() strlen() */
+#include <string.h> /* memset() strcpy() strlen() */
 #include <time.h>
 
 #include "cfg/config.h"
@@ -48,6 +48,7 @@
 #include "utils/fs_limits.h"
 #include "utils/log.h"
 #include "utils/macros.h"
+#include "utils/path.h"
 #include "utils/str.h"
 #include "utils/utf8.h"
 #include "utils/utils.h"
@@ -58,11 +59,26 @@
 #include "quickview.h"
 #include "signals.h"
 #include "status.h"
+#include "term_title.h"
+
+/* State of cancellation request processing. */
+typedef enum
+{
+	CRS_DISABLED,           /* Cancellation is disabled. */
+	CRS_DISABLED_REQUESTED, /* Cancellation is disabled and was requested. */
+	CRS_ENABLED,            /* Cancellation is enabled, but wasn't requested. */
+	CRS_ENABLED_REQUESTED,  /* Cancellation is enabled and was requested. */
+}
+cancellation_request_state;
 
 static const char PRESS_ENTER_MSG[] = "Press ENTER or type command to continue";
 
 static int multiline_status_bar;
 
+/* Whether cancellation was requested.  Used by ui_cancellation_* group of
+ * functions. */
+static cancellation_request_state cancellation_state;
+
 static WINDOW *ltop_line1;
 static WINDOW *ltop_line2;
 static WINDOW *rtop_line1;
@@ -71,12 +87,16 @@ static WINDOW *rtop_line2;
 static void truncate_with_ellipsis(const char msg[], size_t width,
 		char buffer[]);
 static void update_attributes(void);
+static void update_geometry(void);
 static void update_views(int reload);
 static void reload_lists(void);
 static void reload_list(FileView *view);
+static void swap_view_roles(void);
 static void update_view(FileView *win);
 static void update_window_lazy(WINDOW *win);
 static void switch_panes_content(void);
+static int ui_cancellation_enabled(void);
+static int ui_cancellation_disabled(void);
 
 static void _gnuc_noreturn
 finish(const char *message)
@@ -942,6 +962,8 @@ setup_ncurses_interface(void)
 #endif
 #endif
 
+	update_geometry();
+
 	return 1;
 }
 
@@ -1091,33 +1113,11 @@ static void
 resize_all(void)
 {
 	static float prev_x = -1.f, prev_y = -1.f;
+
 	int screen_x, screen_y;
-#ifndef _WIN32
-	struct winsize ws;
 
-	ioctl(0, TIOCGWINSZ, &ws);
-	LOG_INFO_MSG("ws.ws_row = %d; ws.ws_col = %d", ws.ws_row, ws.ws_col);
-
-	/* changed for pdcurses */
-	resize_term(ws.ws_row, ws.ws_col);
-#endif
-
-#ifdef _WIN32
+	update_geometry();
 	getmaxyx(stdscr, screen_y, screen_x);
-	resize_term(screen_y, screen_x);
-#endif
-
-	getmaxyx(stdscr, screen_y, screen_x);
-	cfg.lines = screen_y;
-	cfg.columns = screen_x;
-
-	if(curr_stats.initial_lines == INT_MIN)
-	{
-		curr_stats.initial_lines = screen_y;
-		curr_stats.initial_columns = screen_x;
-	}
-
-	load_geometry();
 
 	LOG_INFO_MSG("screen_y = %d; screen_x = %d", screen_y, screen_x);
 
@@ -1222,6 +1222,39 @@ resize_all(void)
 	curs_set(FALSE);
 }
 
+/* Updates internal data structures to reflect actual terminal geometry. */
+static void
+update_geometry(void)
+{
+	int screen_x, screen_y;
+#ifndef _WIN32
+	struct winsize ws;
+
+	ioctl(0, TIOCGWINSZ, &ws);
+	LOG_INFO_MSG("ws.ws_row = %d; ws.ws_col = %d", ws.ws_row, ws.ws_col);
+
+	/* changed for pdcurses */
+	resize_term(ws.ws_row, ws.ws_col);
+#endif
+
+#ifdef _WIN32
+	getmaxyx(stdscr, screen_y, screen_x);
+	resize_term(screen_y, screen_x);
+#endif
+
+	getmaxyx(stdscr, screen_y, screen_x);
+	cfg.lines = screen_y;
+	cfg.columns = screen_x;
+
+	if(curr_stats.initial_lines == INT_MIN)
+	{
+		curr_stats.initial_lines = screen_y;
+		curr_stats.initial_columns = screen_x;
+	}
+
+	load_geometry();
+}
+
 void
 update_screen(UpdateType update_kind)
 {
@@ -1309,11 +1342,15 @@ reload_lists(void)
 
 	if(curr_stats.number_of_windows == 2)
 	{
-		update_view_title(other_view);
+		ui_view_title_update(other_view);
 		if(curr_stats.view)
+		{
 			quick_view_file(curr_view);
+		}
 		else if(!other_view->explore_mode)
+		{
 			reload_list(other_view);
+		}
 	}
 }
 
@@ -1328,20 +1365,10 @@ reload_list(FileView *view)
 				!(cfg.vifm_info&VIFMINFO_SAVEDIRS) || view->list_pos != 0);
 }
 
-static void
-switch_views(void)
-{
-	FileView *tmp = curr_view;
-	curr_view = other_view;
-	other_view = tmp;
-
-	load_local_options(curr_view);
-}
-
 void
 change_window(void)
 {
-	switch_views();
+	swap_view_roles();
 
 	if(curr_stats.number_of_windows != 1)
 	{
@@ -1350,7 +1377,6 @@ change_window(void)
 			put_inactive_mark(other_view);
 			erase_current_line_bar(other_view);
 		}
-		update_view_title(other_view);
 	}
 
 	if(curr_stats.view && !is_dir_list_loaded(curr_view))
@@ -1358,27 +1384,19 @@ change_window(void)
 		navigate_to(curr_view, curr_view->curr_dir);
 	}
 
-	update_view_title(curr_view);
+	curr_stats.need_update = UT_REDRAW;
+}
 
-	wnoutrefresh(other_view->win);
-	wnoutrefresh(curr_view->win);
+/* Swaps curr_view and other_view pointers and updates things that are bound to
+ * current view, which is obviously changed after swapping. */
+static void
+swap_view_roles(void)
+{
+	FileView *tmp = curr_view;
+	curr_view = other_view;
+	other_view = tmp;
 
-	(void)change_directory(curr_view, curr_view->curr_dir);
-
-	if(curr_stats.number_of_windows == 1 && window_shows_dirlist(curr_view))
-	{
-		load_dir_list(curr_view, 1);
-	}
-
-	if(!curr_view->explore_mode)
-	{
-		redraw_current_view();
-	}
-	werase(status_bar);
-	wnoutrefresh(status_bar);
-
-	if(curr_stats.number_of_windows == 1)
-		update_all_windows();
+	load_local_options(curr_view);
 }
 
 void
@@ -1608,7 +1626,7 @@ update_attributes(void)
 }
 
 void
-wprint(WINDOW *win, const char *str)
+wprint(WINDOW *win, const char str[])
 {
 #ifndef _WIN32
 	waddstr(win, str);
@@ -1625,6 +1643,14 @@ wprint(WINDOW *win, const char *str)
 #endif
 }
 
+void
+wprinta(WINDOW *win, const char str[], int line_attrs)
+{
+	wattron(win, line_attrs);
+	wprint(win, str);
+	wattroff(win, line_attrs);
+}
+
 void
 request_view_update(FileView *view)
 {
@@ -1801,7 +1827,7 @@ format_entry_name(FileView *view, size_t pos, size_t buf_len, char buf[])
 			entry->name[name_len - 1] = '\0';
 		}
 	}
-	snprintf(buf, buf_len + 1, "%s%s%s", prefix, entry->name, suffix);
+	snprintf(buf, buf_len, "%s%s%s", prefix, entry->name, suffix);
 	/* FIXME: remove this hack for directories. */
 	if(type == DIRECTORY)
 	{
@@ -1827,5 +1853,256 @@ ui_view_win_changed(FileView *view)
 	wnoutrefresh(view->win);
 }
 
+void
+ui_view_reset_selection_and_reload(FileView *view)
+{
+	clean_selected_files(view);
+	load_saving_pos(view, 1);
+}
+
+void
+ui_views_reload_visible_filelists(void)
+{
+	if(curr_stats.view)
+	{
+		load_saving_pos(curr_view, 1);
+	}
+	else
+	{
+		ui_views_reload_filelists();
+	}
+}
+
+void
+ui_views_reload_filelists(void)
+{
+	load_saving_pos(curr_view, 1);
+	load_saving_pos(other_view, 1);
+}
+
+void
+ui_views_update_titles(void)
+{
+	ui_view_title_update(&lwin);
+	ui_view_title_update(&rwin);
+}
+
+void
+ui_view_title_update(FileView *view)
+{
+	char *buf;
+	size_t len;
+	int gen_view = get_mode() == VIEW_MODE && !curr_view->explore_mode;
+	FileView *selected = gen_view ? other_view : curr_view;
+
+	if(curr_stats.load_stage < 2)
+	{
+		return;
+	}
+
+	if(view == selected)
+	{
+		col_attr_t col;
+
+		col = cfg.cs.color[TOP_LINE_COLOR];
+		mix_colors(&col, &cfg.cs.color[TOP_LINE_SEL_COLOR]);
+		init_pair(DCOLOR_BASE + TOP_LINE_SEL_COLOR, col.fg, col.bg);
+
+		wbkgdset(view->title, COLOR_PAIR(DCOLOR_BASE + TOP_LINE_SEL_COLOR) |
+				(col.attr & A_REVERSE));
+		wattrset(view->title, col.attr & ~A_REVERSE);
+	}
+	else
+	{
+		wbkgdset(view->title, COLOR_PAIR(DCOLOR_BASE + TOP_LINE_COLOR) |
+				(cfg.cs.color[TOP_LINE_COLOR].attr & A_REVERSE));
+		wattrset(view->title, cfg.cs.color[TOP_LINE_COLOR].attr & ~A_REVERSE);
+		wbkgdset(top_line, COLOR_PAIR(DCOLOR_BASE + TOP_LINE_COLOR) |
+				(cfg.cs.color[TOP_LINE_COLOR].attr & A_REVERSE));
+		wattrset(top_line, cfg.cs.color[TOP_LINE_COLOR].attr & ~A_REVERSE);
+		werase(top_line);
+	}
+	werase(view->title);
+
+	if(curr_stats.load_stage < 2)
+		return;
+
+	buf = replace_home_part(view->curr_dir);
+	if(view == selected)
+	{
+		set_term_title(replace_home_part(view->curr_dir));
+	}
+
+	if(view->explore_mode)
+	{
+		if(!is_root_dir(buf))
+			strcat(buf, "/");
+		strcat(buf, get_current_file_name(view));
+	}
+	else if(curr_stats.view && view == other_view)
+	{
+		strcpy(buf, "File: ");
+		strcat(buf, get_current_file_name(curr_view));
+	}
+
+	len = get_screen_string_length(buf);
+	if(len > view->window_width + 1 && view == selected)
+	{ /* Truncate long directory names */
+		const char *ptr;
+
+		ptr = buf;
+		while(len > view->window_width - 2)
+		{
+			len--;
+			ptr += get_char_width(ptr);
+		}
+
+		wprintw(view->title, "...");
+		wprint(view->title, ptr);
+	}
+	else if(len > view->window_width + 1 && view != selected)
+	{
+		size_t len = get_normal_utf8_string_widthn(buf, view->window_width - 3 + 1);
+		buf[len] = '\0';
+		wprint(view->title, buf);
+		wprintw(view->title, "...");
+	}
+	else
+	{
+		wprint(view->title, buf);
+	}
+
+	wnoutrefresh(view->title);
+}
+
+int
+ui_view_sort_list_contains(const char sort[SORT_OPTION_COUNT], char key)
+{
+	int i = -1;
+	while(++i < SORT_OPTION_COUNT)
+	{
+		const int sort_key = abs(sort[i]);
+		if(sort_key > LAST_SORT_OPTION)
+		{
+			return 0;
+		}
+		else if(sort_key == key)
+		{
+			return 1;
+		}
+	}
+	return 0;
+}
+
+void
+ui_view_sort_list_ensure_well_formed(char sort[SORT_OPTION_COUNT])
+{
+	int found_name_key = 0;
+	int i = -1;
+	while(++i < SORT_OPTION_COUNT)
+	{
+		const int sort_key = abs(sort[i]);
+		if(sort_key > LAST_SORT_OPTION)
+		{
+			break;
+		}
+		else if(sort_key == SORT_BY_NAME || sort_key == SORT_BY_INAME)
+		{
+			found_name_key = 1;
+		}
+	}
+
+	if(!found_name_key && i < SORT_OPTION_COUNT)
+	{
+		sort[i++] = DEFAULT_SORT_KEY;
+	}
+
+	if(i < SORT_OPTION_COUNT)
+	{
+		memset(&sort[i], NO_SORT_OPTION, SORT_OPTION_COUNT - i);
+	}
+}
+
+int
+ui_view_displays_numbers(const FileView *const view)
+{
+	return view->num_type != NT_NONE && !view->ls_view;
+}
+
+int
+ui_view_is_visible(const FileView *const view)
+{
+	return curr_stats.number_of_windows == 2 || curr_view == view;
+}
+
+void
+ui_cancellation_reset(void)
+{
+	assert(ui_cancellation_disabled() && "Can't reset while active.");
+
+	cancellation_state = CRS_DISABLED;
+}
+
+void
+ui_cancellation_enable(void)
+{
+	assert(ui_cancellation_disabled() && "Can't enable twice in a row.");
+
+	cancellation_state = (cancellation_state == CRS_DISABLED)
+	                   ? CRS_ENABLED
+	                   : CRS_ENABLED_REQUESTED;
+
+	/* Temporary disable raw mode of terminal so that Ctrl-C is be handled as
+	 * SIGINT signal rather than as regular input character. */
+	noraw();
+}
+
+void
+ui_cancellation_request(void)
+{
+	if(ui_cancellation_enabled())
+	{
+		cancellation_state = CRS_ENABLED_REQUESTED;
+	}
+}
+
+int
+ui_cancellation_requested(void)
+{
+	return cancellation_state == CRS_ENABLED_REQUESTED
+	    || cancellation_state == CRS_DISABLED_REQUESTED;
+}
+
+void
+ui_cancellation_disable(void)
+{
+	assert(ui_cancellation_enabled() && "Can't disable what disabled.");
+
+	/* Restore raw mode of terminal so that Ctrl-C is be handled as regular input
+	 * character rather than as SIGINT signal. */
+	raw();
+
+	cancellation_state = (cancellation_state == CRS_ENABLED_REQUESTED)
+	                   ? CRS_DISABLED_REQUESTED
+	                   : CRS_DISABLED;
+}
+
+/* Checks whether cancellation processing is enabled.  Returns non-zero if so,
+ * otherwise zero is returned. */
+static int
+ui_cancellation_enabled(void)
+{
+	return cancellation_state == CRS_ENABLED
+	    || cancellation_state == CRS_ENABLED_REQUESTED;
+}
+
+/* Checks whether cancellation processing is disabled.  Returns non-zero if so,
+ * otherwise zero is returned. */
+static int
+ui_cancellation_disabled(void)
+{
+	return !ui_cancellation_enabled();
+}
+
 /* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab cinoptions-=(0 : */
 /* vim: set cinoptions+=t0 : */
