@@ -36,6 +36,7 @@
 #include "utils/str.h"
 #include "ops.h"
 #include "registers.h"
+#include "trash.h"
 
 typedef struct
 {
@@ -93,7 +94,7 @@ static OPS undo_op[] = {
 	OP_REMOVE,   /* OP_SYMLINK */
 	OP_REMOVESL, /* OP_SYMLINK2 */
 	OP_RMDIR,    /* OP_MKDIR */
-	OP_NONE,     /* OP_RMDIR */
+	OP_MKDIR,    /* OP_RMDIR */
 	OP_REMOVE,   /* OP_MKFILE */
 };
 ARRAY_GUARD(undo_op, OP_COUNT);
@@ -151,7 +152,7 @@ static enum
 	{ OPER_1ST, OPER_NON, OPER_NON, OPER_1ST,    /* do   OP_MKDIR */
 		OPER_1ST, OPER_NON, OPER_1ST, OPER_NON, }, /* undo OP_RMDIR */
 	{ OPER_1ST, OPER_NON, OPER_1ST, OPER_NON,    /* do   OP_RMDIR */
-		OPER_NON, OPER_NON, OPER_NON, OPER_NON, }, /* undo OP_NONE  */
+		OPER_1ST, OPER_NON, OPER_NON, OPER_1ST, }, /* undo OP_MKDIR */
 	{ OPER_1ST, OPER_NON, OPER_NON, OPER_1ST,    /* do   OP_MKFILE */
 		OPER_1ST, OPER_NON, OPER_1ST, OPER_NON, }, /* undo OP_REMOVE  */
 };
@@ -192,6 +193,8 @@ static perform_func do_func;
 /* External function, which corrects operation availability and influence on
  * operation checks. */
 static op_available_func op_avail_func;
+/* External optional callback to abort execution of compound operations. */
+static undo_cancel_requested cancel_func;
 /* Number of undo levels, which are not groups but operations. */
 static const int *undo_levels;
 
@@ -200,8 +203,6 @@ static cmd_t cmds = {
 };
 static cmd_t *current = &cmds;
 
-static size_t trash_dir_len;
-
 static int group_opened;
 static long long next_group;
 static group_t *last_group;
@@ -209,6 +210,7 @@ static char *group_msg;
 
 static int command_count;
 
+static int no_function(void);
 static void init_cmd(cmd_t *cmd, OPS op, void *do_data, void *undo_data);
 static void init_entry(cmd_t *cmd, const char **e, int type);
 static void remove_cmd(cmd_t *cmd);
@@ -223,15 +225,21 @@ static char **fill_undolist_nondetail(char **list);
 
 void
 init_undo_list(perform_func exec_func, op_available_func op_avail,
-		const int* max_levels)
+		undo_cancel_requested cancel, const int* max_levels)
 {
 	assert(exec_func != NULL);
 
 	do_func = exec_func;
 	op_avail_func = op_avail;
+	cancel_func = (cancel != NULL) ? cancel : &no_function;
 	undo_levels = max_levels;
+}
 
-	trash_dir_len = strlen(cfg.trash_dir);
+/* Always says no.  Returns zero. */
+static int
+no_function(void)
+{
+	return 0;
 }
 
 void
@@ -454,6 +462,7 @@ undo_group(void)
 {
 	int errors, disbalance, cant_undone;
 	int skip;
+	int cancelled;
 	assert(!group_opened);
 
 	if(current == &cmds)
@@ -499,12 +508,21 @@ undo_group(void)
 		}
 		current = current->prev;
 	}
-	while(current != &cmds && current->group == current->next->group);
+	while(!(cancelled = cancel_func()) && current != &cmds &&
+			current->group == current->next->group);
 
-	if(skip)
+	if(cancelled)
+	{
+		return -7;
+	}
+	else if(skip)
+	{
 		return -6;
-
-	return errors ? -2 : 0;
+	}
+	else
+	{
+		return errors ? -2 : 0;
+	}
 }
 
 static int
@@ -530,6 +548,7 @@ redo_group(void)
 {
 	int errors, disbalance;
 	int skip;
+	int cancelled;
 	assert(!group_opened);
 
 	if(current->next == NULL)
@@ -572,12 +591,21 @@ redo_group(void)
 			}
 		}
 	}
-	while(current->next != NULL && current->group == current->next->group);
+	while(!(cancelled = cancel_func()) && current->next != NULL &&
+			current->group == current->next->group);
 
-	if(skip)
+	if(cancelled)
+	{
+		return -7;
+	}
+	else if(skip)
+	{
 		return -6;
-
-	return errors ? -2 : 0;
+	}
+	else
+	{
+		return errors ? -2 : 0;
+	}
 }
 
 static int
@@ -622,7 +650,7 @@ is_op_possible(const op_t *op)
 		return 0;
 	if(op->dont_exist != NULL && lstat(op->dont_exist, &st) == 0)
 	{
-		if(strnoscmp(op->dst, cfg.trash_dir, trash_dir_len) == 0)
+		if(is_under_trash(op->dst))
 			return -1;
 		return 0;
 	}
@@ -632,23 +660,23 @@ is_op_possible(const op_t *op)
 static void
 change_filename_in_trash(cmd_t *cmd, const char *filename)
 {
-	char *p;
-	int i;
-	char buf[PATH_MAX];
+	const char *name_tail;
+	char *new;
 	char *old;
+	char *const base_dir = strdup(filename);
 
-	p = strchr(filename + trash_dir_len, '_') + 1;
+	remove_last_path_component(base_dir);
 
-	i = -1;
-	do
-	{
-		snprintf(buf, sizeof(buf), "%s/%03i_%s", cfg.trash_dir, ++i, p);
-	}
-	while(path_exists(buf));
-	rename_in_registers(filename, buf);
+	name_tail = get_real_name_from_trash_name(filename);
+	new = gen_trash_name(base_dir, name_tail);
+	assert(new != NULL && "Should always get trash name here.");
+
+	free(base_dir);
 
 	old = cmd->buf2;
-	cmd->buf2 = strdup(buf);
+	cmd->buf2 = new;
+
+	rename_in_registers(filename, new);
 
 	update_entry(&cmd->do_op.src, old, cmd->buf2);
 	update_entry(&cmd->do_op.dst, old, cmd->buf2);
@@ -874,14 +902,12 @@ clean_cmds_with_trash(void)
 
 		if(cur->group->balance < 0)
 		{
-			if(cur->do_op.exists != NULL &&
-					strnoscmp(cur->do_op.exists, cfg.trash_dir, trash_dir_len) == 0)
+			if(cur->do_op.exists != NULL && is_under_trash(cur->do_op.exists))
 				remove_cmd(cur);
 		}
 		else
 		{
-			if(cur->undo_op.exists != NULL &&
-					strnoscmp(cur->undo_op.exists, cfg.trash_dir, trash_dir_len) == 0)
+			if(cur->undo_op.exists != NULL && is_under_trash(cur->undo_op.exists))
 				remove_cmd(cur);
 		}
 		cur = prev;
