@@ -26,6 +26,7 @@
 
 #include <stddef.h> /* size_t */
 #include <stdint.h> /* uint64_t */
+#include <stdio.h> /* FILE */
 #include <wchar.h> /* wchar_t */
 
 /* Regular expressions. */
@@ -33,50 +34,89 @@
 /* Gets flags for compiling a regular expression specified by the pattern taking
  * 'ignorecase' and 'smartcase' options into account.  Returns regex flags. */
 int get_regexp_cflags(const char pattern[]);
+
 /* Decides whether case should be ignored for the pattern.  Considers
  * 'ignorecase' and 'smartcase' options.  Returns non-zero when case should be
  * ignored, otherwise zero is returned. */
 int regexp_should_ignore_case(const char pattern[]);
+
 const char * get_regexp_error(int err, regex_t *re);
 
 /* Shell and program running. */
 
 /* Executes an external command.  Clears the screen up on Windows before running
  * the command.  Returns error code, which is zero on success. */
-int my_system(char command[]);
+int vifm_system(char command[]);
+
 /* Pauses shell.  Assumes that curses interface is off. */
 void pause_shell(void);
+
 /* Called after return from a shellout to provide point to recover UI. */
 void recover_after_shellout(void);
 
 /* Other functions. */
 
+struct mntent;
+
+/* Client of the traverse_mount_points() function.  Should return non-zero to
+ * stop traversal. */
+typedef int (*mptraverser)(struct mntent *entry, void *arg);
+
 int is_on_slow_fs(const char full_path[]);
+
 /* Fills supplied buffer with user friendly representation of file size.
  * Returns non-zero in case resulting string is a shortened variant of size. */
 int friendly_size_notation(uint64_t num, int str_size, char *str);
-const char * enclose_in_dquotes(const char *str);
-int my_chdir(const char *path);
+
+/* Returns pointer to a statically allocated buffer. */
+const char * enclose_in_dquotes(const char str[]);
+
+/* Changes current working directory of the process.  Does nothing if we already
+ * at path.  Returns zero on success, otherwise -1 is returned. */
+int vifm_chdir(const char path[]);
+
 /* Expands all environment variables and tilde in the path.  Allocates
  * memory, that should be freed by the caller. */
 char * expand_path(const char path[]);
+
 /* Expands all environment variables in the str of form "$envvar".  Non-zero
  * escape_vals means escaping suitable for internal use.  Allocates and returns
  * memory that should be freed by the caller. */
 char * expand_envvars(const char str[], int escape_vals);
+
 /* Makes filename unique by adding an unique suffix to it.
  * Returns pointer to a statically allocated buffer */
 const char * make_name_unique(const char filename[]);
+
 /* Returns process identification in a portable way. */
 unsigned int get_pid(void);
+
 /* Finds command name in the command line and writes it to the buf.
  * Raw mode will preserve quotes on Windows.
  * Returns a pointer to the argument list. */
 char * extract_cmd_name(const char line[], int raw, size_t buf_len, char buf[]);
+
 /* Determines columns needed for a wide character.  Returns number of columns,
  * on error default value of 1 is returned. */
 int vifm_wcwidth(wchar_t c);
 
+/* Fills buf of the length buf_len with path to mount point of the path.
+ * Returns non-zero on error, otherwise zero is returned. */
+int get_mount_point(const char path[], size_t buf_len, char buf[]);
+
+/* Calls client traverser for each mount point.  Returns non-zero on error,
+ * otherwise zero is returned. */
+int traverse_mount_points(mptraverser client, void *arg);
+
+/* Waits until non-blocking read operation is available for given file
+ * descriptor (uses f if it's not NULL, otherwise fd is used) that is associated
+ * with a process.  Process operation cancellation requests from a user. */
+void wait_for_data_from(pid_t pid, FILE *f, int fd);
+
+/* Blocks/unblocks SIGCHLD signal.  Returns zero on success, otherwise non-zero
+ * is returned. */
+int set_sigchld(int block);
+
 #ifdef _WIN32
 #include "utils_win.h"
 #else
