@@ -21,27 +21,52 @@
 #include "utils_int.h"
 
 #include <sys/stat.h> /* S_* */
-#include <sys/types.h> /* gid_t mode_t uid_t */
-#include <ctype.h> /* isdigit() */
-#include <errno.h> /* EINTR errno */
+#include <sys/types.h> /* gid_t mode_t pid_t uid_t */
 #include <fcntl.h> /* O_RDONLY open() close() */
 #include <grp.h> /* getgrnam() */
 #include <pwd.h> /* getpwnam() */
+#include <unistd.h> /* dup2() getpid() */
+
+#include <assert.h> /* assert() */
+#include <ctype.h> /* isdigit() */
+#include <errno.h> /* EINTR errno */
 #include <signal.h> /* signal() SIGINT SIGTSTP SIGCHLD SIG_DFL sigset_t
                        sigemptyset() sigaddset() sigprocmask() SIG_BLOCK
                        SIG_UNBLOCK */
-#include <stddef.h> /* NULL */
+#include <stddef.h> /* NULL size_t */
 #include <stdio.h> /* snprintf() */
 #include <stdlib.h> /* atoi() */
 #include <string.h> /* strchr() strlen() strncmp() */
-#include <unistd.h> /* dup2() getpid() */
 
 #include "../cfg/config.h"
 #include "fs_limits.h"
 #include "log.h"
 #include "mntent.h" /* mntent setmntent() getmntent() endmntent() */
 #include "path.h"
+#include "str.h"
+#include "utils.h"
 
+/* Types of mount point information for get_mount_point_traverser_state. */
+typedef enum
+{
+	MI_MOUNT_POINT, /* Path to the mount point. */
+	MI_FS_TYPE,     /* File system name. */
+}
+mntinfo;
+
+/* State for get_mount_point() traverser. */
+typedef struct
+{
+	mntinfo type;     /* Type of information to put into the buffer. */
+	const char *path; /* Path whose mount point we're looking for. */
+	size_t buf_len;   /* Output buffer length. */
+	char *buf;        /* Output buffer. */
+	size_t curr_len;  /* Max length among all found points (len of buf string). */
+}
+get_mount_point_traverser_state;
+
+static void process_cancel_request(pid_t pid);
+static int get_mount_info_traverser(struct mntent *entry, void *arg);
 static int begins_with_list_item(const char pattern[], const char list[]);
 
 void
@@ -59,7 +84,6 @@ run_in_shell_no_cls(char command[])
 	int result;
 	extern char **environ;
 	sig_handler sigtstp_handler;
-	sigset_t sigchld_mask;
 
 	if(command == NULL)
 		return 1;
@@ -69,15 +93,13 @@ run_in_shell_no_cls(char command[])
 	/* We need to block SIGCHLD signal.  One can't just set it to SIG_DFL, because
 	 * it will possibly cause missing of SIGCHLD from a background process
 	 * (job). */
-	sigemptyset(&sigchld_mask);
-	sigaddset(&sigchld_mask, SIGCHLD);
-	sigprocmask(SIG_BLOCK, &sigchld_mask, NULL);
+	(void)set_sigchld(1);
 
 	pid = fork();
 	if(pid == -1)
 	{
 		signal(SIGTSTP, sigtstp_handler);
-		sigprocmask(SIG_UNBLOCK, &sigchld_mask, NULL);
+		(void)set_sigchld(0);
 		return -1;
 	}
 	if(pid == 0)
@@ -86,7 +108,7 @@ run_in_shell_no_cls(char command[])
 
 		signal(SIGTSTP, SIG_DFL);
 		signal(SIGINT, SIG_DFL);
-		sigprocmask(SIG_UNBLOCK, &sigchld_mask, NULL);
+		(void)set_sigchld(0);
 
 		args[0] = cfg.shell;
 		args[1] = "-c";
@@ -95,6 +117,75 @@ run_in_shell_no_cls(char command[])
 		execve(cfg.shell, args, environ);
 		exit(127);
 	}
+
+	result = get_proc_exit_status(pid);
+
+	signal(SIGTSTP, sigtstp_handler);
+	(void)set_sigchld(0);
+
+	return result;
+}
+
+void
+recover_after_shellout(void)
+{
+	/* Do nothing.  No need to recover anything on this platform. */
+}
+
+void
+wait_for_data_from(pid_t pid, FILE *f, int fd)
+{
+	const struct timeval ts_init = { .tv_sec = 0, .tv_usec = 1000 };
+	struct timeval ts;
+
+	fd_set read_ready;
+	FD_ZERO(&read_ready);
+
+	fd = (f != NULL) ? fileno(f) : fd;
+
+	do
+	{
+		process_cancel_request(pid);
+		ts = ts_init;
+		FD_SET(fd, &read_ready);
+	}
+	while(select(fd + 1, &read_ready, NULL, NULL, &ts) == 0);
+
+	/* Inform other parts of the application that cancellation took place. */
+	if(errno == EINTR)
+	{
+		ui_cancellation_request();
+	}
+}
+
+int
+set_sigchld(int block)
+{
+	const int action = block ? SIG_BLOCK : SIG_UNBLOCK;
+	sigset_t sigchld_mask;
+
+	return sigemptyset(&sigchld_mask) == -1
+	    || sigaddset(&sigchld_mask, SIGCHLD) == -1
+	    || sigprocmask(action, &sigchld_mask, NULL) == -1;
+}
+
+/* Checks whether cancelling of current operation is requested and sends SIGINT
+ * to process specified by its process id to request cancellation. */
+static void
+process_cancel_request(pid_t pid)
+{
+	if(ui_cancellation_requested())
+	{
+		if(kill(pid, SIGINT) != 0)
+		{
+			LOG_SERROR_MSG(errno, "Failed to send SIGINT to " PRINTF_PID_T, pid);
+		}
+	}
+}
+
+int
+get_proc_exit_status(pid_t pid)
+{
 	do
 	{
 		int status;
@@ -103,25 +194,15 @@ run_in_shell_no_cls(char command[])
 			if(errno != EINTR)
 			{
 				LOG_SERROR_MSG(errno, "waitpid()");
-				result = -1;
-				break;
+				return -1;
 			}
 		}
 		else
 		{
-			result = status;
-			break;
+			return status;
 		}
-	}while(1);
-	signal(SIGTSTP, sigtstp_handler);
-	sigprocmask(SIG_UNBLOCK, &sigchld_mask, NULL);
-	return result;
-}
-
-void
-recover_after_shellout(void)
-{
-	/* Do nothing.  No need to recover anything on this platform. */
+	}
+	while(1);
 }
 
 /* if err == 1 then use stderr and close stdin and stdout */
@@ -195,36 +276,93 @@ get_perm_string(char buf[], int len, mode_t mode)
 int
 is_on_slow_fs(const char full_path[])
 {
-	FILE *f;
-	struct mntent *ent;
-	size_t len = 0;
-	char max[PATH_MAX] = "";
+	char fs_name[PATH_MAX];
+	get_mount_point_traverser_state state =
+	{
+		.type = MI_FS_TYPE,
+		.path = full_path,
+		.buf_len = sizeof(fs_name),
+		.buf = fs_name,
+		.curr_len = 0UL,
+	};
 
+	/* Empty list optimization. */
 	if(cfg.slow_fs_list[0] == '\0')
 	{
 		return 0;
 	}
 
+	if(traverse_mount_points(&get_mount_info_traverser, &state) == 0)
+	{
+		if(state.curr_len > 0)
+		{
+			return begins_with_list_item(fs_name, cfg.slow_fs_list);
+		}
+	}
+	return 0;
+}
+
+int
+get_mount_point(const char path[], size_t buf_len, char buf[])
+{
+	get_mount_point_traverser_state state =
+	{
+		.type = MI_MOUNT_POINT,
+		.path = path,
+		.buf_len = buf_len,
+		.buf = buf,
+		.curr_len = 0UL,
+	};
+	return traverse_mount_points(&get_mount_info_traverser, &state);
+}
+
+/* traverse_mount_points client that gets mount point info for a given path. */
+static int
+get_mount_info_traverser(struct mntent *entry, void *arg)
+{
+	get_mount_point_traverser_state *const state = arg;
+	if(path_starts_with(state->path, entry->mnt_dir))
+	{
+		const size_t new_len = strlen(entry->mnt_dir);
+		if(new_len > state->curr_len)
+		{
+			state->curr_len = new_len;
+			switch(state->type)
+			{
+				case MI_MOUNT_POINT:
+					copy_str(state->buf, state->buf_len, entry->mnt_dir);
+					break;
+				case MI_FS_TYPE:
+					copy_str(state->buf, state->buf_len, entry->mnt_type);
+					break;
+
+				default:
+					assert(0 && "Unknown mount information type.");
+					break;
+			}
+		}
+	}
+	return 0;
+}
+
+int
+traverse_mount_points(mptraverser client, void *arg)
+{
+	FILE *f;
+	struct mntent *ent;
+
 	if((f = setmntent("/etc/mtab", "r")) == NULL)
 	{
-		return 0;
+		return 1;
 	}
 
 	while((ent = getmntent(f)) != NULL)
 	{
-		if(path_starts_with(full_path, ent->mnt_dir))
-		{
-			const size_t new_len = strlen(ent->mnt_dir);
-			if(new_len > len)
-			{
-				len = new_len;
-				snprintf(max, sizeof(max), "%s", ent->mnt_fsname);
-			}
-		}
+		client(ent, arg);
 	}
 
 	endmntent(f);
-	return (max[0] == '\0') ? 0 : begins_with_list_item(max, cfg.slow_fs_list);
+	return 0;
 }
 
 static int
