@@ -29,6 +29,7 @@
 
 #include <locale.h> /* setlocale */
 #include <stdio.h> /* fputs() puts() */
+#include <stdlib.h> /* exit() */
 #include <string.h>
 
 #include "cfg/config.h"
@@ -68,6 +69,7 @@
 #include "signals.h"
 #include "sort.h"
 #include "status.h"
+#include "trash.h"
 #include "ui.h"
 #include "undo.h"
 #include "version.h"
@@ -78,7 +80,7 @@
 #define CONF_DIR "(%HOME%/.vifm or %APPDATA%/Vifm)"
 #endif
 
-static void quit_on_invalid_arg(void);
+static void quit_on_arg_parsing(void);
 static void parse_recieved_arguments(char *args[]);
 static void remote_cd(FileView *view, const char *path, int handle);
 static void load_scheme(void);
@@ -114,8 +116,10 @@ show_help_msg(void)
 	puts("  If no path is given vifm will start in the current working directory.\n");
 	puts("  vifm --logging");
 	puts("    log some errors to " CONF_DIR "/log.\n");
+#ifdef ENABLE_REMOTE_CMDS
 	puts("  vifm --remote");
 	puts("    passes all arguments that left in command line to active vifm server.\n");
+#endif
 	puts("  vifm -c <command> | +<command>");
 	puts("    run <command> on startup.\n");
 	puts("  vifm --version | -v");
@@ -165,7 +169,7 @@ parse_args(int argc, char *argv[], const char *dir, char *lwin_path,
 	int x;
 	int select = 0;
 
-	(void)my_chdir(dir);
+	(void)vifm_chdir(dir);
 
 	/* Get Command Line Arguments */
 	for(x = 1; x < argc; x++)
@@ -174,14 +178,16 @@ parse_args(int argc, char *argv[], const char *dir, char *lwin_path,
 		{
 			select = 1;
 		}
+#ifdef ENABLE_REMOTE_CMDS
 		else if(!strcmp(argv[x], "--remote"))
 		{
 			if(!ipc_server())
 			{
 				ipc_send(argv + x + 1);
-				quit_on_invalid_arg();
+				quit_on_arg_parsing();
 			}
 		}
+#endif
 		else if(!strcmp(argv[x], "-f"))
 		{
 			cfg.vim_filter = 1;
@@ -192,12 +198,12 @@ parse_args(int argc, char *argv[], const char *dir, char *lwin_path,
 		else if(!strcmp(argv[x], "--version") || !strcmp(argv[x], "-v"))
 		{
 			show_version_msg();
-			quit_on_invalid_arg();
+			quit_on_arg_parsing();
 		}
 		else if(!strcmp(argv[x], "--help") || !strcmp(argv[x], "-h"))
 		{
 			show_help_msg();
-			quit_on_invalid_arg();
+			quit_on_arg_parsing();
 		}
 		else if(!strcmp(argv[x], "--logging"))
 		{
@@ -208,7 +214,7 @@ parse_args(int argc, char *argv[], const char *dir, char *lwin_path,
 			if(x == argc - 1)
 			{
 				puts("Argument missing after \"-c\"");
-				quit_on_invalid_arg();
+				quit_on_arg_parsing();
 			}
 			/* do nothing, it's handeled in exec_startup_commands() */
 			x++;
@@ -235,20 +241,26 @@ parse_args(int argc, char *argv[], const char *dir, char *lwin_path,
 		else if(curr_stats.load_stage == 0)
 		{
 			show_help_msg();
-			quit_on_invalid_arg();
+			quit_on_arg_parsing();
 		}
+#ifdef ENABLE_REMOTE_CMDS
 		else
 		{
 			show_error_msgf("--remote error", "Invalid argument: %s", argv[x]);
 		}
+#endif
 	}
 }
 
+/* Quits during argument parsing when it's allowed (e.g. not for remote
+ * commands). */
 static void
-quit_on_invalid_arg(void)
+quit_on_arg_parsing(void)
 {
 	if(curr_stats.load_stage == 0)
+	{
 		exit(1);
+	}
 }
 
 static void
@@ -306,7 +318,7 @@ main(int argc, char *argv[])
 	init_builtin_functions();
 	update_path_env(1);
 
-	if(init_status() != 0)
+	if(init_status(&cfg) != 0)
 	{
 		puts("Error during session status initialization.");
 		return -1;
@@ -358,7 +370,8 @@ main(int argc, char *argv[])
 
 	colmgr_init(COLOR_PAIRS);
 	init_modes();
-	init_undo_list(&perform_operation, NULL, &cfg.undo_levels);
+	init_undo_list(&perform_operation, NULL, &ui_cancellation_requested,
+			&cfg.undo_levels);
 	load_local_options(curr_view);
 
 	curr_stats.load_stage = 1;
@@ -389,7 +402,9 @@ main(int argc, char *argv[])
 		source_config();
 	}
 
-	(void)create_trash_dir(cfg.trash_dir);
+	/* Ensure trash directories exist, it might not have been called during
+	 * configuration file sourcing if there is no `set trashdir=...` command. */
+	(void)set_trash_dir(cfg.trash_dir);
 
 	check_path_for_file(&lwin, lwin_path, lwin_handle);
 	check_path_for_file(&rwin, rwin_path, rwin_handle);
@@ -552,36 +567,37 @@ static int
 run_converter(int vifm_like_mode)
 {
 #ifndef _WIN32
-	char buf[PATH_MAX];
-	snprintf(buf, sizeof(buf), "vifmrc-converter %d", vifm_like_mode);
-	return shellout(buf, -1, 1);
+	char cmd[PATH_MAX];
+	snprintf(cmd, sizeof(cmd), "vifmrc-converter %d", vifm_like_mode);
+	return shellout(cmd, -1, 0);
 #else
-	TCHAR buf[PATH_MAX + 2];
-	TCHAR *last_path_component;
+	char cmd[2*PATH_MAX];
 	int returned_exit_code;
+	char *name_part;
 
-	if(GetModuleFileName(NULL, buf, ARRAY_LEN(buf)) == 0)
+	if(GetModuleFileName(NULL, cmd, PATH_MAX) == 0)
+	{
 		return -1;
+	}
 
-	/* Remove last path component. */
-	last_path_component = _tcsrchr(buf, _T('\\')) + 1;
-	*last_path_component = _T('\0');
-
+	/* Override last path component. */
+	name_part = strrchr(cmd, '\\');
+	name_part = (name_part == NULL) ? cmd : (name_part + 1);
 	switch(vifm_like_mode)
 	{
 		case 2:
-			_tcscat(buf, _T("vifmrc-converter 2"));
+			strcpy(name_part, "vifmrc-converter 2");
 			break;
 		case 1:
-			_tcscat(buf, _T("vifmrc-converter 1"));
+			strcpy(name_part, "vifmrc-converter 1");
 			break;
 
 		default:
-			_tcscat(buf, _T("vifmrc-converter 0"));
+			strcpy(name_part, "vifmrc-converter 0");
 			break;
 	}
 
-	return win_exec_cmd(buf, &returned_exit_code);
+	return win_exec_cmd(cmd, &returned_exit_code);
 #endif
 }
 
