@@ -6,7 +6,7 @@
 static void
 test_unexistant_bookmark(void)
 {
-	assert_int_equal(1, move_to_bookmark(&lwin, 'b'));
+	assert_int_equal(1, goto_bookmark(&lwin, 'b'));
 }
 
 void
@@ -19,4 +19,5 @@ bookmarks_tests(void)
 	test_fixture_end();
 }
 
-/* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab : */
+/* vim: set tabstop=2 softtabstop=2 shiftwidth=2 noexpandtab cinoptions-=(0 : */
+/* vim: set cinoptions+=t0 : */
